#! /usr/bin/guile -s
!#

;; Export the help tree as HTML.

(use-modules (srfi srfi-13)
             (ice-9 regex)
             (soundscrape)
             (soundscrape ui)
             (soundscrape ui stext html)
             (soundscrape ui stext plain-text)
             (soundscrape ui help)
             (soundscrape ui help-tree)
             (gnome gtk)
             (soundscrape util))

(if (not (eq? (length (program-arguments)) 2))
    (begin
      (format #t "Usage: make-html.scm DOCUMENT-PATH\n")
      (exit 1)))

(define document-path (cadr (program-arguments)))

(if (not (and (file-exists? document-path)
              (file-is-directory? document-path)))
    (error (string-append "Error: " document-path " is not a valid directory")))

(format #t "Building documents in ~A...\n" document-path)

;; A global variable, I couldn't avoid it because of the ref resolver...
(define current-iter (make-fluid))
(define tree the-help-tree) ;; just for convenience

(define (write-html-file path iter)
  (with-fluids*
   (list current-iter) (list iter)
   (lambda ()
     (let* ((children (iter-children tree iter))
            (leaf-node? (null? children))
            (depth ((if leaf-node? 1- noop)
                    (length (get-path tree iter))))
            (header-path (apply build-path (make-list depth ".."))))
       (define (post-processor head body)
         (define (link-tail iter)
           (string-append
            (urlify (get-value tree iter 0))
            (if (zero? (iter-n-children tree iter))
                ".php"
                "/")))
         (define (make-navigation)
           (let* ((siblings (iter-children tree (iter-parent tree iter)))
                  (index (car (reverse (get-path tree iter))))
                  (numsiblings (length siblings))
                  (prev (if (zero? index)
                            #f
                            (list-ref siblings (1- index))))
                  (next (if (eq? index (1- (length siblings)))
                            #f
                            (list-ref siblings (1+ index))))
                  (next-following (if leaf-node?
                                      next
                                      (car children)))
                  (up-path (if leaf-node? "./" "../")))
             `(div :class "book-navigation reversed"
                   ,(if prev
                        `(a :href ,(string-append
                                    (if leaf-node? "" "../")
                                    (link-tail prev))
                            "&lt;")
                        "&lt;") " | "
                   (a :href ,up-path
                      "^") " | "
                   ,(if next
                        `(a :href ,(string-append
                                    (if leaf-node? "" "../")
                                    (link-tail next))
                            "&gt;")
                        "&gt;") " | "
                   (a :href ,(if next-following
                                 (link-tail next-following)
                                 up-path)
                      ":&gt;"))))
         (define (make-subsections)
           `(p :class "subsections"
                 ,@(list-join
                    (map
                     (lambda (node)
                       `(a :href ,(link-tail node)
                           ,(get-value tree node 0)))
                     children)
                    " | ")))
         (list
          (list
           'raw
           (format #f "<?php
 include('~A'); 
 echo get_header('soundscrape: documentation: ~A', '~A');
?>
"
                   (build-path header-path "../lib/template.php")
                   (stext->plain-text-title (cons 'text (cdr head)))
                   (build-path header-path "../")))
          (make-navigation)
          (list 'h2 (stext->plain-text-title (cons 'text (cdr head))))
          (make-subsections)
          body
          (make-navigation)
          '(raw "<?php echo get_footer(); ?>\n")))
       (let* ((file (open-output-file path))
              (stext (get-value tree iter 1))
              (html (stext->html stext `(post-processor . ,post-processor))))
         (display html file)
         (close-port file))))))

(define (maybe-mkdir path)
  (if (or (not (file-exists? path)) (not (file-is-directory? path)))
      (mkdir path)))

(define (urlify str)
  (string-downcase
   (string-map
    (lambda (c)
      (case c
        ((#\space #\/ #\:) #\-)
        (else c)))
    str)))

(define (make-document-path parent-path . iters)
  (apply
   build-path
   parent-path
   (map
    (lambda (iter)
      (urlify (get-value tree iter 0)))
    iters)))

(define (write-iter parent-path iter)
  (let ((children (iter-children tree iter)))
    (cond
     ((null? children)
      ;; A leaf node
      (let ((path (string-append parent-path ".php")))
        (format #t "Making leaf node ~A...\n" path)
        (write-html-file path iter)))
     (else
      ;; A branch
      (maybe-mkdir parent-path)
      (let ((path (build-path parent-path "index.php")))
        (format #t "Making branch node ~A...\n" path)
        (write-html-file path iter))
      (for-each
       (lambda (iter)
         (write-iter (make-document-path parent-path iter) iter))
       children)))))

;; Populate the tree
((module-ref (resolve-module '(soundscrape ui help)) 'populate-help))

;; Similar to select-node.
(define (ref-resolver node-name manual-name)
  (define (find-node-named top name)
    (define (contains-node? node)
      (let loop ((tokens (get-value tree node 1)))
        (cond
         ((null? tokens)
          #f)
         ((and (list? (car tokens))
               (eq? (caar tokens) 'node)
               (equal? (cadar tokens) name))
          #t)
         (else
          (loop (cdr tokens))))))
    (or-map
     (lambda (node)
       (if (contains-node? node)
           node
           (find-node-named node name)))
     (iter-children tree top)))
  
  (let* ((current-iter (fluid-ref current-iter))
         (manual (if manual-name
                     (or (or-map
                          (lambda (iter)
                            (if (string=? (cadar (get-value tree iter 1))
                                          manual-name)
                                iter
                                #f))
                          (iter-children tree #f))
                         (error "Unknown manual:" manual-name))
                     (let find-top ((iter current-iter))
                       (let ((parent (iter-parent tree iter)))
                         (if (not parent)
                             iter
                             (find-top parent))))))
         (node (if (and node-name (not (string=? node-name "top")))
                   (or
                    (find-node-named manual node-name)
                    (error "No such node in manual:" manual-name node-name))
                   manual))
         (leaf-node? (null? (iter-children tree current-iter)))
         (leaf-target? (null? (iter-children tree node)))
         (depth ((if leaf-node? 1- noop)
                 (length (get-path tree current-iter))))
         (top-of-docs-path (apply build-path (make-list depth ".."))))
    (string-append
     (apply
      make-document-path
      top-of-docs-path
      (let loop ((nodes (list node)))
        (let ((parent (iter-parent tree (car nodes))))
          (if (not parent)
              nodes
              (loop (cons parent nodes))))))
     (if leaf-target? ".php" "/"))))
         
;; Install our own ref resolver.
(set! html-ref-resolvers
      (cons ref-resolver html-ref-resolvers))

;; finally, do the work.
(for-each
 (lambda (iter)
   (write-iter (make-document-path document-path iter) iter))
 (iter-children tree #f))
