;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;The core synthesis engine of soundscrape.
;;
;;; Code:

(read-set! keywords 'prefix)

(define-module (soundscrape core)
  :use-module (soundscrape util)
  :use-module (soundscrape gst-util)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-13)
  :use-module (gnome gstreamer)
  :export (<ss-pad>
           <ss-element>
           ss-audiosink
           ss-buffer-frames
           ss-filesink
           ss-float2int
           ss-main-loop
           ss-main-pipeline
           ss-plugin
           ss-sample-rate
           graph-fold
           play
           record
           render
           run))

(use-modules ((srfi srfi-1)
              :select (filter filter-map zip delete-duplicates)))
(use-modules ((gnome gobject primitives)
              :select (gruntime:byte-order)))

;; try setting a scheduler here to see if anything is different
;; if nothing is set it defaults to opt
;; (gst-scheduler-factory-set-default-name "basicwingo")

(define-with-docs ss-plugin
  "The GStreamer plugin that defines soundscrape elements."
  (or
   (gst-plugin-load-file
    (build-path "." "soundscrape" ".libs" "libsoundscrape.so") #f)
   (errorf "~A\n~A\n~A"
           "Unable to load soundscrape plugin.\n"
           "If you are running uninstalled soundscrape, make sure you are running\n"
           "it from the top source directory.")))
(gst-registry-pool-add-plugin ss-plugin)

(define-with-docs <ss-pad>
  "A class for @dfn{pads} (connection points) within a synthesis network."
  (gtype->class (gtype-from-name "SSPad")))
(define-with-docs <ss-element>
  "A class for unit generators."
  (gtype->class (gtype-from-name "SSElement")))

(define-with-docs ss-main-pipeline
  "The main pipeline that contains the soundscrape synthesis network."
  (make "pipeline"))
(define-with-docs ss-audiosink
  "The GStreamer element that will run the audio hardware."
  (make "osssink"))
(define-with-docs ss-filesink
  "The GStreamer element that will record to disk."
  (make "sfsink"))
(define-with-docs ss-float2int
  "The GStreamer element that converts between float and int."
  (make "float2int"))
(define-with-docs ss-main-loop
  "The GLib main loop that the synthesis network runs."
  (g-main-loop-new #f #f))

(define-with-docs ss-sample-rate
  "The sample rate of the soundscrape system. Changes to this
variable will take place when running the next synthesis network."
  44100)
(define-with-docs ss-buffer-frames
  "The number of frames per synthesis buffer. Changes to this
variable will take place when running the next synthesis network."
  256)

;(gst-debug-set-threshold-for-name "GST_CAPS" 6)
;(gst-debug-set-threshold-for-name "GST_PROPERTIES" 6)

;; defined as a function to allow the user to change the properties
(define (ss-fixed-caps)
  (gst-caps-new
   "ss-fixed-caps"
   "audio/x-raw-float"
   (gst-props-new `("endianness" int-type ,gruntime:byte-order)
                  '("width" int-type 32)
                  `("rate" int-type ,ss-sample-rate)
                  '("channels" int-type 1)
                  `("buffer-frames" int-type ,ss-buffer-frames))))

;; catch all errors from the pipeline with this handler
(connect ss-main-pipeline 'error
         (lambda (pipe e str) (errorf "ERROR:~A: ~A" (get-name e) str)))

;; quit the main loop when we catch a change from 'playing to 'paused
(connect ss-main-pipeline 'state-change
         (lambda (pipe old new)
           (if (and (eq? old 8)
                    (eq? new 4)) ;; these numbers are hacks...
               (g-main-loop-quit ss-main-loop))))

(define (graph-elements graph result)
  (let* ((parents (delete-duplicates 
                   (filter (lambda (e) (not (member e result))) (map get-parent graph))))
         (sink-pads (apply
                     append
                     (map (lambda (pad-list)
                            (filter (lambda (pad)
                                      (eq? (genum->symbol (get-direction pad)) 'sink))
                                    pad-list))
                          (map get-pad-list parents))))
         (peers (filter-map get-peer sink-pads)))
    (if (eq? parents '())
        result
        (graph-elements peers (append parents result)))))

(define-public (graph-fold proc init graph)
  "Call a procedure @var{proc} for each element in @var{graph},
accumulating arguments starting with @var{init}. Similar to
@code{hash-fold}."
  (let ((result init)
        (elements (graph-elements graph '())))
    (for-each
     (lambda (e) (set! result (apply proc (list e result))))
     elements)
    result))

(define (run graph setup-graph cleanup-graph)
  "Run the synthesis network defined by @var{graph}. Before running,
call @var{setup-graph} with @var{graph} as its argument. After running,
call @var{cleanup-graph}, again with @var{graph} as the argument."
  (if (not (or (is-a? graph <gst-pad>)
               (and (list? graph)
                    (not (eq? graph '()))
                    (and-map (lambda (x) (is-a? x <gst-pad>)) graph))))
      (error "Bad synthesis network. Try using the `audio' or `control' functions."))
  (let* ((graph (if (list? graph) graph (list graph)))
         (pads graph)
         (elements (graph-fold cons '() graph)))
    ;; my first use of dynamic-wind :-)
    (dynamic-wind
        (lambda ()
          (for-each
           (lambda (element)
             (if (not (add ss-main-pipeline element))
                 (errorf "Element ~A failed to add to the pipeline, aborting."
                         (get-name element))))
           elements)
          ;(set-state ss-main-pipeline 'paused) ;; bring everything to PAUSED so we can link
          (setup-graph pads)
          (attach-iterator ss-main-pipeline)
          (set-state ss-main-pipeline 'playing))
        (lambda () (g-main-loop-run ss-main-loop))
        (lambda () 
          (set-state ss-main-pipeline 'paused)
          (remove-iterator ss-main-pipeline)
          (for-each
           (lambda (element)
             (for-each
              (lambda (pad)
                (and-let* ((peer (get-peer pad))) (unlink pad peer)))
              (get-pad-list element))
             (remove ss-main-pipeline element)
             (set-state element 'null))
           (get-list ss-main-pipeline))
          (if cleanup-graph (cleanup-graph))))))

(define (play . args)
  "Construct a graph of unit generators and play it through
the sound card"
  (let-positional-keywords args
                           ((graph #f)
                            (duration #f))
    (let ((clock-id (if duration
                        (new-single-shot-id (get-clock ss-audiosink)
                                            (inexact->exact (* duration 1e9)))
                        #f)))
      (run graph
           (lambda (pads) ; setup-graph-proc
             (add ss-main-pipeline ss-float2int)
             (add ss-main-pipeline ss-audiosink)
             (if duration
                 (gst-clock-id-wait-async
                  clock-id
                  (lambda (time) (g-main-loop-quit ss-main-loop) #t)))
             (for-each
              (lambda (pad)
                (or (link-filtered pad (get-pad ss-float2int "sink%d") (ss-fixed-caps))
                    (errorf "Failed to link ~A to float2int, aborting..." pad)))
              pads)
             (or (link (get-pad ss-float2int "src") (get-pad ss-audiosink "sink"))
                 (errorf "Failed to link float2int to the audiosink, aborting...")))

           (lambda () ; cleanup-graph
             (if duration (gst-clock-id-unschedule clock-id))
             (for-each
              (lambda (pad)
                (remove-pad ss-float2int pad))
              (filter-direction (get-pad-list ss-float2int) 'sink)))))))

;; the function "write" is already used
(define (render . args)
  "Construct a graph of unit generators and write
the output to disk"
  (let-positional-keywords args
                           ((graph #f)
                            (duration #f)
                            (location #f)
                            (type 'wav)
                            (format 'signed-16-bit-pcm))
    (let ((clock-id (if duration
                        (new-single-shot-id (get-clock ss-filesink)
                                            (inexact->exact (* duration 1e9)))
                        #f)))
      (run graph
           (lambda (pads) ; setup-graph
             (add ss-main-pipeline ss-filesink)
             (set ss-filesink 'location location)
             (set ss-filesink 'major-type type)
             (set ss-filesink 'minor-type format)
             (if duration
                 (gst-clock-id-wait-async
                  clock-id
                  (lambda (time) (g-main-loop-quit ss-main-loop) #t)))
             (for-each
              (lambda (pad)
                (or (link-filtered pad (get-pad ss-filesink "sink%d") (ss-fixed-caps))
                    (errorf "Failed to link ~A to the filesink, aborting..." pad)))
              pads))
           
           (lambda () ; cleanup-graph
             (if duration (gst-clock-id-unschedule clock-id))
             (for-each
              (lambda (pad)
                (release-request-pad ss-filesink pad))
              (get-pad-list ss-filesink)))))))

(define (record . args)
  "Construct a graph of unit generators, play it, and write
the output to disk in realtime"
  (let-positional-keywords args
                           ((graph #f)
                            (duration #f)
                            (location #f)
                            (type 'wav)
                            (format 'signed-16-bit-pcm))
    (let ((clock-id (if duration
                        (new-single-shot-id (get-clock ss-filesink)
                                            (inexact->exact (* duration 1e9)))
                        #f)))
      (run graph
           (lambda (pads) ; setup-graph
             (add ss-main-pipeline ss-filesink)
             (add ss-main-pipeline ss-float2int)
             (add ss-main-pipeline ss-audiosink)
             (set ss-filesink 'location location)
             (set ss-filesink 'major-type type)
             (set ss-filesink 'minor-type format)
             (if duration
                 (gst-clock-id-wait-async
                  clock-id
                  (lambda (time) (g-main-loop-quit ss-main-loop) #t)))
             (for-each
              (lambda (pair)
                (link (car pair) (get-pad ss-filesink "sink%d"))
                (link (cdr pair) (get-pad ss-float2int "sink%d")))
              (map
               (lambda (pad)
                 (let ((tee (make "tee")))
                   (add ss-main-pipeline tee)
                   (or (link-filtered pad (get-pad tee "sink") (ss-fixed-caps))
                       (errorf "Failed to link ~A to the tee, aborting..." pad))
                   (cons (get-pad tee "src%d") (get-pad tee "src%d"))))
               pads))
             (or (link ss-float2int ss-audiosink)
                 (errorf "Failed to link float2int to the audiosink, aborting..." pad)))
           
           (lambda () ; cleanup-graph
             (if duration (gst-clock-id-unschedule clock-id))
             (for-each
              (lambda (pad)
                (release-request-pad ss-filesink pad))
              (get-pad-list ss-filesink))
             (for-each
              (lambda (pad)
                (remove-pad ss-float2int pad))
              (filter-direction (get-pad-list ss-float2int) 'sink)))))))
