;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Support routines for constructing unit generators.
;;
;;The bindings exported by this module are intended for internal use.
;;
;;; Code:

(define-module (soundscrape element-support)
  :use-module (soundscrape util)
  :use-module (soundscrape gst-util)
  :use-module (soundscrape core)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-13)
  :use-module (gnome gstreamer)
  :export (ss-multichannel-expand
           ss-make
           ss-link-pads
           make-documentation
           choose-factory
           make-operator))

(use-modules ((srfi srfi-1)
             :select (filter filter-map zip delete-duplicates)))

;; Perform multichannel expansion if any argument is a list. Quoted
;; lists are protected from expansion.
(define (ss-multichannel-expand proc args)
  "If any element of @var{args} is a list, return a list of multiple
invocations of @var{proc}, as many as the longest list. Scalar elements
are simply copied for each invocation. References into lists with fewer
elements than the longest list are wrapped around.
@example
(ss-multichannel-expand + '((1 2) (4 8)))
 ==> (5 10)
(ss-multichannel-expand + '((1 2) (4 8) 16))
 ==> (21 26)
(ss-multichannel-expand + '((1 2) (4 8) 16 (32 64 128)))
 ==> (53 90 149)
@end example
If no member of @var{args} is a list, return @code{#f}."
  (let* ((quoted
          (let loop ((args args))
            (cond ((eq? args '())
                   '())
                  ((and (keyword? (car args))
                        (list? (cadr args))
                        (eq? (caadr args) 'quote))
                   (cons* (car args) (cadr args) (loop (cddr args))))
                  ((and (list? (car args))
                        (eq? (caar args) 'quote))
                   (cons (car args) (loop (cdr args))))
                  (else
                   (loop (cdr args))))))
         (to-expand (filter (lambda (x) (not (memq x quoted))) args)))
    (if (or-map list? to-expand)
        (let loop ((n 0)
                   (nmax (apply max (map any-length to-expand))))
          (if (eq? n nmax)
              '()
              (cons
               (apply proc
                      (append (map (lambda (x) (wrapping-ref x n)) to-expand)
                              quoted)) ;; FIXME: for the moment, pass
                                       ;; quoted lists with kwargs
               (loop (1+ n) nmax))))
        #f)))

(define (ss-make factory-chooser rate . args)
  "Create a unit generator chosen by calling @var{factory-chooser} on
@var{args}, performing multichannel expansion as appropriate."
  (or
   (ss-multichannel-expand
    (lambda args (apply ss-make factory-chooser rate args))
    args)
   (let ((element (create (factory-chooser args) #f)))
     (set element 'ss-rate rate)
     (ss-link-pads element args))))

(define (ss-link-pads element args)
  "For each input of @var{element}, sets the pad's value or links to the
output of another modulating element as appropriate. Returns either one
output or a list of multiple outputs."
  (let* ((pad-list (get-pad-list element))
         (src-list (filter-direction pad-list 'src))
         (sink-list (filter-direction pad-list 'sink))
         (process-sink (lambda (pad arg)
                         (cond ((is-a? arg <gst-pad>)
                                (link arg pad))
                               ((real? arg)
                                (set pad 'value arg))
                               (else
                                (throw 'bad-arg)))))
         (positional-args (until keyword? args))
         (keyword-args (list-tail args (length positional-args))))

    (if (> (length positional-args) (length sink-list))
        (throw 'too-many-positional-args))
    (for-each
     (lambda (pair)
       (if (cadr pair)
           (process-sink (car pair) (cadr pair))))
     (zip sink-list positional-args))

    (let loop ((kwargs keyword-args))
      (cond
       ((eq? kwargs '()) #f) ; finished
       ((and (keyword? (car kwargs)) (not (keyword? (cadr kwargs))))
        (if (eq? (car kwargs) :properties)
            (for-each
             (lambda (kvp) (set element (car kvp) (cdr kvp)))
             (cadadr kwargs))
            (let ((pad (get-pad element (keyword->string (car kwargs)))))
              (if (not pad)
                  (throw 'bad-keyword))
              (process-sink pad (cadr kwargs))
              (loop (cddr kwargs)))))
       (else
        (throw 'bad-args kwargs))))

    (if (eq? (length src-list) 1)
        (car src-list)
        src-list)))

(define (make-documentation factory)
  "Make documentation for an element with GStreamer factory or factories
@var{factory}. The resulting documentation will be formatted as
@ref{(soundscrape ui stext)}."
  (let* ((name (if (list? factory)
                   (cons 'text
                         (list-join
                          (map (lambda (f) (list 'var (get-name f))) factory)
                          ", "))
                   (list 'var (get-name factory))))
         (factory (if (list? factory) (car factory) factory)) ; 'car factory', hehe
         (pad-templates (get-pad-templates factory)))
    `(text
      ,(texinfo-escape (get-description factory))
      "\n" "\n" "A Soundscrape unit generator."

      (subsection "Inputs")
      ,(cons
        '(table (code))
        (let ((items (map
                      (lambda (template)
                        (list
                         `(item ,(texinfo-escape (get-name-template template)))
                         "Default: "
                         (number->string (let ((v (get template 'default-value)))
                                           (if (< (abs v) 1e-40) 0 v)))))
                      (filter-direction pad-templates 'sink))))
          (if (null? items)
              '((item "[none]"))
              (apply append items))))

      (subsection "Outputs")
      ,(cons
        '(table (code))
        (let ((items (map
                      (lambda (template)
                        `((item ,(texinfo-escape (get-name-template template)))))
                      (filter-direction pad-templates 'src))))
          (if (null? items)
              '((item "[none]"))
              (apply append items))))

      (subsection "Properties")
      ,(cons
        '(table (code))
        (let ((items (map
                      (lambda (pspec)
                        (list
                         `(item ,(symbol->string (gparam-spec:name pspec)))
                         (string-append (gparam-spec:blurb pspec) "\n")
                         (string-append "value-type: "
                                        (gtype-name (gparam-spec:value-type pspec))
                                        "\n")
                         (string-join
                          (map
                           (lambda (pair)
                             (string-append
                              (symbol->string (keyword->symbol (car pair))) ": "
                              (object->string (cadr pair))))
                           (gparam-spec:args pspec))
                          "\n")))
                      (map
                       gparam->pspec-struct
                       (vector->list
                        (gobject-class-get-properties
                         (gtype->class
                          (get-element-type factory))))))))
          (if (null? items)
              '((item "[none]"))
              (apply append items))))

      (subsection "Details")
      ((table (code))
       (item "Author")
       ,(texinfo-escape (get-author factory))
       (item "GStreamer element factory")
       ,name))))

;; nasty, but useful
(define (choose-factory choices rate-char args)
  "Given a choice between different factories, choose the most
appropriate one based on the rates of the input @var{args} and the
desired output rate, expressed as @var{rate-char} (either @code{#\\a} or
@code{#\\c})."
  (call-with-current-continuation
   (lambda (exit)
     (if (not (list? choices))
         (exit choices))
     (let* ((templates (get-pad-templates (car choices)))
            (sink-list (filter-direction templates 'sink))
            (src-list (filter-direction templates 'src))
            (sink-spec "")
            (src-spec "")
            (positional-args (until keyword? args))
            (keyword-args (list-tail args (length positional-args))))

       (define (pop-positional)
         (if (null? positional-args)
             #f
             (let ((ret (car positional-args)))
               (set! positional-args (cdr positional-args))
               ret)))

       ;; Make two lists, one for the src pads and one for the sink
       ;; pads. The lists will contain either 'a' or 'c' for each pad
       ;; depending on whether it is audio-rate or control-rate. These
       ;; lists will then be compared with the factory name to choose
       ;; the best match.

       (set! src-spec
             (map
              (lambda (src) rate-char)
              (map get-name-template src-list)))

       (set! sink-spec
             (map
              (lambda (name)
                (let ((value (or (get-keyword (string->keyword name) keyword-args #f)
                                 (pop-positional))))
                  (cond
                   ((and (is-a? value <ss-pad>)
                         (eq? (genum->symbol (get (get-parent value) 'ss-rate))
                              'control))
                    #\c)
                   ((is-a? value <gst-pad>)
                    #\a)
                   (else
                    #\c))))
              (map get-name-template sink-list)))
       
       ;(display (list sink-spec src-spec)) (newline)
       
       ;; Find an appropriate factory, or just choose the first one.
       (or (or-map
            (lambda (factory)
              (define (remove-names spec)
                (let loop ((rest (string->list spec)))
                  (if (null? rest)
                      '()
                      (cons (cadr rest) (loop (cddr rest))))))

              (let* ((split (string-split (get-name factory) #\-))
                     (sinks (remove-names (cadr split)))
                     (srcs (remove-names (if (> (length split) 2) (caddr split) ""))))
                (if (and (equal? sinks (list-head sink-spec (length sinks)))
                         (equal? srcs (list-head src-spec (length srcs))))
                    factory
                    #f)))
            choices)
           (car choices))))))

(define (make-operator factory symbol num-args)
  "Make a new operator with named @var{symbol}. This operator will
override the definition in the root module, and will invoke a
soundscrape unit generator if any argument to it is not a number."
  (let* ((old-var (or (module-variable the-root-module symbol)
                      (module-variable
                       (nested-ref the-root-module '(app modules soundscrape operators))
                       symbol)))
         (old-val (if old-var (variable-ref old-var) #f)))
    (define (operator . args)
      (or
       (ss-multichannel-expand operator args)
       (cond
        ((and old-val (and-map number? args))
         (apply old-val args))
        ((not num-args)
         (let ((element (create factory #f)))
           (for-each
            (lambda args (get-pad element "sink%d"))
            args)
           (set element 'ss-rate 'audio) ;; fixme
           (ss-link-pads element args)))
        ((and num-args (eq? (length args) num-args))
         (let ((element (create factory #f)))
           (set element 'ss-rate 'audio) ;; fixme
           (ss-link-pads element args)))
        (else
         (errorf "Expected ~A arguments to operator ~A, but only got ~A."
                 num-args symbol (length args))))))
    ;; craziness!
    (module-define! the-root-module symbol operator)
    operator))

