;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;This module defines all of the available soundscrape elements.
;;
;;In soundscrape, unit generators are constructed with the @code{audio}
;;and @code{control} generic methods. (This method of calling is in
;;contrast to traditional object-oriented systems such as SuperCollider;
;;generic functions are Scheme's equivalent to object methods.)
;;
;;The first argument to these methods should be a unit generator class.
;;A list of all available classes can be found in the documentation for
;;@code{ss-all-ugens}.
;;
;;Following that are arguments to connect the input ports of the
;;generator to values, other generators, or lists of either, in which
;;case multichannel expansion is performed. You may pass arguments in
;;the order given in that generator's help, or with keywords.
;;
;;Examples:
;;@example
;;@;; track the frequency of a signal, in this case a sine wave
;;(audio freqTracker
;;       :tracking-speed 0.1
;;       :input (sine 'audio :frequency 440))
;;
;;@;; the same thing with positional arguments -- see (help freqTracker)
;;(audio freqTracker 0.1 (sine 'audio 440))
;;
;;@;; you can set object properties with the special :properties
;;@;; keyword; it accepts a quoted assoc-list of properties to set
;;(control plug :properties ''((lag-time . 0.1) (value . 300)))
;;@end example
;;
;;Try @code{(help @var{name})} for help on a specific unit generator.
;;
;;; Code:

(define-module (soundscrape elements)
  :use-module (gnome gstreamer)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-13)
  :use-module (soundscrape util)
  :use-module (soundscrape core)
  :use-module (soundscrape element-support)
  :export (audio control ss-all-ugens mix))

;; Documentation for this one is set after all plugins are loaded.
(define ss-all-ugens (make-hash-table 31))

;; We're actually going to put the created definitions in a new module.
;; This is both for cleanliness, as well as making the graphical help
;; less cluttered. We need to call beautify-user-module! to finish the
;; initialization process.
(define ss-ugens-module (resolve-module '(soundscrape ugens)))
(beautify-user-module! ss-ugens-module)

(define ss-element-factory (make-object-property))

;; Should these be changed to ar and cr to be more concise?
(define-generic-with-docs audio
  "A generic function used to create unit generators operating
at audio rate.")

(define-generic-with-docs control
  "A generic function used to create unit generators operating
at control rate.")

;; The general case.
(define-method (audio (ugen <gtype-class>) . args) ;; GAH! Need to make a metaclass.
  (apply ss-make (lambda (args) (ss-element-factory ugen)) 'audio args))

(define-method (control (ugen <gtype-class>) . args)
  (apply ss-make (lambda (args) (ss-element-factory ugen)) 'control args))

;; The case where you have to choose a factory depending on the args.
;; This is a detail that is normally hidden from the user.
(define-method (audio (ugen-list <list>) . args)
  (apply ss-make 
         (lambda (args) (choose-factory (ss-element-factory ugen-list) #\a args))
         'audio args))

(define-method (control (ugen-list <list>) . args)
  (apply ss-make 
         (lambda (args) (choose-factory (ss-element-factory ugen-list) #\c args))
         'control args))

;; A specific implementation for the disk-in ugen.
(let* ((factory (gst-element-factory-find "sfsrc")))
  (ensure-loaded factory)
  (let ((disk-in (gtype->class (get-element-type factory))))
    (module-define! ss-ugens-module 'disk-in disk-in)
    (hashv-create-handle! ss-all-ugens 'disk-in factory)
    (set! (ss-element-factory disk-in) factory)
    ;; what the fuck -- does goops not have class methods?
    (define-method (audio (in (class-of disk-in)) location . args)
      (let ((in (create (ss-element-factory in) #f)))
        (let-positional-keywords args ((loop? #f)
                                       (start-frame 0))
          (set in 'location location)
          (set in 'loop loop?)
          (let ((ret (get-pad-list in)))
            (if (null? ret) (error "Could not open file for input:" location) ret)))))))

;; We can implement mix as a function, given the fact that + accepts
;; multiple arguments. Nice, eh?
(define (mix channels)
  (define (wrapping-zip . args)
    (let ((nmax (apply max (map any-length args))))
      (let lp ((n 0))
        (if (eq? n nmax)
            '()
            (cons (map (lambda (l) (wrapping-ref l n)) args) (lp (1+ n)))))))
  (map (lambda (l) (apply + l)) (apply wrapping-zip channels)))

;; Now we begin with the plugins offered by libsoundscrape.so.

;; can this list be automagically generated?
(define ss-ladspa-choice-list (make-hash-table 31))
(for-each
 (lambda (pair)
   (hash-create-handle! ss-ladspa-choice-list (car pair) (cdr pair)))
 '(("branch" "branch-ia-oaoa" "branch-ic-ococ")
   ("fmod" "fmod-fama-oa" "fmod-famc-oa" "fmod-fcma-oa" "fmod-fcmc-oc")
   ("impulse" "impulse-fc")
   ("lp4pole" "lp4pole-faraia-oa" "lp4pole-fcrcia-oa")
   ("pulse" "pulse-fapa-oa" "pulse-fcpa-oa" "pulse-fapc-oa" "pulse-fcpc-oa")
   ("random-osc" "random-fasa-oa" "random-fasc-oa" "random-fcsa-oa" "random-fcsc-oa")
   ("saw" "sawtooth-fa-oa" "sawtooth-fc-oa")
   ("sine" "sine-faaa" "sine-fcac" "sine-faac" "sine-fcaa")
   ("square" "square-fa-oa" "square-fc-oa")
   ("syncsquare" "syncsquare-fcga-oa" "syncsquare-faga-oa")
   ("triangle" "triangle-fasa-oa" "triangle-fcsa-oa" "triangle-fasc-oa" "triangle-fcsc-oa")))

(define ss-factory-name-transformations (make-hash-table 31))
(for-each
 (lambda (pair)
   (hash-create-handle! ss-factory-name-transformations (car pair) (cdr pair)))
 '(("ss-plug" . "plug")
   ("ss-float2int" . #f)
   ("am" . #f)
   ("amp" . #f)
   ("amp-gaia-oa" . #f)
   ("amp-gcia-oa" . #f)
   ("amp-mono" . #f)
   ("amp-stereo" . #f)
   ("dcRemove" . #f)
   ("identity-audio" . #f)
   ("identity-control" . #f)
   ("interpolator" . #f)
   ("mixer" . #f)
   ("noise-source-white" . "noise-white")
   ("null-ai" . #f)
   ("null-ao" . #f)
   ("null-ci" . #f)
   ("null-co" . #f)
   ("product-iaia-oa" . #f)
   ("product-iaic-oa" . #f)
   ("product-icic-oc" . #f)
   ("split" . #f)
   ("sum-iaia-oa" . #f)
   ("sum-iaic-oa" . #f)
   ("sum-icic-oc" . #f)))

(define <ss-multi-op> (gtype->class (gtype-from-name "SSMultiOp")))
(define <ss-binary-op> (gtype->class (gtype-from-name "SSBinaryOp")))
(define <ss-unary-op> (gtype->class (gtype-from-name "SSUnaryOp")))

(for-each
 (lambda (name)
   (call-with-current-continuation
    (lambda (next)
      (let ((factory (gst-element-factory-find name))
            (symbol #f)
            (ugen #f))
        (if (not factory)
            (gruntime-error "Could not find an element factory named ~A" name)
            (gst-plugin-feature-ensure-loaded factory))
        
        (hash-fold (lambda (key val x) (if (member name val) (next #f)))
                   #f ss-ladspa-choice-list)

        (set! symbol (string->symbol
                      (or (hash-ref ss-factory-name-transformations name name)
                          (next #f))))
        
        (cond
         ((class-is-a? (gtype->class (gst-element-factory-get-element-type factory))
                       <ss-multi-op>)
          (set! ugen (make-operator factory symbol #f))
          (module-define! ss-ugens-module symbol ugen))
         ((class-is-a? (gtype->class (gst-element-factory-get-element-type factory))
                       <ss-binary-op>)
          (set! ugen (make-operator factory symbol 2))
          (module-define! ss-ugens-module symbol ugen))
         ((class-is-a? (gtype->class (gst-element-factory-get-element-type factory))
                       <ss-unary-op>)
          (set! ugen (make-operator factory symbol 1))
          (module-define! ss-ugens-module symbol ugen))
         ((module-locally-bound? ss-ugens-module symbol)
          (next #f))
         (else
          (set! ugen (gtype->class (get-element-type factory)))
          (module-define! ss-ugens-module symbol ugen)))

        (set-object-property! ugen 'documentation (make-documentation factory))
        (set! (ss-element-factory ugen) factory)
        
        (hashv-create-handle! ss-all-ugens symbol factory)))))
 (map get-name (gst-plugin-get-feature-list ss-plugin)))

(hash-fold
 (lambda (composite-name instance-names x)
   (let ((factories (map gst-element-factory-find instance-names))
         (symbol (string->symbol composite-name))
         (ugen #f))
     (if (and-map identity factories)
         (begin
           (set! ugen (map (lambda (f) (gtype->class (get-element-type f))) factories))
           (set-object-property! ugen 'documentation (make-documentation factories))
           (set! (ss-element-factory ugen) factories)

           (hashv-create-handle! ss-all-ugens symbol factories)
           
           (module-define! ss-ugens-module symbol ugen)))))
 #f ss-ladspa-choice-list)

;; Now we define specific implementations of audio and control for
;; libsoundscrape plugins.
(define-method (audio (ugen <gtype-class>) (closure <procedure>) . args)
  ;; lack of class methods is really irritating.
  (cond
   ((eq? ugen sequencer)
    (let ((seq (create (ss-element-factory ugen) #f)))
      (set seq 'lambda closure)
      (set seq 'ss-rate 'audio)
      (ss-link-pads seq args)))
   (else
    (error "No applicable method." ugen))))

(define-method (control (ugen <gtype-class>) (closure <procedure>) . args)
  ;; lack of class methods is really irritating.
  (cond
   ((eq? ugen sequencer)
    (let ((seq (create (ss-element-factory ugen) #f)))
      (set seq 'lambda closure)
      (set seq 'ss-rate 'control)
      (ss-link-pads seq args)))
   (else
    (error "No applicable method." ugen))))

(module-export!
 ss-ugens-module
 (hash-fold (lambda (key value fold) (cons key fold)) '() ss-all-ugens))

(set-object-property!
 ss-all-ugens 'documentation
 `(text
   "A hash table of all available unit generators. The "
   "key is the generator's name (a symbol), and the value is the "
   "corresponding GStreamer element factory."
   ((itemize (bullet))
    ,@(apply
       append
       (sort! 
        (hash-fold
         (lambda (key value fold)
           (cons
            `((item (code ,(symbol->string key)))
              ,(if (list? value)
                   (string-join
                    (map get-description value)
                    ", ")
                   (get-description value)))
            fold))
         '() ss-all-ugens)
        (lambda args (apply string-ci<? (map (lambda (x) (car (cdadar x))) args))))))
   "To get help on a specific unit generator, try " (code "(help " (var "name") ")") "."))
