;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Envelopes are used to give sounds dynamic character and to cause
;;them to end. They are defined as discretely sampled functions. Each
;;transition between levels occurs within a specified time, and can be
;;interpolated either linearly, exponentially, or as sine segments.
;;
;;Common kinds of envelope can be made with the @code{linen},
;;@code{triangle}, @code{sine}, and @code{perc} convenience functions,
;;all of which have the @code{envelope:} prefix.
;;
;;Some kinds of envelope, like @code{adsr} and @code{asr}, have a
;;sustain portion. You can indicate the index of this level when
;;creating an envelope.
;;
;;To use an envelope in a synthesis network, use the @code{enveloptor}
;;unit generator. See its help for more information.
;;
;;; Code:

(define-module (soundscrape envelopes)
  :use-module (oop goops)
  :export (<envelope>
           envelope:linen
           envelope:triangle
           envelope:sine
           envelope:adsr
           envelope:ar))

;; how to represent envelopes? hm. will these be the first native goops
;; objects I create?

(define-class <envelope> ()
  (levels :init-value '(0))
  (times :init-value '())
  (curvature :init-value 'linear)
  (sustain :init-value #f))
(set-object-property! <envelope> 'documentation
  "A class representing an envelope. Has class variables @code{:levels},
@code{:times}, @code{:curvature}, and @code{:sustain}. The envelope is
described by @var{levels} and transition times given in @var{times}. As
the first element of @var{levels} is the starting level, @var{levels} is
one element longer than @var{times}. @var{curvature} can be either
@code{linear}, @code{exponential}, or @code{sine}. If @var{sustain} is
not @code{#f}, it is interpreted as the index into @var{levels} of the
sustain phase.")

(define (envelope:linen attack sustain release level curve)
  (make <envelope>
    :levels (list 0 level level 0)
    :times (list attack sustain release)
    :curvature curve))

(define (envelope:triangle duration level)
  (make <envelope>
    :levels (list 0 level 0)
    :times (list (/ duration 2) (/ duration 2))))

(define (envelope:sine duration level)
  (make <envelope>
    :levels (list 0 level 0)
    :times (list (/ duration 2) (/ duration 2))
    :curvature 'sine))

(define (envelope:adsr attack decay sustain release level curve)
  (make <envelope>
    :levels (list 0 level (* level sustain) 0)
    :times (list attack decay release)
    :sustain 2
    :curvature curve))

(define (envelope:ar attack release level curve)
  (make <envelope>
    :levels (list 0 level 0)
    :times (list attack release)
    :sustain 1
    :curvature curve))
