;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;An "environment" in the soundscrape sense is a lexical environment
;;in the scheme sense. An empty environment only knows the symbols
;;exported by the util, orchestra, operator, envelope, and misc
;;soundscrape modules.
;;
;;In soundscrape, they can also be used as dynamic musical parameters.
;;When they are used this way, they are called `events'. Some examples
;;should be given here. See
;;@uref{http://www.audiosynth.com/schtmldocs/Documentation/Streams-Patterns-Events/}
;;for more information.
;;
;;; Code:

(define-module (soundscrape environments)
  :use-module (soundscrape util)
  :use-module (soundscrape orchestras)
  :use-module (soundscrape operators)
  :use-module (soundscrape envelopes)
  :use-module (soundscrape misc)
  :export (make-empty-environment
           make-environment
           with-environment
           display-environment
           call-with-environment
           estream:empty
           estream:proto-event
           estream:merge!
           estream:binder))

;; A lexical environment is a list whose last element is an
;; eval-closure. We capture that closure so that we can use other
;; functions defined by soundscrape.
(define (make-empty-environment)
  "Make an empty environment. The environment will, however, contain
bindings exported by soundscrape."
  (last-pair (the-environment)))

;; Just like let* does, make-environment conses symbol-value pairs onto
;; the lexical environment. You can get or set values from this
;; environment using local-eval.
(defmacro make-environment args
  (let loop ((l args) (ret (make-empty-environment)))
    (if (null? l)
        (list 'quote ret)
        (loop (cddr l) (cons (cons (car l) (local-eval (cadr l) ret)) ret)))))
(set-object-property! make-environment 'documentation
  "Make an environment and populate it. @var{args} is a list of symbols
and values. This is a syntax, so @var{args} are not evaluated before
calling this syntax; rather, values are evaluated in the lexical
environment that is being created. This allows binding values to
reference other bindings.")

(defmacro with-environment (env . body)
  `(local-eval
    ,(list 'quote `(begin ,@body))
    ,env))
(set-object-property! with-environment 'documentation
  "Execute @var{body} within the lexical environment @var{env}.")

(define (display-environment env)
  "Display a list of SYMBOL => VALUE pairs."
  (let loop ((l env))
    (if (pair? (car l))
        (begin
          (format #t "~A => ~A\n" (caar l) (cdar l))
          (loop (cdr l))))))

(define (call-with-environment env proc . args)
  "Call @var{proc} with arguments @var{args}. If any argument to
@var{proc} is not specified in @var{args}, get it from environment
@var{env}."
  (let* ((formals (cadr (procedure-source proc)))
         (positional (until keyword? args))
         (keyword (list-cdr-ref args (length positional)))
         (unset-sym (gensym))
         (unset? (lambda (x) (eq? x unset-sym))))
    (if (not (list? formals))
        (error "Procedure must accept a fixed number of args"))
    (if (> (length positional) (length formals))
        (error "Too many positional args"))
    (if (not (even? (length keyword)))
        (error "Bad keyword arguments"))
    (let ((newargs (append positional
                           (make-list (- (length formals) (length positional)) unset-sym))))
      (let loop ((keyword keyword))
        (if (not (null? keyword))
            (let ((i (list-index formals (keyword->symbol (car keyword)))))
              (if (not i)
                  (error "Invalid keyword" (car keyword)))
              (list-set! newargs i (cadr keyword))
              (loop (cddr keyword)))))
      (do ((i 0 (+ i 1))) ((eq? i (length newargs)) #f)
        (if (unset? (list-ref newargs i))
            ;; the list-set! will throw an error if the symbol is not bound...
            (list-set! newargs i (local-eval (list-ref formals i) env))))
      (apply proc newargs))))

(define (estream:empty . args)
  "A stream that returns new, empty events."
  (make-empty-environment))

(define (estream:proto-event . args)
  "A stream that returns new events. Document me!"
  (make-environment
    value (lambda (x) (if procedure? x) (x) x)

    reverse #f
    tempo #f
    dur 1.0
    stretch 1.0
    legato 0.8
    sustain (lambda () (* dir legato stretch))
    lag 0.0
    strum 0.0
   
    amp (lambda () (db->amp db))
    db -20.0
    velocity 64 ; MIDI units 0-127
    pan 0.0
    channels 2
   
    mtranspose 0
    gtranspose 0
    ctranspose 0
   
    octave 5.0
    root 0
    degree 0
    scale '(0 2 4 5 7 9 11)
    steps-per-octave 12
    detune 0
   
    note (lambda () (degree->key (+ degree mtranspose) scale steps-per-octave))
    midi-note (lambda () (/ (* (+ (value note) gtranspose
                                  (* octave steps-per-octave) root)
                               12)
                            steps-per-octave))
    freq (lambda () (midi->cps (+ (value midi-note) ctranspose)))
   
    env (envelope:asr 0.01 1.0 0.5)
   
    chan-offset 0
    instrument 'sine-test
    ;; 'wave-table (apply wavetable:sine 1024 (map (lambda (x) (/ 1 x)) '(1 2 3 4 5 6)))
    orchestra (make-orchestra
                (make-instrument
                 'sine-test
                 (lambda (freq amp pan env) ; wavetable
                   (audio pan2
                          (audio sine freq (control env-gen env amp))
                          pan))))
    graph-func (lambda () (orchestra-ref orchestra instrument))
   
    finish (lambda ()
             (set! freq (+ (value freq) detune))
             (set! amp (value amp))
             (set! sustain (value sustain)))))

;; The second stream takes precedence.
(define (estream:merge! s1 s2)
  "Merge two event streams. @var{s2} takes precedence. This procedure
modifies the incoming events to produce the results."
  (lambda args
    (let ((e1 (s1))
          (e2 (s2)))
      (if (not (and e1 e2))
          #f
          (if (eq? 1 (length e2))
              e1
              (begin
                (set-cdr! (list-cdr-ref e2 (- (length e2) 2)) e1)
                e2))))))

(define (estream:binder estream . args)
  "Returns a new stream that binds symbols to values from value streams
(not event streams)."
  (let* ((alist (let lp ((args args))
                  (if (null? args)
                      '()
                      (cons (cons (car args) (cadr args)) (lp (cddr args))))))
         (value (lambda (x) (if (procedure? x) (x) x)))
         (reset (lambda (x) (if (procedure? x) (x 'reset) #f)))
         (get-values (lambda () (map (lambda (pair) (value (cdr pair))) alist))))
    (yielding-thunk
     (map reset (map cdr alist))
     (let loop ((vals (get-values)) (proto (estream)))
       (if (or (not (and-map identity vals)) (not proto))
           (while #t (yield #f)))
       (yield (append (map cons (map car alist) vals) proto))
       (loop (get-values) (estream))))))
