;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;This module implements some miscellaneous useful procedures.
;;
;;; Code:

(define-module (soundscrape misc)
  :export (rrand coin choose yielding-thunk repeat shuffle!))

(define (rrand x y)
  "Returns a random number within [@var{x},@var{y})."
  (+ x (random (if (< x y) (- y x) (- x y)))))

(define (coin x)
  "The probability of this function returning @code{#t} is $var{x}."
  (> (random:uniform) (if (> (abs x) 1.) (frac (abs x)) (abs x))))

(define (choose l)
  "Chooses an element of @var{l} at random."
  (cond
   ((list? l)
    (list-ref l (random (length l))))
   ((vector? l)
    (vector-ref l (random (vector-length l))))
   (else
    (error "Wrong type" l))))

;; This could be more simple, but we have to protect against returning
;; from continuations other than the one being called now now.
(defmacro yielding-thunk BODY
  `(let ((resume #f)
         (return #f))
     (define (yield x)
       (call-with-current-continuation
        (lambda (cont)
          (set! resume cont)
          (return x))))
     (lambda args
       (if (not (null? args))
           (set! resume #f)
           (call-with-current-continuation
            (lambda (%return)
              (set! return %return)
              (if resume
                  (resume #f)
                  (let %loop ()
                    ,@BODY
                    (%loop)))))))))
(set-object-property!
 yielding-thunk 'documentation
  "Usage: (yielding-thunk BODY)

Similar to lambda, yielding-thunk is a syntax that creates a procedure.
The difference is that yielding-thunk provides a routine called `yield'
that can perform a nonlocal exit from the function. The next time the
procedure is called, execution resumes where yield was called. The
procedure exits only when yield is called; if the end of the procedure
is reached, the procedure is restarted.

If any arguments are passed, the state of the procedure is reset.

Example:
  (define f (yielding-thunk (yield 1) (yield 2) #f))
  (f) => 1
  (f) => 2
  (f) => 1

To reset the thunk, pass 'reset as the first argument.
")

(defmacro repeat (REPEATVAR NUM-REPEATS . BODY)
  (if (not (symbol? REPEATVAR))
      (error "Argument 0 must be a symbol:" REPEATVAR))
  (let ((loop (gensym)))
    `(if (not ,NUM-REPEATS)
         (while #t ,@BODY)
         (begin
           (if (or (not (integer? ,NUM-REPEATS)) (< ,NUM-REPEATS 0))
               (error "Argument 1 must be a nonnegative integer:" ,NUM-REPEATS))
           (let ,loop ((,REPEATVAR ,NUM-REPEATS))
                (if (not (zero? ,REPEATVAR))
                    (begin
                      ,@BODY
                      (,loop (1- ,REPEATVAR)))))))))

;; don't know about efficiency, but it's a first try and my brain hurts
;; right now.
(define (shuffle! l)
  "Shuffles the elements of @var{l}, possibly modifying the structure of
its argument in the process."
  (define (delete-ref! l i)
    (if (eq? i 0)
        (cdr l)
        (begin
          (set-cdr! (list-cdr-ref l (1- i)) (list-cdr-ref l (1+ i)))
          l)))
  (let loop ((in l) (out '()))
    (if (null? in)
        out
        (let* ((n (random (length in)))
               (val (list-ref in n)))
          (loop (delete-ref! in n) (cons val out))))))
