;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;This module defines the scheme versions of operators also implemented
;;with unit generators.
;;
;;; Code:

(define-module (soundscrape operators))

(define rlog2 (/ 1 (log 2)))
(define (%log2 x) (* x rlog2))

;; Many of the operators are already defined for us by guile -- those
;; that aren't are defined here

;; Unary ops first
(define-public (1/ x)
  "Reciprocal"
  (/ 1 x))
(define-public ceil
  ceiling)
(define-public (frac x)
  "Fractional part of @var{x}"
  (- x (floor x)))
(define-public (sign x)
  "-1 if @var{x} is negative, 1 otherwise"
  (if (< x 0) -1 1))
(define-public (squared x)
  "@code{(* @var{x} @var{x})}"
  (* x x))
(define-public (cube x)
  "@code{(* @var{x} @var{x} @var{x})}"
  (* x x x))
(define-public (midi->cps x)
  "MIDI note -> Hz"
  (* 440 (expt 2 (/ (- x 69) 12))))
(define-public (cps->midi x)
  "Hz -> MIDI note"
  (+ (* (%log2 (/ x 440)) 12) 64))
(define-public (midi->ratio x)
  "MIDI note range -> frequency ratio"
  (expt 2 (/ x 12)))
(define-public (ratio->midi x)
  "frequency ratio -> MIDI note range"
  (* 12 (%log2 x)))
(define-public (oct->cps x)
  "Octave -> Hz"
  (* 440 (expt 2 (- x 4.75))))
(define-public (cps->oct x)
  "Hz -> Octave"
  (+ (%log2 (/ x 440)) 4.75))
(define-public (amp->db x)
  "Amplitude -> Decibels"
  (* (log10 x) 20))
(define-public (db->amp x)
  "Decibels -> Amplitude"
  (expt 10 (/ x 20)))
                                           
;; within the closure, the old definition is in place
(set! log (lambda (x)
            "Like normal @code{log}, but when @var{x} is negative,
return @code{(- (log (- @var{x})))}."
            (cond ((< x 0) (- (log (- a)))) ((eq? x 0) 0) (else (log x)))))
                                           
(define-public (log2 x)
  "Similar to @code{log}, but base 2."
  (cond ((< x 0) (- (%log2 (- x)))) ((eq? x 0) 0) (else (%log2 x))))
(define-public (log10 x)
  "Similar to @code{log}, but base 10."
  (cond ((< x 0) (- (log10 (- x)))) ((eq? x 0) 0) (else (log10 x))))
(define-public (distort x)
  "@code{(/ @var{x} (+ 1 (abs @var{x}))))}"
  (/ x (+ 1 (abs x))))
(define-public (softclip x)
  "clip the output to +/- 1, but don't mess with @code{(abs @var{x}) < 0.5}"
  (if (< (abs x) 0.5) x (/ (- (abs x) 0.25) x)))

;; Now binary ops
(define-public mod modulo)
(define-public (trunc a b)
  "Truncate @var{a} to be a multiple of @var{b}."
  (* (floor (/ a b)) b))
(set! round (lambda args
              (case (length args)
                "Round the first argument to a multiple of the second,
or to a whole number if only one argument is given."
                ((1) (round (car args)))
                ((2) (* (round (/ (car args) (cadr args))) (cadr args)))
                (else (error "wrong number of args")))))
(define-public (amclip a b)
  "Two-quadrant multiply, @code{(* @var{a} @var{b})} if @code{(> b 0)}."
  (if (> b 0) (* a b) 0))
(define-public (scaleneg a b)
  "Scale @var{a} by @var{b} if @var{a} is negative. Otherwise just
return @var{a}."
  (if (< a 0) (* a b) a))
(define-public (clip2 a b)
  "Clip @var{a} to +/-@var{b}."
  (if (a < (- (abs b))) (- (abs b)) (if (a > (abs b)) (abs b) a)))
(define-public (excess a b)
  "@code{(- @var{a} (clip2 @var{a} @var{b}))}"
  (- a (clip2 a b)))
(define-public (hypot a b)
  "Hypotenuse of triangle with sides @var{a} and @var{b}."
  (sqrt (+ (* a a) (* b b))))
(define-public atan2 atan)
(define-public (difsqr a b)
  "Difference of squares."
  (- (* a a) (* b b)))
(define-public (sqrdif a b)
  "Square of difference."
  (squared (- a b)))
(define-public (sqrsum a b)
  "Square of sum."
  (squared (+ a b)))
(define-public (absdif a b)
  "Absolute value of difference."
  (abs (- b a)))

;; The lone ternary op...
(define-public (degree->key degree scale gamut)
  "Return a note based in @var{gamut} from the @var{deg}th element of
list @var{scale}."
  (+ (* (quotient degree (length scale)) gamut)
     (list-ref scale (modulo degree (length scale)))))
