/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-binary-op.c: Binary operations
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Some code based on gstadder.c from gst-plugins:
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *               2000 Wim Taymans <wim.taymans@chello.be>
 *               2001 Thomas <thomas@apestaart.org>
 *
 * Some code based on BinaryOpUgens.cpp from SuperCollider:
 * Copyright (C) 2002 James McCartney
 */

#include <config.h>
#include "ss-binary-op.h"
#include <math.h>

/* a lot of this could probably be optimized if I used routines from
 * SC_InlineBinaryOp.h */

/* declare the guts of the operations here in the beginning */
static gfloat do_minus_cc (gfloat a, gfloat b)
{
    return a - b;
}
static void do_minus_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] -= b[i];
}
static void do_minus_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] -= b;
}
static void do_minus_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a - b[i];
}
static gfloat do_divide_cc (gfloat a, gfloat b)
{
    return b != 0.f ? a / b : 0.f;
}
static void do_divide_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b[i] != 0.f ? a[i] / b[i] : 0.f;
}
static void do_divide_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    if (b == 0.f)
        for (i=0; i<how_many; i++)
            a[i] = 0.f;
    else
        for (i=0; i<how_many; i++)
            a[i] /= b;
}
static void do_divide_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = b[i] != 0.f ? a / b[i] : 0.f;
}
static gfloat do_mod_cc (gfloat a, gfloat b)
{
    return b != 0.f ? fmod (a, b) : 0.f;
}
static void do_mod_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b[i] != 0.f ? (float)fmod (a[i], b[i]) : 0.f;
}
static void do_mod_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    if (b == 0.f)
        for (i=0; i<how_many; i++)
            a[i] = 0.f;
    else
        for (i=0; i<how_many; i++)
            a[i] = (float)fmod (a[i], b);
}
static void do_mod_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = b[i] != 0.f ? (float)fmod (a, b[i]) : 0.f;
}
static gfloat do_expt_cc (gfloat a, gfloat b)
{
    return a < 0.f ? -((float)pow (-a, b)) : (float)pow (a, b);
}
static void do_expt_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ?
            -((float)pow (-a[i], b[i])) :
            (float)pow (a[i], b[i]);
}
static void do_expt_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ?
            -((float)pow (-a[i], b)) :
            (float)pow (a[i], b);
}
static void do_expt_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    if (a < 0.f)
        for (i=0; i<how_many; i++)
            b[i] = -((float)pow (-a, b[i]));
    else
        for (i=0; i<how_many; i++)
            (float)pow (a, b[i]);
}
static gfloat do_lt_cc (gfloat a, gfloat b)
{
    return a < b ? 1.f : 0.f;
}
static void do_lt_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < b[i] ? 1.f : 0.f;
}
static void do_lt_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < b ? 1.f : 0.f;
}
static void do_lt_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a < b[i] ? 1.f : 0.f;
}
static gfloat do_gt_cc (gfloat a, gfloat b)
{
    return a > b ? 1.f : 0.f;
}
static void do_gt_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] > b[i] ? 1.f : 0.f;
}
static void do_gt_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] > b ? 1.f : 0.f;
}
static void do_gt_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a > b[i] ? 1.f : 0.f;
}
static gfloat do_lte_cc (gfloat a, gfloat b)
{
    return a <= b ? 1.f : 0.f;
}
static void do_lte_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] <= b[i] ? 1.f : 0.f;
}
static void do_lte_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] <= b ? 1.f : 0.f;
}
static void do_lte_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a <= b[i] ? 1.f : 0.f;
}
static gfloat do_gte_cc (gfloat a, gfloat b)
{
    return a >= b ? 1.f : 0.f;
}
static void do_gte_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] >= b[i] ? 1.f : 0.f;
}
static void do_gte_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] >= b ? 1.f : 0.f;
}
static void do_gte_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a >= b[i] ? 1.f : 0.f;
}
static gfloat do_trunc_cc (gfloat a, gfloat b)
{
    /* truncate a to multiples of b */
    return b == 0.f ? a : floor (a / b) * b;
}
static void do_trunc_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b[i] == 0.f ? a[i] : floor (a[i] / b[i]) * b[i];
}
static void do_trunc_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b == 0.f ? a[i] : floor (a[i] / b) * b;
}
static void do_trunc_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = b[i] == 0.f ? a : floor (a / b[i]) * b[i];
}
static gfloat do_round_cc (gfloat a, gfloat b)
{
    return b == 0.f ? a : floor (a / b + 0.5f) * b;
}
static void do_round_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b[i] == 0.f ? a[i] : floor (a[i] / b[i] + 0.5f) * b[i];
}
static void do_round_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b == 0.f ? a[i] : floor (a[i] / b + 0.5f) * b;
}
static void do_round_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = b[i] == 0.f ? a : floor (a / b[i] + 0.5f) * b[i];
}
static gfloat do_amclip_cc (gfloat a, gfloat b)
{
    return b > 0.f ? a * b : 0.f;
}
static void do_amclip_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] *= b[i] > 0.f ? b[i] : 0.f;
}
static void do_amclip_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    b = b > 0.f ? b : 0.f;
    for (i=0; i<how_many; i++)
        a[i] *= b;
}
static void do_amclip_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] *= b[i] > 0.f ? a : 0.f;
}
static gfloat do_scaleneg_cc (gfloat a, gfloat b)
{
    return a < 0.f ? a * b : a;
}
static void do_scaleneg_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] *= a[i] < 0.f ? b[i] : 1.f;
}
static void do_scaleneg_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] *= a[i] < 0.f ? b : 1.f;
}
static void do_scaleneg_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    if (a < 0.f)
        for (i=0; i<how_many; i++)
            b[i] *= a;
    else
        for (i=0; i<how_many; i++)
            b[i] = a;
}
static gfloat do_clip2_cc (gfloat a, gfloat b)
{
    b = b < 0.f ? -b : b;
    return CLAMP (a, -b, b);
}
static void do_clip2_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = b[i] < 0.f
            ? CLAMP (a[i], b[i], -b[i])
            : CLAMP (a[i], -b[i], b[i]);
}
static void do_clip2_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    b = b < 0.f ? -b : b;
    for (i=0; i<how_many; i++)
        a[i] = CLAMP (a[i], -b, b);
}
static void do_clip2_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = b[i] < 0.f
            ? CLAMP (a, b[i], -b[i])
            : CLAMP (a, -b[i], b[i]);
}
static gfloat do_excess_cc (gfloat a, gfloat b)
{
    b = b < 0.f ? -b : b;
    return a - CLAMP (a, -b, b);
}
static void do_excess_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] -= b[i] < 0.f
            ? CLAMP (a[i], b[i], -b[i])
            : CLAMP (a[i], -b[i], b[i]);
}
static void do_excess_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    b = b < 0.f ? -b : b;
    for (i=0; i<how_many; i++)
        a[i] -= CLAMP (a[i], -b, b);
}
static void do_excess_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a - (b[i] < 0.f
                    ? CLAMP (a, b[i], -b[i])
                    : CLAMP (a, -b[i], b[i]));
}
static gfloat do_hypot_cc (gfloat a, gfloat b)
{
    return hypot (a, b);
}
static void do_hypot_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = hypot (a[i], b[i]);
}
static void do_hypot_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = hypot (a[i], b);
}
static void do_hypot_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = hypot (a, b[i]);
}
static gfloat do_atan2_cc (gfloat a, gfloat b)
{
    return atan2 (a, b);
}
static void do_atan2_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = atan2 (a[i], b[i]);
}
static void do_atan2_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = atan2 (a[i], b);
}
static void do_atan2_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = atan2 (a, b[i]);
}
static gfloat do_difsqr_cc (gfloat a, gfloat b)
{
    return a * a - b * b;
}
static void do_difsqr_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] * a[i] - b[i] * b[i];
}
static void do_difsqr_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    gfloat bsqr = b * b;
    for (i=0; i<how_many; i++)
        a[i] = a[i] * a[i] - bsqr;
}
static void do_difsqr_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    gfloat asqr = a * a;
    for (i=0; i<how_many; i++)
        b[i] = asqr - b[i] * b[i];
}
static gfloat do_sqrdif_cc (gfloat a, gfloat b)
{
    a -= b;
    return a * a;
}
static void do_sqrdif_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        a[i] -= b[i];
        a[i] *= a[i];
    }
}
static void do_sqrdif_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        a[i] -= b;
        a[i] *= a[i];
    }
}
static void do_sqrdif_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        b[i] -= a;
        b[i] *= b[i];
    }
}
static gfloat do_sqrsum_cc (gfloat a, gfloat b)
{
    a += b;
    return a * a;
}
static void do_sqrsum_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        a[i] += b[i];
        a[i] *= a[i];
    }
}
static void do_sqrsum_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        a[i] += b;
        a[i] *= a[i];
    }
}
static void do_sqrsum_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) {
        b[i] += a;
        b[i] *= b[i];
    }
}
static gfloat do_absdif_cc (gfloat a, gfloat b)
{
    return a < b ? b - a : a - b;
}
static void do_absdif_aa (gfloat *a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++) 
        a[i] = a[i] < b[i] ? b[i] - a[i] : a[i] - b[i];
}
static void do_absdif_ac (gfloat *a, gfloat b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < b ? b - a[i] : a[i] - b;
}
static void do_absdif_ca (gfloat a, gfloat *b, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        b[i] = a < b[i] ? b[i] - a : a - b[i];
}

typedef struct {
    gchar *factory_name;
    gfloat (*do_op_cc) (gfloat a, gfloat b);
    void (*do_op_aa) (gfloat *a, gfloat *b, gint how_many);
    void (*do_op_ac) (gfloat *a, gfloat b, gint how_many);
    void (*do_op_ca) (gfloat a, gfloat *b, gint how_many);
    gchar *description;
    GType type;
} SSBinaryOpDetails;

static SSBinaryOpDetails details[] = {
    {"-", do_minus_cc, do_minus_aa, do_minus_ac, do_minus_ca, "Subtraction", 0},
    {"/", do_divide_cc, do_divide_aa, do_divide_ac, do_divide_ca, "Division", 0},
    {"mod", do_mod_cc, do_mod_aa, do_mod_ac, do_mod_ca, "Mod", 0},
    {"expt", do_expt_cc, do_expt_aa, do_expt_ac, do_expt_ca, "Exponentiation", 0},
    {"<", do_lt_cc, do_lt_aa, do_lt_ac, do_lt_ca, "LT", 0},
    {">", do_gt_cc, do_gt_aa, do_gt_ac, do_gt_ca, "GT", 0},
    {"<=", do_lte_cc, do_lte_aa, do_lte_ac, do_lte_ca, "LTE", 0},
    {">=", do_gte_cc, do_gte_aa, do_gte_ac, do_gte_ca, "GTE", 0},
    {"trunc", do_trunc_cc, do_trunc_aa, do_trunc_ac, do_trunc_ca, "Trunc", 0},
    {"round", do_round_cc, do_round_aa, do_round_ac, do_round_ca, "Floor", 0},
    {"amclip", do_amclip_cc, do_amclip_aa, do_amclip_ac, do_amclip_ca, "AmClip", 0},
    {"scaleneg", do_scaleneg_cc, do_scaleneg_aa, do_scaleneg_ac, do_scaleneg_ca, "ScaleNeg", 0},
    {"clip2",do_clip2_cc, do_clip2_aa, do_clip2_ac, do_clip2_ca, "Clip2", 0},
    {"excess",do_excess_cc, do_excess_aa, do_excess_ac, do_excess_ca, "Excess", 0},
    {"hypot",do_hypot_cc, do_hypot_aa, do_hypot_ac, do_hypot_ca, "Hypot", 0},
    {"atan2",do_atan2_cc, do_atan2_aa, do_atan2_ac, do_atan2_ca, "ATan2", 0},
    {"difsqr",do_difsqr_cc, do_difsqr_aa, do_difsqr_ac, do_difsqr_ca, "DifSquares", 0},
    {"sqrdif",do_sqrdif_cc, do_sqrdif_aa, do_sqrdif_ac, do_sqrdif_ca, "SquareDif", 0},
    {"sqrsum",do_sqrsum_cc, do_sqrsum_aa, do_sqrsum_ac, do_sqrsum_ca, "SquareSum", 0},
    {"absdif",do_absdif_cc, do_absdif_aa, do_absdif_ac, do_absdif_ca, "AbsDif", 0},
    {NULL, NULL, NULL, NULL, NULL, NULL, 0}
};

/* then the generic support */
SS_PAD_TEMPLATE_SRC_FACTORY (ss_binary_op_src_factory, "output");
SS_PAD_TEMPLATE_SINK_FACTORY (ss_binary_op_sink0_factory, "input-0", 0.f);
SS_PAD_TEMPLATE_SINK_FACTORY (ss_binary_op_sink1_factory, "input-1", 0.f);


static void 		ss_binary_op_class_init		(SSBinaryOpClass *klass);
static void 		ss_binary_op_init		(SSBinaryOp *binary_op);

static void 		ss_binary_op_loop  		(GstElement *element);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_binary_op_get_type, SS_TYPE_ELEMENT,
                          "SSBinaryOp", SSBinaryOpClass, SSBinaryOp,
                          NULL /* abstract type */, NULL /* abstract type */);

static void
ss_binary_op_class_init (SSBinaryOpClass *klass) 
{
    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);
}

static void 
ss_binary_op_init (SSBinaryOp *this) 
{
    SSBinaryOpDetails *d;

    for (d=details; d; d++) {
        if (G_TYPE_FROM_INSTANCE (this) == d->type) {
            this->do_op_cc = d->do_op_cc;
            this->do_op_aa = d->do_op_aa;
            this->do_op_ac = d->do_op_ac;
            this->do_op_ca = d->do_op_ca;
            break;
        }
    }

    gst_element_set_loop_function (GST_ELEMENT (this), ss_binary_op_loop);

    this->srcpad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_binary_op_src_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->srcpad);

    this->sink[0] = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_binary_op_sink0_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->sink[0]);

    this->sink[1] = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_binary_op_sink1_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->sink[1]);
}

static void
ss_binary_op_loop (GstElement *element)
{
    SSBinaryOp *this;
    gint j;
    gint buffer_frames;
    gint num_to_write;
    GstData *buf_out = NULL, *control_out = NULL;
    GstData *buf_in[2];
    gfloat *data_out, *data_in[2];
    GstPad *pad;
    gfloat control_value[2];
    gboolean eos, audio_rate = FALSE;
    SSPad *spad;

    this = (SSBinaryOp*)element;

    num_to_write = buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;
    if (!buffer_frames) {
        gst_element_error (element, "Caps were never set, bailing...");
        return;
    }

    eos = FALSE;

    for (j = 0; j < 2; j++) {
        pad = this->sink[j];
        spad = (SSPad*)pad;
        control_value[j] = 0.f;
        data_in[j] = NULL;
      
        if (GST_PAD_IS_USABLE (pad))
            buf_in[j] = gst_pad_pull (pad);
        else
            buf_in[j] = NULL;

        if (buf_in[j] && GST_IS_EVENT (buf_in[j])) {
            /* stupidly assuming it's an eos event */
            num_to_write = 0;
            eos = TRUE;
            gst_data_unref (buf_in[j]);
            break;
        } else if (!buf_in[j] || SS_IS_CONTROL (buf_in[j])) {
            /* a control input */
            if (buf_in[j]) {
                control_value[j] = SS_CONTROL_VALUE (buf_in[j]);

                if (!control_out)
                    control_out = buf_in[j];
                else
                    gst_data_unref (buf_in[j]);
            } else {
                control_value[j] = spad->value;
            }
        } else {
            /* an audio input, we have a buffer */
            audio_rate = TRUE;

            if (!buf_out) {
                /* reuse the new input buffer */
                buf_out = buf_in[j];
                data_out = data_in[j] = SS_BUFFER_DATA (buf_in[j]);
            } else {
                data_in[j] = SS_BUFFER_DATA (buf_in[j]);
            }
            num_to_write = MIN (num_to_write, SS_BUFFER_GET_FRAMES (buf_in[j]));
        }
    }

    if (eos) {
        if (buf_out) {
            if (data_in[1] && buf_out != buf_in[1])
                gst_data_unref (buf_in[1]);
            gst_data_unref (buf_out);
        }
        gst_pad_push (this->srcpad, GST_DATA (gst_event_new (GST_EVENT_EOS)));
        gst_element_set_eos (element);
    } else if (audio_rate) {
        /* audio rate */
        SS_BUFFER_SET_FRAMES (buf_out, num_to_write);
        /* operate on the control inputs at the last minute */
        if (data_in[0] && ! data_in[1])
            this->do_op_ac (data_in[0], control_value[1], num_to_write);
        else if (data_in[0] && data_in[1])
            this->do_op_aa (data_in[0], data_in[1], num_to_write);
        else if (!data_in[0] && data_in[1])
            this->do_op_ca (control_value[0], data_in[1], num_to_write);
        else
            g_assert_not_reached ();
        
        if (data_in[1] && buf_out != buf_in[1])
            gst_data_unref (buf_in[1]);
        if (control_out)
            gst_data_unref (control_out);

        ss_pad_push ((SSPad*)this->srcpad, buf_out);
    } else {
        /* control rate */
        GstData *out = control_out ? control_out : ss_control_new ((SSPad*)this->srcpad);
        SS_CONTROL_VALUE (out) = this->do_op_cc (control_value[0], control_value[1]);
        ss_pad_push ((SSPad*)this->srcpad, out);
    }
}

gboolean
init_ss_binary_op (GstPlugin *plugin)
{
    SSBinaryOpDetails *d;
    GstElementClass *class;

    for (d=details; d->factory_name!=NULL; d++) {
        GstElementDetails *edetails;
        static const GTypeInfo type_info = {
            sizeof (SSBinaryOpClass), NULL, NULL,
            (GClassInitFunc) ss_binary_op_class_init, NULL, NULL,
            sizeof (SSBinaryOp), 0,
            (GInstanceInitFunc) ss_binary_op_init,
        };

        edetails = g_new0 (GstElementDetails, 1);
        edetails->longname = d->factory_name;
        edetails->klass = "Soundscrape/Operators/Binary";
        edetails->description = d->description;
        edetails->author = "Andy Wingo <wingo at pobox dot com>";

        d->type = g_type_register_static (SS_TYPE_BINARY_OP,
                                          g_strdup_printf ("SSBinaryOp%s", d->description),
                                          &type_info, 0);
        class = g_type_class_ref (d->type);

        gst_element_class_set_details (class, edetails);
        gst_element_class_add_pad_template (class, ss_binary_op_sink0_factory ());
        gst_element_class_add_pad_template (class, ss_binary_op_sink1_factory ());
        gst_element_class_add_pad_template (class, ss_binary_op_src_factory ());

        gst_element_register (plugin, d->factory_name, GST_RANK_NONE, d->type);
    }
    
    return TRUE;
}

