/* Soundscrape
 * ss-buffer-pool.c: Buffer pool for soundscrape buffers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Based on gstbufferpool-default.c,
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *                    2000 Wim Taymans <wtay@chello.be>
 */


#include <gst/gstbuffer.h>
#include <gst/gstinfo.h>
#include <gst/gstmemchunk.h>

#include "ss-buffer-pool.h"


static GstBuffer*	ss_buffer_pool_buffer_new	(GstBufferPool *pool,
                                                         guint64 offset, guint size,
                                                         gpointer user_data);
static void		ss_buffer_pool_buffer_free	(GstBufferPool *pool,
                                                         GstBuffer *buffer,
                                                         gpointer user_data);
static void		ss_buffer_pool_free		(GstData *pool);


static GstMemChunk *data_chunk = NULL;
static guint buffer_frames = 0;
static GstBufferPool *ss_buffer_pool = NULL;

#define POOL_SIZE 30

GstBufferPool*
ss_buffer_pool_set_buffer_frames (guint nframes)
{
  guint real_buffer_size;
  
  g_return_val_if_fail (nframes > 0, NULL);

  if (buffer_frames == nframes)
    return ss_buffer_pool;
  
  buffer_frames = nframes;

  /* round up to the nearest 32 bytes for cache-line and other efficiencies */
  real_buffer_size = (((buffer_frames * sizeof(gfloat) - 1) / 32) + 1) * 32;
  data_chunk = gst_mem_chunk_new ("SSBufferPool", real_buffer_size, 
                                  real_buffer_size * POOL_SIZE, G_ALLOC_AND_FREE);
    
  ss_buffer_pool = gst_buffer_pool_new (ss_buffer_pool_free,
                                        NULL, /* pool copy */
                                        ss_buffer_pool_buffer_new,
                                        NULL, /* buffer copy */
                                        ss_buffer_pool_buffer_free,
                                        NULL);
  
  return ss_buffer_pool;
}

static GstBuffer* 
ss_buffer_pool_buffer_new (GstBufferPool *pool, guint64 offset,
                            guint size, gpointer user_data)
{
  GstBuffer *buffer;
  
  buffer = gst_buffer_new ();
  
  GST_BUFFER_DATA (buffer) = gst_mem_chunk_alloc (data_chunk);
  
  GST_BUFFER_SIZE (buffer)    = buffer_frames * sizeof(gfloat);
  GST_BUFFER_MAXSIZE (buffer) = buffer_frames * sizeof(gfloat);
  
  return buffer;
}

static void
ss_buffer_pool_buffer_free (GstBufferPool *pool, GstBuffer *buffer, gpointer user_data)
{
  gpointer data = GST_BUFFER_DATA (buffer);
  
  gst_mem_chunk_free (data_chunk, data);

  GST_BUFFER_DATA (buffer) = NULL;

  gst_buffer_default_free (buffer);
}

static void
ss_buffer_pool_free (GstData *data) 
{
  /* the soundscrape buffer pool should never be freed */
  g_assert_not_reached ();
}

GstData*
ss_buffer_new (void)
{
    return GST_DATA (gst_buffer_new_from_pool (ss_buffer_pool, 0, 0));
}

GstData*
ss_buffer_new_constant_val (gfloat val)
{
    gint i;
    gfloat *data;
    GstData *new_buf = ss_buffer_new ();
    
    data = SS_BUFFER_DATA (new_buf);
    for (i=0; i<buffer_frames; i++)
        data[i] = val;
    
    return new_buf;
}

GstData*
ss_buffer_new_lin_interp (gfloat start, gfloat end)
{
    gfloat *data, slope;
    gint i;
    GstData *new_buf;

    if (start == end)
        return ss_buffer_new_constant_val (start);
    
    new_buf = ss_buffer_new ();
    slope = (start - end) / buffer_frames;
    
    data = SS_BUFFER_DATA (new_buf);
    for (i=0; i<buffer_frames; i++) {
        data[i] = start;
        start += slope;
    }
    
    return GST_DATA (new_buf);
}
