/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-enveloptor.c: Apply an envelope to incoming date
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Inspiration taken from SuperCollider's EnvGen.
 */

#include <config.h>
#include "ss-enveloptor.h"
#include <guile-gnome-gobject-primitives.h>
#include <math.h>


static GstElementDetails enveloptor_details = {
  "Enveloptor",
  "Soundscrape/Dynamic",
  "Apply an envelope to incoming data",
  "Andy Wingo <wingo at pobox dot com>",
};

SS_PAD_TEMPLATE_SRC_FACTORY (ss_enveloptor_src_factory, "output");
SS_PAD_TEMPLATE_SINK_FACTORY (ss_enveloptor_input_factory, "input", 0.f);
SS_PAD_TEMPLATE_SINK_FACTORY (ss_enveloptor_trigger_factory, "trigger", 0.f);


static void 	ss_enveloptor_class_init	(SSEnveloptorClass *klass);
static void 	ss_enveloptor_init		(SSEnveloptor *enveloptor);

static void 	ss_enveloptor_loop  		(GstElement *element);

static void	ss_enveloptor_cleanup		(GstElement *element);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_enveloptor_get_type, SS_TYPE_ELEMENT,
                          "SSEnveloptor", SSEnveloptorClass, SSEnveloptor,
                          ss_enveloptor_class_init, ss_enveloptor_init);

static void
ss_enveloptor_class_init (SSEnveloptorClass *klass) 
{
    GstElementClass *eclass = GST_ELEMENT_CLASS (klass);
    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);

    SS_ELEMENT_CLASS (klass)->cleanup = ss_enveloptor_cleanup;

    gst_element_class_set_details (eclass, &enveloptor_details);
    gst_element_class_add_pad_template (eclass, ss_enveloptor_src_factory ());
    gst_element_class_add_pad_template (eclass, ss_enveloptor_input_factory ());
    gst_element_class_add_pad_template (eclass, ss_enveloptor_trigger_factory ());
}

static void 
ss_enveloptor_init (SSEnveloptor *this) 
{
    GstElement *element = GST_ELEMENT (this);

    ss_enveloptor_cleanup (element);
    
    this->src = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_enveloptor_src_factory ()));
    gst_element_add_pad (element, this->src);

    this->trigger = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_enveloptor_trigger_factory ()));
    gst_element_add_pad (element, this->trigger);

    this->input = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_enveloptor_input_factory ()));
    gst_element_add_pad (element, this->input);

    gst_element_set_loop_function (element, ss_enveloptor_loop);
}

static void
ss_enveloptor_loop (GstElement *element)
{
    SSEnveloptor *this = (SSEnveloptor*)element;
    gint buffer_frames;
    GstData *buf_in = NULL;
    gfloat *levels = this->levels;
    //gfloat *times = this->times;
    SSWarp shape;
    gfloat level, final_level;
        
    buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;

    if (!levels) {
        gst_element_error (element, "Envelope levels were never set, bailing...");
        return;
    }
    if (!buffer_frames) {
        gst_element_error (element, "Caps were never set, bailing...");
        return;
    }

    if (GST_PAD_IS_USABLE (this->trigger)) {
        gst_element_error (element, "Sustain is not implemented in the enveloptor");
        return;
    }

    /* these guarantee we have the proper kind of buffer */
    if (SS_ELEMENT_IS_RATE (this, CONTROL))
        buf_in = ss_pad_update_control (SS_PAD (this->input));
    else
        buf_in = ss_pad_pull (SS_PAD (this->input), buffer_frames);

    if (buf_in && GST_IS_EVENT (buf_in)) {
        gst_pad_event_default (this->src, GST_EVENT (buf_in));
        return;
    }

    if (!buf_in || SS_IS_CONTROL (buf_in)) {
        /* control rate input and output */
        if (buf_in)
            ss_control_set_pad (buf_in, this->src);
        else
            buf_in = ss_control_new (SS_PAD (this->src));
        
        if (this->count < 0) {
            /* we're going into a new stage */
            this->stage++;
            if (this->stage >= this->nstages) {
                /* we're finished, yo */
                gst_pad_push (this->src, GST_DATA (gst_event_new (GST_EVENT_EOS)));
                gst_element_set_eos (element);
                return;
            }
            this->count = MAX (this->times[this->stage] * SS_ELEMENT (this)->rate, 1);
            shape = (this->count == 1) ? SS_WARP_LINEAR : this->warp;
            this->level = level = this->levels[this->stage];
            final_level = this->levels[this->stage + 1];

            switch (shape) {
            case SS_WARP_LINEAR:
                this->grow = final_level - level;
                break;
            case SS_WARP_EXPONENTIAL:
                this->grow = pow (final_level / level, 1. / this->count);
                break;
            default:
                gst_element_error (element, "unsupported warp");
                break;
            }
        } else {
            switch (shape) {
            case SS_WARP_LINEAR:
                this->level += this->grow;
                break;
            case SS_WARP_EXPONENTIAL:
                this->level *= this->grow;
                break;
            default:
                gst_element_error (element, "unsupported warp");
                break;
            }
        }
        
        SS_PAD_VALUE (this->src) = SS_PAD_VALUE (this->input) * this->level;

        ss_pad_push (SS_PAD (this->src), buf_in);

        this->count -= buffer_frames;
    } else {
        /* audio rate input and output */
        gint i;
        gfloat *data = SS_BUFFER_DATA (buf_in);
        
        if (this->count < 0) {
            //new_stage_audio:
            /* we're going into a new stage */
            this->stage++;
            if (this->stage >= this->nstages) {
                /* we're finished, yo */
                gst_pad_push (this->src, GST_DATA (gst_event_new (GST_EVENT_EOS)));
                gst_element_set_eos (element);
                return;
            }
            this->count = MAX (this->times[this->stage] * SS_ELEMENT (this)->rate, 1);
            shape = (this->count == 1) ? SS_WARP_LINEAR : this->warp;
            this->level = level = this->levels[this->stage];
            final_level = this->levels[this->stage + 1];

            switch (shape) {
            case SS_WARP_LINEAR:
                this->grow = final_level - level;
                break;
            case SS_WARP_EXPONENTIAL:
                this->grow = pow (final_level / level, 1. / this->count);
                break;
            default:
                gst_element_error (element, "unsupported warp");
                break;
            }
        } else {
            switch (shape) {
            case SS_WARP_LINEAR:
                this->level += this->grow;
                break;
            case SS_WARP_EXPONENTIAL:
                this->level *= this->grow;
                break;
            default:
                gst_element_error (element, "unsupported warp");
                break;
            }
        }
        
        SS_PAD_VALUE (this->src) = SS_PAD_VALUE (this->input) * this->level;

        ss_pad_push (SS_PAD (this->src), buf_in);

        this->count -= buffer_frames;

        /* fixme: implement properly for the case where the buffer can't be
           reused */
        buffer_frames = MIN (buffer_frames, SS_BUFFER_GET_FRAMES (buf_in));
        for (i=0; i<buffer_frames; i++) {
            data[i] *= level;
        }

        ss_pad_push (SS_PAD (this->src), buf_in);
    }
}

static void
ss_enveloptor_cleanup (GstElement *element)
{
    SSEnveloptor *this = (SSEnveloptor*)element;

    g_free (this->levels);
    this->levels = NULL;
    g_free (this->times);
    this->times = NULL;
    this->warp = SS_WARP_LINEAR;
    this->sustain = -1;
    this->count = -1;
    this->nstages = 0;
}

void
ss_enveloptor_set_envelope (SSEnveloptor *e, SCM envelope)
#define FUNC_NAME "%ss-enveloptor-set-envelope"
{
    SCM slevels, stimes, swarp, ssustain;
    //GList *levels = NULL, *times = NULL, *walk = NULL;
    //SSWarp warp = SS_WARP_LINEAR;
    //gint sustain = -1;

    slevels = scm_slot_ref (envelope, scm_str2symbol ("levels"));
    stimes = scm_slot_ref (envelope, scm_str2symbol ("times"));
    swarp = scm_slot_ref (envelope, scm_str2symbol ("curvature"));
    ssustain = scm_slot_ref (envelope, scm_str2symbol ("sustain"));

    SCM_VALIDATE_NONEMPTYLIST (1, slevels);
    SCM_VALIDATE_NONEMPTYLIST (1, stimes);
    //SCM_ASSERT (
    for (; SCM_NNULLP (slevels); slevels = SCM_CDR (slevels)) {
        gfloat level;
        SCM_VALIDATE_FLOAT_COPY (1, SCM_CAR (slevels), level);
        //levels = g_list_append (levels, level);
    }
    
    for (; SCM_NNULLP (slevels); slevels = SCM_CDR (slevels)) {
        gfloat level;
        SCM_VALIDATE_FLOAT_COPY (1, SCM_CAR (slevels), level);
        //levels = g_list_append (levels, level);
    }
}
#undef FUNC_NAME

gboolean 
init_ss_enveloptor (GstPlugin *plugin) 
{
    return gst_element_register (plugin, "enveloptor", GST_RANK_NONE, SS_TYPE_ENVELOPTOR);
}
