/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-ladspa.c: Wrap LADSPA plugins as SSElements
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Based on gstladspa.c:
 * Copyright (C) 2001 Steve Baker, <stevebaker_org@yahoo.co.uk>.
 */

#include <string.h>
#include <math.h>
#include <gst/control/control.h>

#include "ss-ladspa.h"
#include <ladspa.h>
#include "ss-ladspa-utils.h"


static void			ss_ladspa_class_init	(SSLADSPAClass *klass);
static void			ss_ladspa_init		(SSLADSPA *ladspa);

static GstPadLinkReturn		ss_ladspa_link		(GstPad *pad, GstCaps *caps);

static gboolean			ss_ladspa_instantiate	(SSLADSPA *ladspa);
static void			ss_ladspa_activate	(SSLADSPA *ladspa);
static void			ss_ladspa_deactivate	(SSLADSPA *ladspa);

static GstElementStateReturn	ss_ladspa_change_state	(GstElement *element);
static void			ss_ladspa_loop		(GstElement *element);


static GstElementClass *parent_class = NULL;

static GHashTable *ladspa_descriptors;

GST_DEBUG_CATEGORY_STATIC (ladspa_debug);
#define SSL_INFO(...) \
    GST_CAT_LEVEL_LOG (ladspa_debug, GST_LEVEL_INFO, NULL, __VA_ARGS__)
#define SSL_INFO_OBJ(obj,...) \
    GST_CAT_LEVEL_LOG (ladspa_debug, GST_LEVEL_INFO, obj, __VA_ARGS__)


static void
ss_ladspa_class_init (SSLADSPAClass *klass)
{
    GObjectClass *gobject_class;
    GstElementClass *gstelement_class;
    LADSPA_Descriptor *desc;
    gint i, sink, src, audioin, audioout, controlin, controlout;

    gobject_class = (GObjectClass*)klass;
    gstelement_class = (GstElementClass*)klass;

    gstelement_class->change_state = ss_ladspa_change_state;

    /* look up and store the ladspa descriptor */
    klass->descriptor = g_hash_table_lookup(ladspa_descriptors,
                                            GINT_TO_POINTER (G_TYPE_FROM_CLASS (klass)));
    desc = klass->descriptor;

    klass->numports = desc->PortCount;

    /* walk through the ports, count the input and output ports */
    for (i=0; i<desc->PortCount; i++) {
        if (LADSPA_IS_PORT_INPUT (desc->PortDescriptors[i])) {
            if (LADSPA_IS_PORT_CONTROL (desc->PortDescriptors[i]))
                klass->numcontrol_sinks++;
            else
                klass->numaudio_sinks++;
            klass->numsinks++;
        } else {
            if (LADSPA_IS_PORT_CONTROL (desc->PortDescriptors[i]))
                klass->numcontrol_srcs++;
            else
                klass->numaudio_srcs++;
            klass->numsrcs++;
        }
    }

    klass->sink_ports = g_new0 (gint, klass->numsinks);
    klass->audio_sinks = g_new0 (gint, klass->numaudio_sinks);
    klass->control_sinks = g_new0 (gint, klass->numcontrol_sinks);

    klass->src_ports = g_new0 (gint, klass->numsrcs);
    klass->audio_srcs = g_new0 (gint, klass->numaudio_srcs);
    klass->control_srcs = g_new0 (gint, klass->numcontrol_srcs);

    sink = src = audioin = audioout = controlin = controlout = 0;

    /* walk through the ports, note the portnums for srcpads and sinkpads */
    for (i=0; i<desc->PortCount; i++) {
        if (LADSPA_IS_PORT_INPUT (desc->PortDescriptors[i])) {
            if (LADSPA_IS_PORT_CONTROL (desc->PortDescriptors[i]))
                klass->control_sinks[controlin++] = sink;
            else
                klass->audio_sinks[audioin++] = sink;
            klass->sink_ports[sink++] = i;
        } else {
            if (LADSPA_IS_PORT_CONTROL (desc->PortDescriptors[i]))
                klass->control_srcs[controlout++] = src;
            else
                klass->audio_srcs[audioout++] = src;
            klass->src_ports[src++] = i;
        }
    }
}

static void
ss_ladspa_init (SSLADSPA *ladspa)
{
    SSLADSPAClass *oclass;
    LADSPA_Descriptor *desc;
    GList *l;
    gint i, sinkcount, srccount;

    oclass = (SSLADSPAClass*)(G_OBJECT_GET_CLASS (ladspa));

    desc = oclass->descriptor;
    ladspa->descriptor = oclass->descriptor;
  
    ladspa->srcs = g_new0 (GstPad*,oclass->numsrcs);
    ladspa->sinks = g_new0 (GstPad*,oclass->numsinks);
  
    srccount = sinkcount = 0;

    l = GST_ELEMENT_CLASS (oclass)->padtemplates;
    for (i=0; i<oclass->numports; i++,l=l->next) {
        /* these are the ports corresponding to ladspa pads */
        GstPad *pad = ss_pad_new_from_template ((SSPadTemplate*)l->data);

        gst_element_add_pad (GST_ELEMENT (ladspa), pad);
        gst_pad_set_link_function (pad, ss_ladspa_link);

        if (GST_PAD_IS_SINK (pad))
            ladspa->sinks[sinkcount++] = pad;
        else
            ladspa->srcs[srccount++] = pad;
    }
    
    for (; l; l=l->next) {
        GstPad *pad = ss_pad_new_from_template ((SSPadTemplate*)l->data);

        gst_element_add_pad (GST_ELEMENT (ladspa), pad);
        gst_pad_set_link_function (pad, ss_ladspa_link);

        if (strcmp (GST_PAD_TEMPLATE_NAME_TEMPLATE (l->data), "add") == 0)
            ladspa->add = pad;
        else
            ladspa->mul = pad;
    }

    ladspa->controls_in = g_new0 (GstData*, oclass->numcontrol_sinks);
    ladspa->controls_out = g_new0 (GstData*, oclass->numcontrol_srcs);
    ladspa->buffers_in  = g_new0 (GstData*, sinkcount);
    ladspa->buffers_out = g_new0 (GstData*, srccount);

    /* nonzero default needed to instantiate() some plugins */
    SS_ELEMENT_CAST (ladspa)->rate = 44100;

    ladspa->activated = FALSE;
    ladspa->inplace_broken = LADSPA_IS_INPLACE_BROKEN (ladspa->descriptor->Properties);

    /* because we allow values to be defaulted if a sink pad is not linked, we
       need to always use the loop function. perhaps we could change this if we
       connected to the ::linked signal... */

    SSL_INFO_OBJ (ladspa, "loop mode with %d sink pads and %d src pads",
                  sinkcount, srccount);

    gst_element_set_loop_function (GST_ELEMENT (ladspa), ss_ladspa_loop);

    ss_ladspa_instantiate (ladspa);
}

static GstPadLinkReturn
ss_ladspa_link (GstPad *pad, GstCaps *caps)
{
    SSLADSPA *ladspa = (SSLADSPA *) GST_PAD_PARENT (pad);
    SSElement *sselement = SS_ELEMENT_CAST (ladspa);
    SSLADSPAClass *oclass = (SSLADSPAClass *) (G_OBJECT_GET_CLASS (ladspa));
    guint i;
    gint rate, buffer_frames;

    if (GST_CAPS_IS_FIXED (caps)) {
        gst_caps_get_int (caps, "rate", &rate);
        gst_caps_get_int (caps, "buffer-frames", &buffer_frames);

        /* have to instantiate ladspa plugin when sample rate changes (groan) */
        if (sselement->rate != rate) {
            sselement->rate = rate;
            if (! ss_ladspa_instantiate (ladspa))
                return GST_PAD_LINK_REFUSED;
        }

        for (i=0; i<oclass->numsrcs; i++)
            if (pad != ladspa->srcs[i]
                && gst_pad_try_set_caps (ladspa->srcs[i], caps) <= 0)
                return GST_PAD_LINK_REFUSED;
        for (i=0; i<oclass->numsinks; i++)
            if (pad != ladspa->sinks[i]
                && gst_pad_try_set_caps (ladspa->sinks[i], caps) <= 0)
                return GST_PAD_LINK_REFUSED;
        if (ladspa->add && pad != ladspa->add
            && gst_pad_try_set_caps (ladspa->add, caps) <= 0)
            return GST_PAD_LINK_REFUSED;
        if (ladspa->mul && pad != ladspa->mul
            && gst_pad_try_set_caps (ladspa->mul, caps) <= 0)
            return GST_PAD_LINK_REFUSED;

        ss_buffer_pool_set_buffer_frames (buffer_frames);

        sselement->rate = rate;
        sselement->buffer_frames = buffer_frames;

        return GST_PAD_LINK_OK;
    }

    return GST_PAD_LINK_DELAYED;
}

static gboolean
ss_ladspa_instantiate (SSLADSPA *ladspa)
{
    LADSPA_Descriptor *desc;
    int i;
    SSLADSPAClass *oclass = (SSLADSPAClass*)(G_OBJECT_GET_CLASS (ladspa));
    gboolean was_activated;
  
    desc = ladspa->descriptor;
  
    /* check for old handle */
    was_activated = ladspa->activated;
    if (ladspa->handle != NULL) {
        ss_ladspa_deactivate(ladspa);
        desc->cleanup(ladspa->handle);
    }
        
    /* instantiate the plugin */ 
    SSL_INFO_OBJ (ladspa, "instantiating the plugin");
  
    ladspa->handle = desc->instantiate (desc, SS_ELEMENT_CAST (ladspa)->rate);
    g_return_val_if_fail (ladspa->handle != NULL, FALSE);

    /* connect control ports directly to the pad values */
    for (i=0; i<oclass->numcontrol_sinks; i++)
        desc->connect_port(ladspa->handle,
                           oclass->sink_ports[oclass->control_sinks[i]],
                           &(SS_PAD (ladspa->sinks[oclass->control_sinks[i]])->value));
    for (i=0; i<oclass->numcontrol_srcs; i++)
        desc->connect_port(ladspa->handle,
                           oclass->src_ports[oclass->control_srcs[i]],
                           &(SS_PAD (ladspa->srcs[oclass->control_srcs[i]])->value));
    
    /* reactivate if it was activated before the reinstantiation */
    if (was_activated)
        ss_ladspa_activate(ladspa);

    return TRUE;
}

static GstElementStateReturn
ss_ladspa_change_state (GstElement *element)
{
    SSLADSPA *ladspa = (SSLADSPA*)element;

    switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_READY_TO_PAUSED:
        ss_ladspa_activate (ladspa);
        break;
    case GST_STATE_PAUSED_TO_READY:
        ss_ladspa_deactivate (ladspa);
        break;
    default:
        break;
    }

    if (GST_ELEMENT_CLASS (parent_class)->change_state)
        return GST_ELEMENT_CLASS (parent_class)->change_state (element);

    return GST_STATE_SUCCESS;
}

static void
ss_ladspa_activate(SSLADSPA *ladspa)
{
    LADSPA_Descriptor *desc = ladspa->descriptor;
  
    if (ladspa->activated)
        ss_ladspa_deactivate (ladspa);
  
    SSL_INFO_OBJ (ladspa, "activating");

    /* activate the plugin (function might be null) */
    if (desc->activate != NULL)
        desc->activate (ladspa->handle);

    ladspa->activated = TRUE;
}

static void
ss_ladspa_deactivate(SSLADSPA *ladspa)
{
    LADSPA_Descriptor *desc = ladspa->descriptor;

    SSL_INFO_OBJ (ladspa, "deactivating");

    /* deactivate the plugin (function might be null) */
    if (ladspa->activated && (desc->deactivate != NULL))
        desc->deactivate (ladspa->handle);

    ladspa->activated = FALSE;
}

static void
ss_ladspa_loop(GstElement *element)
{
    guint	  i, j, num_srcs, num_sinks, frames_to_process;
    GstData	**controls_in, **controls_out;
    GstData	**buffers_in, **buffers_out;
    SSElementRate rate;

    SSLADSPA       *ladspa = (SSLADSPA *)element;
    LADSPA_Descriptor *desc = ladspa->descriptor;
    SSLADSPAClass  *oclass = (SSLADSPAClass*)(G_OBJECT_GET_CLASS (ladspa));

    num_sinks = oclass->numsinks;
    num_srcs = oclass->numsrcs;
    rate = SS_ELEMENT_RATE (element);
    controls_in = ladspa->controls_in;
    controls_out = ladspa->controls_out;
    buffers_in = ladspa->buffers_in;
    buffers_out = ladspa->buffers_out;

    frames_to_process = SS_ELEMENT (ladspa)->buffer_frames;

    if (!frames_to_process)
        gst_element_error (element, "Caps were never set, bailing...");

    /* get input buffers from the source pads */
    for (i=0; i<oclass->numaudio_sinks; i++) {
        buffers_in[i] = ss_pad_pull (SS_PAD (ladspa->sinks[oclass->audio_sinks[i]]),
                                     frames_to_process);
        if (GST_IS_EVENT (buffers_in[i]))
            goto push_eos_on_all_pads;

        frames_to_process = MIN (frames_to_process, SS_BUFFER_GET_FRAMES (buffers_in[i]));

        desc->connect_port (ladspa->handle,
                            oclass->sink_ports[oclass->audio_sinks[i]],
                            SS_BUFFER_DATA (buffers_in[i]));
    }

    /* control ports are connected to the pads' values already, we just need to
     * update the values */
    for (i=0; i<oclass->numcontrol_sinks; i++)
        controls_in[i] =
            ss_pad_update_control (SS_PAD (ladspa->sinks[oclass->control_sinks[i]]));
    
    /* prepare audio output buffers */
    for (i=0,j=0; i<oclass->numaudio_srcs; i++) {
        if (!ladspa->inplace_broken && j < oclass->numaudio_sinks)
            buffers_out[i] = buffers_in[j++];
        else
            buffers_out[i] = ss_buffer_new ();

        desc->connect_port (ladspa->handle,
                            oclass->src_ports[oclass->audio_srcs[i]],
                            SS_BUFFER_DATA (buffers_out[i]));
    }
    
    /* clean up the control inputs, reusing them if possible */
    for (i=0,j=0; i<oclass->numcontrol_sinks; i++) {
        if (controls_in[i]) {
            if (rate == SS_ELEMENT_RATE_CONTROL && j<oclass->numcontrol_srcs) {
                controls_out[j++] = controls_in[i];
            } else {
                gst_data_unref (controls_in[i]);
                controls_in[i] = NULL;
            }
        }
    }
    for (j=0; j<oclass->numcontrol_srcs; j++)
        if (controls_out[j])
            /* connect the control to the source pad's value */
            ss_control_set_pad (controls_out[j], ladspa->srcs[oclass->control_srcs[j]]);
        else
            /* make a new control buffer if necessary */
            controls_out[j] = ss_control_new (SS_PAD (ladspa->srcs[oclass->control_srcs[j]]));

    /* truncate the buffers if necessary (e.g., if one pad is about to eos) */
    if (frames_to_process != SS_ELEMENT (ladspa)->buffer_frames) {
        g_return_if_fail (frames_to_process != 0);

        if (rate == SS_ELEMENT_RATE_AUDIO)
            for (i=0; i<num_srcs; i++)
                if (buffers_out[i])
                    SS_BUFFER_SET_FRAMES (buffers_out[i], frames_to_process);
    }

    /* do it */
    desc->run (ladspa->handle, frames_to_process);

    /* take care of :mul and :add for audio buffers (controls can be done
     * later) */
    if (ladspa->mul)
        ss_pad_do_mul (SS_PAD (ladspa->mul), buffers_out, oclass->numaudio_srcs);
    if (ladspa->add)
        ss_pad_do_add (SS_PAD (ladspa->add), buffers_out, oclass->numaudio_srcs);

    /* clean up, get ready to push... */
    for (i=0; i<oclass->numaudio_sinks; i++) {
        if (i >= oclass->numaudio_srcs || buffers_in[i] != buffers_out[i])
            gst_data_unref (buffers_in[i]);
        buffers_in[i] = NULL;
    }
    
    /* let's push, yo */
    if (rate == SS_ELEMENT_RATE_CONTROL) {
        /* first we push control sources */
        for (i=0; i<oclass->numcontrol_srcs; i++) {
            /* we need to take care of :mul and :add still */
            if (ladspa->mul)
                SS_CONTROL_VALUE (controls_out[i]) *= SS_PAD_VALUE (ladspa->mul);
            if (ladspa->add)
                SS_CONTROL_VALUE (controls_out[i]) += SS_PAD_VALUE (ladspa->add);
            
            ss_pad_push (SS_PAD (ladspa->srcs[oclass->control_srcs[i]]), 
                         controls_out[i]);
            controls_out[i] = NULL;
        }

        /* if we're told to operate at control rate but we have audio source ports,
           update the src pads' values and make new control buffers */
        for (i=0; i<oclass->numaudio_srcs; i++) {
            SSPad *spad = SS_PAD (ladspa->srcs[oclass->audio_srcs[i]]);
            
            SS_PAD_VALUE (spad) = SS_BUFFER_DATA (buffers_out[i])[0];

            gst_data_unref (buffers_out[i]);
            buffers_out[i] = NULL;

            ss_pad_push (spad, ss_control_new (spad));
        }
    } else {
        /* first we push audio rate sources */
        for (i=0; i<oclass->numaudio_srcs; i++) {
            ss_pad_push (SS_PAD (ladspa->srcs[oclass->audio_srcs[i]]),
                         buffers_out[i]);
            buffers_out[i] = NULL;
        }

        /* if we're told to operate at audio rate but we have control source
           ports, just dup the control value for the length of the buffer */
        for (i=0; i<oclass->numcontrol_srcs; i++) {
            gfloat *data;
            gfloat value;
            GstData *buffer;

            buffer = ss_buffer_new ();
            data = SS_BUFFER_DATA (buffer);
            value = SS_PAD_VALUE (ladspa->srcs[oclass->control_srcs[i]]);

            /* take care of :mul and :add now */
            if (ladspa->mul)
                value *= SS_PAD_VALUE (ladspa->mul);
            if (ladspa->add)
                value += SS_PAD_VALUE (ladspa->add);
            
            for (j=1; j<frames_to_process; j++)
                data[j] = value;

            ss_pad_push (SS_PAD (ladspa->srcs[oclass->control_srcs[i]]),
                         buffer);
        }
    }

    return;

push_eos_on_all_pads:
    for (i=0; i<num_srcs; i++)
        gst_pad_push (ladspa->srcs[i], GST_DATA (gst_event_new (GST_EVENT_EOS)));
    gst_element_set_eos (element);
    /* FIXME: unref any live buffers */
}

static void
ss_ladspa_class_uniquify_pad_names (GstElementClass *class) 
{
    int de_dupifier = 1;

    /* ay, I wish I could do this in scheme... */
    GList *walk, *l, *dups = NULL;

    for (walk = class->padtemplates; walk; walk = walk->next)
        for (l = class->padtemplates; l != walk; l = l->next)
            if (strcmp (GST_PAD_TEMPLATE (l->data)->name_template,
                        GST_PAD_TEMPLATE (walk->data)->name_template) == 0)
                dups = g_list_append (g_list_append (dups, walk->data), l->data);
    
    /* it's possible to have some duplicate duplicates (!), so go through and
     * remove those -- am I crazy? */
    l = NULL;
    for (walk = dups; walk; walk = walk->next)
        if (g_list_find (walk->next, walk->data))
            l = g_list_append (l, walk->data);
    for (walk = l; walk; walk = walk->next)
        dups = g_list_remove (dups, walk->data);
    for (walk = dups; walk; walk = walk->next) {
        gchar *oldname = GST_PAD_TEMPLATE (walk->data)->name_template;
        GST_PAD_TEMPLATE (walk->data)->name_template = g_strdup_printf ("%s-%d", oldname, de_dupifier++);
        g_free (oldname);
    }

    g_list_free (l);
    g_list_free (dups);
}

static GstPadTemplate*
ladspa_sink_factory (gchar *name, gfloat default_value)
{
    GstPadTemplate *ret = ss_pad_template_sink_new (name, default_value);
    g_free (name);
    return ret;
}

static GstPadTemplate*
ladspa_src_factory (gchar *name)
{
    GstPadTemplate *ret = ss_pad_template_src_new (name);
    g_free (name);
    return ret;
}

static gchar*
ss_ladspa_sanitize_name (const gchar *name)
{
    char *paren;
    char *ret;
    int i;
    
    /* returns a new copy */
    ret = g_ascii_strdown (name, -1);
    
    /* find out if there is a (unitname) at the end of the argname and get rid
       of it */
    paren = g_strrstr (ret, " (");
    if (paren != NULL) {
        *paren = '\0';
    }

    g_strcanon (ret, G_CSET_a_2_z G_CSET_DIGITS "-", '-');
    
    /* chomp trailing dashes */
    for (i=strlen(ret)-1; ret[i]=='-' && i>0; i--)
        ret[i] = '\0';

    if (strcmp (ret, "amplitude") == 0) {
        g_free (ret);
        return g_strdup ("mul");
    }

    return ret;
}

static gfloat
ss_ladspa_get_default_value (LADSPA_PortRangeHint hint) 
{
    gfloat def, lowerbound, upperbound;
    gboolean toggled, logarithmic, integer, lower, upper;
    gint hintdesc;

    def = upperbound = lowerbound = 0.0f;
    toggled = logarithmic = integer = lower = upper = FALSE;
    hintdesc = hint.HintDescriptor;

#ifdef LADSPA_IS_HINT_HAS_DEFAULT
    /* get the various bits */
    if (LADSPA_IS_HINT_TOGGLED(hintdesc))
        toggled = TRUE;
    if (LADSPA_IS_HINT_LOGARITHMIC(hintdesc))
        logarithmic = TRUE;
    if (LADSPA_IS_HINT_INTEGER(hintdesc))
        integer = TRUE;

    /* grab the bounds */
    if (LADSPA_IS_HINT_BOUNDED_BELOW(hintdesc)) {
        lower = TRUE;
        lowerbound = hint.LowerBound;
    }
    
    if (LADSPA_IS_HINT_BOUNDED_ABOVE(hintdesc)) {
        upper = TRUE;
        upperbound = hint.UpperBound;
    }

    def = lowerbound;

    if (LADSPA_IS_HINT_HAS_DEFAULT (hintdesc)) {
        if (LADSPA_IS_HINT_DEFAULT_MINIMUM (hintdesc))
            def = lowerbound;
        else if (LADSPA_IS_HINT_DEFAULT_LOW (hintdesc))
            if (LADSPA_IS_HINT_LOGARITHMIC (hintdesc))
                def = exp (0.75*log(lowerbound) +
                           0.25*log(upperbound));
            else
                def = (0.75*lowerbound +
                       0.25*upperbound);
        else if (LADSPA_IS_HINT_DEFAULT_MIDDLE (hintdesc))
            if (LADSPA_IS_HINT_LOGARITHMIC (hintdesc))
                def = exp (0.5*log(lowerbound) +
                           0.5*log(upperbound));
            else
                def = (0.5*lowerbound +
                       0.5*upperbound);
        else if (LADSPA_IS_HINT_DEFAULT_HIGH (hintdesc))
            if (LADSPA_IS_HINT_LOGARITHMIC (hintdesc))
                def = exp (0.25*log(lowerbound) +
                           0.75*log(upperbound));
            else
                def = (0.25*lowerbound +
                       0.75*upperbound);
        else if (LADSPA_IS_HINT_DEFAULT_MAXIMUM (hintdesc))
            def = upperbound;
        else if (LADSPA_IS_HINT_DEFAULT_0 (hintdesc))
            def = 0.0;
        else if (LADSPA_IS_HINT_DEFAULT_1 (hintdesc))
            def = 1.0;
        else if (LADSPA_IS_HINT_DEFAULT_100 (hintdesc))
            def = 100.0;
        else if (LADSPA_IS_HINT_DEFAULT_440 (hintdesc))
            def = 440.0;
    }
#endif /* LADSPA_IS_HINT_HAS_DEFAULT */

    return def;
}

static void
ladspa_describe_plugin(const char *pcFullFilename,
                       void *pvPluginHandle,
                       LADSPA_Descriptor_Function pfDescriptorFunction)
{
    const LADSPA_Descriptor *desc;
    int i,j;
  
    GstElementDetails *details;
    GTypeInfo typeinfo = {
        sizeof(SSLADSPAClass),
        NULL,
        NULL,
        (GClassInitFunc)ss_ladspa_class_init,
        NULL,
        NULL,
        sizeof(SSLADSPA),
        0,
        (GInstanceInitFunc)ss_ladspa_init,
    };
    GType type;

    /* walk through all the plugins in this pluginlibrary */
    i = 0;
    while ((desc = pfDescriptorFunction(i++))) {
        gchar *type_name, *pad_name;
        gboolean provides_mul = FALSE, provides_add = FALSE;
        GList *l;
        GstElementClass *class;

        /* construct the type */
        type_name = g_strdup_printf("ssl-%s",desc->Label);
        g_strcanon (type_name, G_CSET_A_2_Z G_CSET_a_2_z G_CSET_DIGITS "-+", '-');
        if (g_type_from_name(type_name)) {
            /* for some reason the g_warning here will screw things up. besides,
               the user doesn't care if she has two copies of a plugin. */
            /* g_warning ("Type %s already registered", type_name); */
            g_free(type_name);
            continue;
        }
        /* create the type now */
        type = g_type_register_static(SS_TYPE_ELEMENT, type_name, &typeinfo, 0);

        /* add this plugin to the hash */
        g_hash_table_insert(ladspa_descriptors,
                            GINT_TO_POINTER(type),
                            (gpointer)desc);
    
        class = g_type_class_ref (type);

        /* construct the element details struct */
        details = g_new0(GstElementDetails,1);
        details->longname = g_strdup(desc->Name);
        details->klass = "Soundscrape/LADSPA";
        /* where to put this? details->license = g_strdup (desc->Copyright);*/
        details->description = g_strdup (details->longname);
        details->author = g_strdup(desc->Maker);
        gst_element_class_set_details (GST_ELEMENT_CLASS (class), details);

        for (j=0;j<desc->PortCount;j++) {
            pad_name = ss_ladspa_sanitize_name (desc->PortNames[j]);

            if (LADSPA_IS_PORT_INPUT(desc->PortDescriptors[j])) {
                float def = ss_ladspa_get_default_value (desc->PortRangeHints[j]);
                gst_element_class_add_pad_template (class,
                                                    ladspa_sink_factory (pad_name, def));
            }
            
            if (LADSPA_IS_PORT_OUTPUT(desc->PortDescriptors[j]))
                gst_element_class_add_pad_template (class,
                                                    ladspa_src_factory (pad_name));
        }
        
        ss_ladspa_class_uniquify_pad_names (class);

        /* FIXME: a provided add or mul should be moved to the correct place */
        for (l=class->padtemplates; l; l=l->next)
            if (strcmp (GST_PAD_TEMPLATE_NAME_TEMPLATE (l->data), "mul") == 0)
                provides_mul=TRUE;
        for (l=class->padtemplates; l; l=l->next)
            if (strcmp (GST_PAD_TEMPLATE_NAME_TEMPLATE (l->data), "add") == 0)
                provides_add=TRUE;

        /* should only add these if there is an output, but oh well */
        if (!provides_mul)
            gst_element_class_add_pad_template (class, ladspa_sink_factory (g_strdup ("mul"), 1.0));
        if (!provides_add)
            gst_element_class_add_pad_template (class, ladspa_sink_factory (g_strdup ("add"), 0.0));

        /* register the plugin with gstreamer, trim "ssl-" */
        if (!gst_element_register (ss_plugin, type_name+4, GST_RANK_NONE, type))
            g_warning ("Generator %s failed to register", type_name+4);
    }
}

gboolean
init_ss_ladspa (GstPlugin *plugin)
{
    GST_DEBUG_CATEGORY_INIT (ladspa_debug, "ss-ladspa",
                             GST_DEBUG_FG_WHITE | GST_DEBUG_BG_GREEN,
                             "Soundscrape LADSPA");

    ladspa_descriptors = g_hash_table_new(NULL,NULL);
    parent_class = g_type_class_ref(SS_TYPE_ELEMENT);

    LADSPAPluginSearch(ladspa_describe_plugin);

    return TRUE;
}
