/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-multi-op.c: Commutative operations taking a variable number of arguments
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Some code based on gstadder.c from gst-plugins:
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *               2000 Wim Taymans <wim.taymans@chello.be>
 *               2001 Thomas <thomas@apestaart.org>
 */

#include <config.h>
#include "ss-multi-op.h"

/* declare the guts of the operations here in the beginning */
static gfloat do_plus_cc (gfloat existing, gfloat new)
{
    return existing + new;
}
static void do_plus_aa (gfloat *existing, gfloat *new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] += new[i];
}
static void do_plus_ac (gfloat *existing, gfloat new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] += new;
}
static gfloat do_times_cc (gfloat existing, gfloat new)
{
    return existing * new;
}
static void do_times_aa (gfloat *existing, gfloat *new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] *= new[i];
}
static void do_times_ac (gfloat *existing, gfloat new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] *= new;
}
static gfloat do_min_cc (gfloat existing, gfloat new)
{
    return existing < new ? existing : new;
}
static void do_min_aa (gfloat *existing, gfloat *new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] = existing[i] < new[i] ? existing[i] : new[i];
}
static void do_min_ac (gfloat *existing, gfloat new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] = existing[i] < new ? existing[i] : new;
}
static gfloat do_max_cc (gfloat existing, gfloat new)
{
    return existing > new ? existing : new;
}
static void do_max_aa (gfloat *existing, gfloat *new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] = existing[i] > new[i] ? existing[i] : new[i];
}
static void do_max_ac (gfloat *existing, gfloat new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] = existing[i] > new ? existing[i] : new;
}
static gfloat do_sumsqr_c (gfloat new)
{
    return new * new;
}
static void do_sumsqr_a (gfloat *existing, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] *= existing[i];
}
static gfloat do_sumsqr_cc (gfloat existing, gfloat new)
{
    return existing + new * new;
}
static void do_sumsqr_aa (gfloat *existing, gfloat *new, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        existing[i] += new[i] * new[i];
}
static void do_sumsqr_ac (gfloat *existing, gfloat new, gint how_many) 
{
    gint i;
    gfloat newsqr = new * new;
    for (i=0; i<how_many; i++)
        existing[i] += newsqr;
}

typedef struct {
    gchar *factory_name;
    gfloat (*do_op_c) (gfloat new);
    void (*do_op_a) (gfloat *existing, gint how_many);
    gfloat (*do_op_cc) (gfloat existing, gfloat new);
    void (*do_op_aa) (gfloat *existing, gfloat *new, gint how_many);
    void (*do_op_ac) (gfloat *existing, gfloat new, gint how_many);
    gchar *description;
    GType type;
} SSMultiOpDetails;

static SSMultiOpDetails details[] = {
    {"+", NULL, NULL, do_plus_cc, do_plus_aa, do_plus_ac, "Addition", 0},
    {"*", NULL, NULL, do_times_cc, do_times_aa, do_times_ac, "Multiplication", 0},
    {"min", NULL, NULL, do_min_cc, do_min_aa, do_min_ac, "Minimum", 0},
    {"max", NULL, NULL, do_max_cc, do_max_aa, do_max_ac, "Maximum", 0},
    {"sumsqr", do_sumsqr_c, do_sumsqr_a, do_sumsqr_cc, do_sumsqr_aa, do_sumsqr_ac, "SumSquares", 0},
    {NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0}
};

/* then the generic support */
SS_PAD_TEMPLATE_SRC_FACTORY (ss_multi_op_src_factory, "src");

static GstPadTemplate*
ss_multi_op_sink_factory (gfloat default_value) 
{
    static GstPadTemplate *templ;
  
    if (!templ) {
        templ = ss_pad_template_sink_new ("sink%d", default_value);
        GST_PAD_TEMPLATE_PRESENCE (templ) = GST_PAD_REQUEST;
    }
  
    return templ;
}


static void 		ss_multi_op_class_init		(SSMultiOpClass *klass);
static void 		ss_multi_op_init		(SSMultiOp *multi_op);

static GstPad* 		ss_multi_op_request_new_pad 	(GstElement *element, GstPadTemplate *temp,
                                                         const gchar *unused);

static void 		ss_multi_op_loop  		(GstElement *element);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_multi_op_get_type, SS_TYPE_ELEMENT,
                          "SSMultiOp", SSMultiOpClass, SSMultiOp,
                          NULL /* abstract type */, NULL /* abstract type */);

static void
ss_multi_op_class_init (SSMultiOpClass *klass) 
{
    GstElementClass *gstelement_class = (GstElementClass *) klass;

    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);

    gstelement_class->request_new_pad = ss_multi_op_request_new_pad;
}

static void 
ss_multi_op_init (SSMultiOp *this) 
{
    SSMultiOpDetails *d;

    for (d=details; d; d++) {
        if (G_TYPE_FROM_INSTANCE (this) == d->type) {
            this->do_op_c = d->do_op_c;
            this->do_op_a = d->do_op_a;
            this->do_op_cc = d->do_op_cc;
            this->do_op_aa = d->do_op_aa;
            this->do_op_ac = d->do_op_ac;
            break;
        }
    }

    this->srcpad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_multi_op_src_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->srcpad);

    gst_element_set_loop_function (GST_ELEMENT (this), ss_multi_op_loop);
}

static GstPad*
ss_multi_op_request_new_pad (GstElement *element, GstPadTemplate *templ, 
                             const gchar *unused) 
{
    static gint count = 0;
    SSMultiOp *this;
    GstPad *pad;
    gchar *name;

    this = SS_MULTI_OP (element);

    pad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_multi_op_sink_factory (0.0)));
    name = g_strdup_printf ("sink%d", count++);
    gst_object_set_name (GST_OBJECT (pad), name);
    g_free (name);
    
    gst_element_add_pad (GST_ELEMENT (this), pad);

    /* add the input_channel to the list of input channels */
  
    this->sinkpads = g_list_append (this->sinkpads, pad);
    this->numsinkpads++;
  
    return GST_PAD (pad);
}

static void
ss_multi_op_loop (GstElement *element)
{
    SSMultiOp *this;
    gint buffer_frames;
    gint num_to_write;
    GstData *buf_out = NULL;
    GstData *buf_in = NULL;
    gfloat *data_out, *data_in;
    GList *l;
    GstPad *pad;
    gfloat control_value;
    gboolean control_set, eos;
    SSPad *spad;

    this = (SSMultiOp*)element;

    if (g_list_length (this->sinkpads) < 2) {
        gst_element_error (element,
                           "You must connect at least two pads to multiop elements.");
        return;
    }

    num_to_write = buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;

    if (!buffer_frames) {
        gst_element_error (element, "Caps were never set, bailing...");
        return;
    }

    control_value = 0.f;
    control_set = FALSE;
    eos = FALSE;

    for (l=this->sinkpads; l; l=l->next) {
        pad = GST_PAD_CAST (l->data);
        spad = (SSPad*)pad;
      
        if (GST_PAD_IS_USABLE (pad))
            buf_in = gst_pad_pull (pad);
        else
            buf_in = NULL;

        if (buf_in && GST_IS_EVENT (buf_in)) {
            /* stupidly assuming it's an eos event */
            num_to_write = 0;
            eos = TRUE;
            gst_data_unref (buf_in);
            break;
        } else if (!buf_in || SS_IS_CONTROL (buf_in)) {
            /* a control input */
            if (buf_in) {
                spad->value = SS_CONTROL_VALUE (buf_in);
                gst_data_unref (buf_in);
            }
            
            if (!control_set) {
                control_value = this->do_op_c ? this->do_op_c (spad->value) : spad->value;
                control_set = TRUE;
            } else {
                control_value = this->do_op_cc (control_value, spad->value);
            }
        } else {
            /* an audio input, we have a buffer */
            num_to_write = MIN (num_to_write, SS_BUFFER_GET_FRAMES (buf_in));

            if (!buf_out) {
                /* reuse the new input buffer */
                buf_out = buf_in;
                data_out = data_in = SS_BUFFER_DATA (buf_in);
                
                /* if, like sumsqr, we have to do something to the first buffer,
                   do it now. */
                if (this->do_op_a)
                    this->do_op_a (data_out, num_to_write);
            } else {
                /* all subsequent pads apply a binary operation between the
                 * existing value and the new buffer. */
                data_in = SS_BUFFER_DATA (buf_in);
                this->do_op_aa (data_out, data_in, num_to_write);
                gst_data_unref (buf_in);
            }
        }
    }

    if (eos) {
        if (buf_out)
            gst_data_unref (buf_out);
        gst_pad_push (this->srcpad, GST_DATA (gst_event_new (GST_EVENT_EOS)));
        gst_element_set_eos (element);
    } else if (buf_out) {
        /* audio rate */
        SS_BUFFER_SET_FRAMES (buf_out, num_to_write);
        /* operate on the control inputs at the last minute */
        if (control_set)
            this->do_op_ac (data_out, control_value, num_to_write);
        ss_pad_push (SS_PAD (this->srcpad), buf_out);
    } else if (control_set) {
        /* control rate */
        /* fixme: should be able to reuse a buffer */
        GstData *control = ss_control_new ((SSPad*)this->srcpad);
        SS_CONTROL_VALUE (control) = control_value;
        ss_pad_push ((SSPad*)this->srcpad, control);
    } else {
        /* error, we have no inputs -- shouldn't happen */
        gst_element_error (element, "MultiOp ugen has no inputs!");
    }
}

gboolean
init_ss_multi_op (GstPlugin *plugin)
{
    SSMultiOpDetails *d;
    GstElementClass *class;

    for (d=details; d->factory_name!=NULL; d++) {
        GstElementDetails *edetails;
        static const GTypeInfo type_info = {
            sizeof (SSMultiOpClass), NULL, NULL,
            (GClassInitFunc) ss_multi_op_class_init, NULL, NULL,
            sizeof (SSMultiOp), 0,
            (GInstanceInitFunc) ss_multi_op_init,
        };

        edetails = g_new0 (GstElementDetails, 1);
        edetails->longname = d->factory_name;
        edetails->klass = "Soundscrape/Operators/Multi";
        edetails->description = d->description;
        edetails->author = "Andy Wingo <wingo at pobox dot com>";

        d->type = g_type_register_static (SS_TYPE_MULTI_OP,
                                          g_strdup_printf ("SSMultiOp%s", d->description),
                                          &type_info, 0);
        class = g_type_class_ref (d->type);

        gst_element_class_set_details (class, edetails);
        gst_element_class_add_pad_template (class, ss_multi_op_sink_factory (0.0));
        gst_element_class_add_pad_template (class, ss_multi_op_src_factory ());
        
        gst_element_register (plugin, d->factory_name, GST_RANK_NONE, d->type);
    }
    
    return TRUE;
}

