/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-pad.c: Subclass of GstRealPad supporting default values
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 */

#include <gst/gstpad.h>

#include "ss-control.h"
#include "ss-pad.h"

#define LIN_INTERPOLATE(val1, valdiff, i, frames)                               \
    (val1 + (1.0 - ((frames - i - 1.0) / (frames - 1.0))) * valdiff);

/* TODO: detect the need for interpolation using discontinuous events? */


static void	ss_pad_class_init		(SSPadClass *klass);
static void	ss_pad_init			(SSPad *pad);
static void	ss_pad_template_class_init	(SSPadTemplateClass *klass);
static void	ss_pad_template_init		(SSPadTemplate *pad);

static void	ss_pad_set_property		(GObject *object, guint prop_id,
                                                 const GValue *value, 
						 GParamSpec *pspec);
static void	ss_pad_get_property		(GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);
static void	ss_pad_template_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
						 GParamSpec *pspec);
static void	ss_pad_template_get_property	(GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);
static GstPadLinkReturn ss_pad_default_link	(GstPad *pad, GstCaps *caps);


static GstPadClass *pad_parent_class = NULL;
static GstPadTemplateClass *template_parent_class = NULL;

enum {
    PAD_PROP_0,
    PAD_PROP_VALUE,
    /* FILL ME */
};

enum {
    TEMPLATE_PROP_0,
    TEMPLATE_PROP_DEFAULT_VALUE,
    /* FILL ME */
};


GST_CAPS_FACTORY (ss_caps_factory,
                  GST_CAPS_NEW ("ss-caps",
                                "audio/x-raw-float",
                                "rate",		GST_PROPS_INT_RANGE (4000, 96000),
                                "endianness",	GST_PROPS_INT (G_BYTE_ORDER),
                                "width",     	GST_PROPS_INT (32),
                                "buffer-frames",GST_PROPS_INT_RANGE (1, G_MAXINT),
                                "channels",	GST_PROPS_INT (1)));


SS_TYPE_MAKE_BOILERPLATE (ss_pad_get_type, GST_TYPE_REAL_PAD,
                          "SSPad", SSPadClass, SSPad,
                          ss_pad_class_init, ss_pad_init);

SS_TYPE_MAKE_BOILERPLATE (ss_pad_template_get_type, GST_TYPE_PAD_TEMPLATE,
                          "SSPadTemplate", SSPadTemplateClass, SSPadTemplate,
                          ss_pad_template_class_init, ss_pad_template_init);

static void
ss_pad_class_init (SSPadClass *klass)
{
    GObjectClass *gobject_class;
    GParamSpec *pspec;

    gobject_class = (GObjectClass*)klass;

    pad_parent_class = g_type_class_ref (GST_TYPE_REAL_PAD);

    gobject_class->set_property  = GST_DEBUG_FUNCPTR (ss_pad_set_property);
    gobject_class->get_property  = GST_DEBUG_FUNCPTR (ss_pad_get_property);

    pspec = g_param_spec_float ("value", "Value", "Value of pad if not connected",
                                -G_MAXFLOAT, G_MAXFLOAT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PAD_PROP_VALUE, pspec);
}

static void
ss_pad_init (SSPad *pad)
{
}

static void
ss_pad_template_class_init (SSPadTemplateClass *klass)
{
    GObjectClass *gobject_class;
    GParamSpec *pspec;

    gobject_class = (GObjectClass*)klass;

    template_parent_class = g_type_class_ref (GST_TYPE_PAD_TEMPLATE);

    gobject_class->set_property  = GST_DEBUG_FUNCPTR (ss_pad_template_set_property);
    gobject_class->get_property  = GST_DEBUG_FUNCPTR (ss_pad_template_get_property);

    pspec = g_param_spec_float ("default-value", "Default value",
                                "Default value of pad if not connected",
                                -G_MAXFLOAT, G_MAXFLOAT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     TEMPLATE_PROP_DEFAULT_VALUE, pspec);
}

static void
ss_pad_template_init (SSPadTemplate *pad)
{
}

/* keep in mind the return value might be an event */
GstData*
ss_pad_pull (SSPad *spad, gint buffer_frames)
{
    GstData *ret = NULL;
    gfloat oldvalue = SS_PAD_VALUE (spad);

    if (GST_PAD_PEER (spad)) {
        ret = gst_pad_pull (GST_PAD_CAST (spad));

        /* we set spad->value even for buffers just to be consistent -- not sure
           if this is useful though */
        if (SS_IS_CONTROL (ret)) {
            SS_PAD_VALUE (spad) = SS_CONTROL_VALUE (ret);
            gst_data_unref (ret);
            ret = NULL;
        } else if (GST_IS_BUFFER (ret)) {
            SS_PAD_VALUE (spad) = SS_BUFFER_DATA (ret)[0];
            return ret;
        } else {
            /* an event */
            return ret;
        }
    }

    /* don't interpolate on the first buffer */
    oldvalue = spad->need_interp ? oldvalue : SS_PAD_VALUE (spad);
    spad->need_interp = TRUE;
    return ss_buffer_new_lin_interp (oldvalue, SS_PAD_VALUE (spad));
}

GstData*
ss_pad_update_control (SSPad *spad)
{
    GstData *ret = NULL;

    if (GST_PAD_PEER (spad)) {
        ret = GST_DATA (gst_pad_pull (GST_PAD_CAST (spad)));

        if (SS_IS_CONTROL (ret)) {
            SS_PAD_VALUE (spad) = SS_CONTROL_VALUE (ret);
            /* don't null the buffer, we might be able to reuse it */
        } else if (GST_IS_BUFFER (ret)) {
            SS_PAD_VALUE (spad) = SS_BUFFER_DATA (ret)[0];
            gst_data_unref (ret);
            ret = NULL;
        } else {
            /* if it's an event, do nothing -- for now at least */
            gst_data_unref (ret);
            ret = NULL;
        }
    }

    return ret;
}

void
ss_pad_push (SSPad *spad, GstData *data) 
{
    /* set the pad's value so that we can ss_pad_poll () the pad */
    if (SS_IS_CONTROL (data))
        SS_PAD_VALUE (spad) = SS_CONTROL_VALUE (data);
    else if (GST_IS_BUFFER (data))
        SS_PAD_VALUE (spad) = SS_BUFFER_DATA (data)[SS_BUFFER_GET_FRAMES (data)-1];

    gst_pad_push (GST_PAD_CAST (spad), data);
}

#define SS_PAD_DO_OP(name, opequals, noop_value)                                \
void                                                                            \
ss_pad_do_##name (SSPad *spad, GstData **out, gint n)                           \
{                                                                               \
    gfloat oldvalue = SS_PAD_VALUE (spad);                                      \
    gfloat newvalue = 0, valuediff = 0;                                         \
                                                                                \
    if (GST_PAD_PEER (spad)) {                                                  \
        GstData *data = GST_DATA (gst_pad_pull (GST_PAD_CAST (spad)));          \
                                                                                \
        /* set SS_PAD_VALUE (spad) in all cases */                              \
        if (SS_IS_CONTROL (data)) {                                             \
            SS_PAD_VALUE (spad) = SS_CONTROL_VALUE (data);                      \
            gst_data_unref (data);                                              \
        } else if (GST_IS_BUFFER (data)) {                                      \
            gint i, j, frames;                                                  \
            gfloat *outd = NULL, *values = SS_BUFFER_DATA (data);               \
                                                                                \
            SS_PAD_VALUE (spad) = values[0];                                    \
            for (i=0; i<n; i++) {                                               \
                if (SS_IS_CONTROL (out[i])) {                                   \
                    /* undersample */                                           \
                    SS_CONTROL_VALUE (out[i]) opequals values[0];               \
                } else {                                                        \
                    frames = SS_BUFFER_GET_FRAMES (out[i]);                     \
                    outd = SS_BUFFER_DATA (out[i]);                             \
                    g_return_if_fail (SS_BUFFER_GET_FRAMES (out[i]) == frames); \
                    for (j=0; j<frames; j++)                                    \
                        outd[j] opequals values[j];                             \
                }                                                               \
            }                                                                   \
            gst_data_unref (data);                                              \
            return;                                                             \
        } else {                                                                \
            /* might be an event -- let's print out something for debugging     \
             * purposes */                                                      \
            g_print ("got an event in ss_pad_do_op on pad %s:%s\n",             \
                     GST_DEBUG_PAD_NAME (spad));                                \
            gst_data_unref (data);                                              \
        }                                                                       \
    }                                                                           \
                                                                                \
    /* don't interpolate on the first buffer */                                 \
    newvalue = SS_PAD_VALUE (spad);                                             \
    valuediff = spad->need_interp ? newvalue - oldvalue : 0.f;                  \
    spad->need_interp = TRUE;                                                   \
    if (newvalue != noop_value || valuediff != 0.f) {                           \
        gfloat *data, slope;                                                    \
        gint i, j, frames;                                                      \
                                                                                \
        for (i=0; i<n; i++) {                                                   \
            if (SS_IS_CONTROL (out[i])) {                                       \
                SS_CONTROL_VALUE (out[i]) opequals oldvalue + valuediff;        \
            } else {                                                            \
                frames = SS_BUFFER_GET_FRAMES (out[i]);                         \
                data = SS_BUFFER_DATA (out[i]);                                 \
                slope = valuediff / frames;                                     \
                for (j=0; j<frames; j++) {                                      \
                    data[j] opequals oldvalue;                                  \
                    oldvalue += slope;                                          \
                }                                                               \
            }                                                                   \
        }                                                                       \
    }                                                                           \
}

SS_PAD_DO_OP (add, +=, 0.f);
SS_PAD_DO_OP (mul, *=, 1.f);

static void
ss_pad_set_property (GObject *object, guint prop_id, 
                     const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
    case PAD_PROP_VALUE:
        SS_PAD_VALUE (object) = g_value_get_float (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
ss_pad_get_property (GObject *object, guint prop_id, 
                     GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
    case PAD_PROP_VALUE:
        g_value_set_float (value, SS_PAD_VALUE (object));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
ss_pad_template_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
    case TEMPLATE_PROP_DEFAULT_VALUE:
        SS_PAD_TEMPLATE (object)->default_value = g_value_get_float (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
ss_pad_template_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
    case TEMPLATE_PROP_DEFAULT_VALUE:
        g_value_set_float (value, SS_PAD_TEMPLATE (object)->default_value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

GstPad*
ss_pad_new_from_template (SSPadTemplate *templ)
{
    SSPad *pad;

    g_return_val_if_fail (templ != NULL, NULL);

    pad = g_object_new (ss_pad_get_type (), NULL);
    gst_object_set_name (GST_OBJECT (pad), GST_PAD_TEMPLATE_NAME_TEMPLATE (templ));
    GST_RPAD_DIRECTION (GST_REAL_PAD (pad)) = GST_PAD_TEMPLATE (templ)->direction;
  
    gst_object_ref (GST_OBJECT (templ));
    GST_PAD_PAD_TEMPLATE (pad) = GST_PAD_TEMPLATE (templ);

    pad->value = templ->default_value;

    gst_pad_set_link_function (GST_PAD_CAST (pad), ss_pad_default_link);

    return GST_PAD (pad);
}

GstPadTemplate*
ss_pad_template_src_new (const gchar *name)
{
    SSPadTemplate *new;
    GstCaps *thecaps = ss_caps_factory ();

    g_return_val_if_fail (name != NULL, NULL);

    new = g_object_new (ss_pad_template_get_type (),
                        "name", name,
                        NULL);

    GST_PAD_TEMPLATE_NAME_TEMPLATE (new) = g_strdup (name);
    GST_PAD_TEMPLATE_DIRECTION (new) = GST_PAD_SRC;
    GST_PAD_TEMPLATE_PRESENCE (new) = GST_PAD_ALWAYS;

    GST_PAD_TEMPLATE_CAPS (new) = thecaps;
    gst_caps_ref (thecaps);
    gst_caps_sink (thecaps);

    return GST_PAD_TEMPLATE (new);
}

GstPadTemplate*
ss_pad_template_sink_new (const gchar *name, gfloat default_value)
{
    SSPadTemplate *new;
    GstCaps *thecaps = ss_caps_factory ();

    g_return_val_if_fail (name != NULL, NULL);

    new = g_object_new (ss_pad_template_get_type (),
                        "name", name,
                        NULL);

    GST_PAD_TEMPLATE_NAME_TEMPLATE (new) = g_strdup (name);
    GST_PAD_TEMPLATE_DIRECTION (new) = GST_PAD_SINK;
    GST_PAD_TEMPLATE_PRESENCE (new) = GST_PAD_ALWAYS;

    GST_PAD_TEMPLATE_CAPS (new) = thecaps;
    gst_caps_ref (thecaps);
    gst_caps_sink (thecaps);

    new->default_value = default_value;

    return GST_PAD_TEMPLATE (new);
}

static GstPadLinkReturn
ss_pad_default_link (GstPad *pad, GstCaps *caps)
{
    SSElement *this;
    const GList *pads;
    GstPad *p;
  
    this = SS_ELEMENT (GST_PAD_PARENT (pad));

    if (GST_CAPS_IS_FIXED (caps)) {
        gst_caps_get_int (caps, "rate", &this->rate);
        gst_caps_get_int (caps, "buffer-frames", &this->buffer_frames);

        for (pads=GST_ELEMENT_CAST (this)->pads; pads; pads=pads->next) {
            p = (GstPad*)pads->data;
            if (p != pad) {
                if (gst_pad_try_set_caps (p, caps) <= 0) {
                    GST_DEBUG ("Sorry kiddo, %s:%s refused caps", GST_DEBUG_PAD_NAME (p));
                    return GST_PAD_LINK_REFUSED;
                }
            }
        }

        ss_buffer_pool_set_buffer_frames (this->buffer_frames);

        return GST_PAD_LINK_OK;
    } else {
        return GST_PAD_LINK_DELAYED;
    }
}

