/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-plug.c: Receive data from external sources
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 */


#include <math.h>

#include "ss-plug.h"
#include <config.h>

static GstElementDetails plug_details = {
  "Plug",
  "Soundscrape/Controls",
  "Soundscrape plug to receive data from external sources",
  "Andy Wingo <wingo at pobox dot com>",
};

enum {
    PROP_0,
    PROP_VALUE,
    PROP_LAG_TIME
};

SS_PAD_TEMPLATE_SRC_FACTORY (plug_src_factory, "src");

static void		ss_plug_class_init	(SSPlugClass *klass);
static void		ss_plug_init		(SSPlug *this);

static void		ss_plug_set_property	(GObject *object, guint prop_id, 
                                                 const GValue *value, GParamSpec *pspec);
static void		ss_plug_get_property	(GObject *object, guint prop_id, 
                                                 GValue *value, GParamSpec *pspec);

static GstData*		ss_plug_get		(GstPad *pad);


static GstElementClass *parent_class = NULL;

GST_DEBUG_CATEGORY_STATIC (ssplug_debug);
#define INFO_OBJ(obj,...) \
    GST_CAT_LEVEL_LOG (ssplug_debug, GST_LEVEL_INFO, obj, __VA_ARGS__)

SS_TYPE_MAKE_BOILERPLATE (ss_plug_get_type, SS_TYPE_ELEMENT,
                          "SSPlug", SSPlugClass, SSPlug,
                          ss_plug_class_init, ss_plug_init);

static void
ss_plug_class_init (SSPlugClass *klass) 
{
    GObjectClass *gobject_class;
    GParamSpec *pspec;

    gobject_class = (GObjectClass*)klass;

    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);

    gobject_class->set_property = ss_plug_set_property;
    gobject_class->get_property = ss_plug_get_property;

    pspec = g_param_spec_float ("value", "Value", "Value of this",
                                -G_MAXFLOAT, G_MAXFLOAT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_VALUE, pspec);
    pspec = g_param_spec_float ("lag-time", "Lag time", "Lag time for input smoothing",
                                0, 1, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_LAG_TIME, pspec);

    gst_element_class_add_pad_template (GST_ELEMENT_CLASS (klass), plug_src_factory ());
    gst_element_class_set_details (GST_ELEMENT_CLASS (klass), &plug_details);
}

static void 
ss_plug_init (SSPlug *this) 
{
    this->value = 0;
    this->lag_time = 0;
    this->last_value = 0;
    this->multiplier = 0;

    this->src = ss_pad_new_from_template ((SSPadTemplate*)plug_src_factory ());
    gst_pad_set_get_function (this->src, ss_plug_get);
    gst_element_add_pad (GST_ELEMENT (this), this->src);
}

static void
ss_plug_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
    SSPlug *this;

    this = SS_PLUG (object);

    switch (prop_id) {
    case PROP_VALUE:
        this->value = g_value_get_float (value);
        INFO_OBJ (this, "value set to %f", this->value);
        break;
    case PROP_LAG_TIME:
        this->lag_time = g_value_get_float (value);
        INFO_OBJ (this, "lag time set to %f", this->value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void 
ss_plug_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
    SSPlug *this;
   
    this = SS_PLUG (object);
   
    switch (prop_id) {
    case PROP_VALUE:
        g_value_set_float (value, this->value);
        break;
    case PROP_LAG_TIME:
        g_value_set_float (value, this->lag_time);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static GstData*
ss_plug_get (GstPad *pad)
{
    SSPlug *this;
    GstData *control;

    this = SS_PLUG (GST_OBJECT_PARENT (pad));

    control = ss_control_new ((SSPad*)pad);
    /* for the moment, no smoothing */
    SS_CONTROL_VALUE (control) = this->value;

    return control;
}

gboolean 
init_ss_plug (GstPlugin *plugin) 
{
    GST_DEBUG_CATEGORY_INIT (ssplug_debug, "ss-plug",
                             GST_DEBUG_FG_BLACK | GST_DEBUG_BG_CYAN | GST_DEBUG_BOLD,
                             "Soundscrape plug debug");

    return gst_element_register (plugin, "ss-plug", GST_RANK_NONE, SS_TYPE_PLUG);
}
