/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-sequencer.c: Sequenced, triggered control data
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 */

#include <config.h>
#include "ss-sequencer.h"
#include <guile-gnome-gobject-primitives.h>


static GstElementDetails sequencer_details = {
  "Sequencer",
  "Soundscrape/Controls",
  "Sequenced, triggered control data",
  "Andy Wingo <wingo at pobox dot com>",
};

enum {
    PROP_0,
    PROP_LAMBDA
};

SS_PAD_TEMPLATE_SRC_FACTORY (ss_sequencer_src_factory, "output");
SS_PAD_TEMPLATE_SINK_FACTORY (ss_sequencer_sink_factory, "trigger", 0.f);


static void 	ss_sequencer_class_init		(SSSequencerClass *klass);
static void 	ss_sequencer_init		(SSSequencer *sequencer);

static void	ss_sequencer_set_property	(GObject *object, guint prop_id, 
                                                 const GValue *value, GParamSpec *pspec);
static void	ss_sequencer_get_property	(GObject *object, guint prop_id, 
                                                 GValue *value, GParamSpec *pspec);

static void 	ss_sequencer_loop  		(GstElement *element);

static GstElementStateReturn ss_sequencer_change_state	(GstElement *element);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_sequencer_get_type, SS_TYPE_ELEMENT,
                          "SSSequencer", SSSequencerClass, SSSequencer,
                          ss_sequencer_class_init, ss_sequencer_init);

static void
ss_sequencer_class_init (SSSequencerClass *klass) 
{
    GParamSpec *pspec;
    GObjectClass *gobject_class = (GObjectClass*)klass;
    GstElementClass *eclass = (GstElementClass*)klass;
    
    gobject_class->set_property = ss_sequencer_set_property;
    gobject_class->get_property = ss_sequencer_get_property;

    pspec = g_param_spec_boxed ("lambda", "lambda", "Scheme procedure to get new control value",
                                G_TYPE_BOXED_SCM, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_LAMBDA, pspec);

    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);

    eclass->change_state = ss_sequencer_change_state;

    gst_element_class_add_pad_template (eclass, ss_sequencer_src_factory ());
    gst_element_class_add_pad_template (eclass, ss_sequencer_sink_factory ());
    gst_element_class_set_details (eclass, &sequencer_details);
}

static void 
ss_sequencer_init (SSSequencer *this) 
{
    GstElement *element = GST_ELEMENT (this);

    this->lambda = SCM_BOOL_F;

    this->src = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_sequencer_src_factory ()));
    gst_element_add_pad (element, this->src);

    this->trigger = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_sequencer_sink_factory ()));
    gst_element_add_pad (element, this->trigger);

    gst_element_set_loop_function (element, ss_sequencer_loop);
}

static void call_lambda (SSSequencer *this) 
{
    SS_PAD_VALUE (this->src) = scm_num2float (scm_call_0 (this->lambda),
                                              0, "ss-sequencer-callback");
}

static void
ss_sequencer_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
    SSSequencer *this;

    this = SS_SEQUENCER (object);

    switch (prop_id) {
    case PROP_LAMBDA:
        if (SCM_NFALSEP (this->lambda))
            g_boxed_free (G_TYPE_BOXED_SCM, SCM_PACK (this->lambda));
        this->lambda = SCM_PACK (g_value_dup_boxed (value));
/*
        irritatingly enough, we can't call the lambda here, in the case that it
        captures the continuation -- that makes it return multiple times from
        this function, which pisses off the gobject property system. hmm.
        if (SCM_NFALSEP (this->lambda))
          call_lambda (this); */
        
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void 
ss_sequencer_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
    SSSequencer *this;
   
    this = SS_SEQUENCER (object);
   
    switch (prop_id) {
    case PROP_LAMBDA:
        g_value_set_boxed (value, SCM_PACK (this->lambda));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
ss_sequencer_loop (GstElement *element)
{
    SSSequencer *this;
    gint buffer_frames;
    GstData *buf_in = NULL;
    gfloat trigger_val;

    this = (SSSequencer*)element;
    trigger_val = SS_PAD_VALUE (this->trigger);

    buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;

    if (SCM_FALSEP (this->lambda)) {
        gst_element_error (element, "Sequencer::lambda was never set");
        return;
    }
    if (!buffer_frames) {
        gst_element_error (element, "Caps were never set, bailing...");
        return;
    }
    if (this->need_call) {
        if (SCM_NFALSEP (this->lambda)) {
            /* if we have a resettable routine, we have to reset it from within
             * the dynamic context of the iteration due to the effect of
             * continuation capturing */
            if (SCM_CLOSUREP (this->lambda) && !SCM_NULLP (SCM_CLOSURE_FORMALS (this->lambda)))
                scm_call_1 (this->lambda, SCM_BOOL_F);
            call_lambda (this);
        }
        this->need_call = FALSE;
    }

    if (GST_PAD_IS_USABLE (this->trigger)) {
        buf_in = GST_DATA (gst_pad_pull (this->trigger));
    } else {
        /* the trigger isn't connected, this is kindof pointless... */
        if (SS_ELEMENT_RATE (this) == SS_ELEMENT_RATE_CONTROL)
            ss_pad_push (SS_PAD (this->src), ss_control_new (SS_PAD (this->src)));
        else
            ss_pad_push (SS_PAD (this->src),
                         ss_buffer_new_constant_val (SS_PAD_VALUE (this->src)));
        return;
    }

    if (GST_IS_EVENT (buf_in)) {
        gst_pad_event_default (this->src, GST_EVENT (buf_in));
    } else if (SS_ELEMENT_CAST (this)->ss_rate == SS_ELEMENT_RATE_CONTROL) {
        /* control rate output */
        if (SS_IS_CONTROL (buf_in)) {
            /* control rate input */
            if (trigger_val <= 0.f && SS_CONTROL_VALUE (buf_in) > 0.f)
                call_lambda (this);
            SS_PAD_VALUE (this->trigger) = SS_CONTROL_VALUE (buf_in);

            ss_control_set_pad (buf_in, this->src);
            ss_pad_push (SS_PAD (this->src), buf_in);
        } else {
            /* audio rate trigger */
            gint i;
            gfloat *data_in = SS_BUFFER_DATA (buf_in);
            
            buffer_frames = MIN (buffer_frames, SS_BUFFER_GET_FRAMES (buf_in));
            for (i=0; i<buffer_frames; i++) {
                if (trigger_val <= 0.f && data_in[i] > 0.f)
                    call_lambda (this);
                trigger_val = data_in[i];
            }

            SS_PAD_VALUE (this->trigger) = trigger_val;
            gst_data_unref (GST_DATA (buf_in));
            ss_pad_push (SS_PAD (this->src), ss_control_new (SS_PAD (this->src)));
        }
    } else {
        /* audio rate output */
        if (SS_IS_CONTROL (buf_in)) {
            /* control rate input -- legal but wasteful... */

            if (trigger_val <= 0.f && SS_CONTROL_VALUE (buf_in) > 0.f)
                call_lambda (this);
            SS_PAD_VALUE (this->trigger) = SS_CONTROL_VALUE (buf_in);
            gst_data_unref (buf_in);

            ss_pad_push (SS_PAD (this->src),
                         ss_buffer_new_constant_val (SS_PAD_VALUE (this->src)));
        } else {
            /* audio rate input */
            gint i;
            gfloat *data = SS_BUFFER_DATA (buf_in);
            
            /* fixme: implement properly for the case where the buffer can't be
               reused */
            buffer_frames = MIN (buffer_frames, SS_BUFFER_GET_FRAMES (buf_in));
            for (i=0; i<buffer_frames; i++) {
                if (trigger_val <= 0.f && data[i] > 0.f)
                    call_lambda (this);
                trigger_val = data[i];
                data[i] = SS_PAD_VALUE (this->src);
            }

            SS_PAD_VALUE (this->trigger) = trigger_val;
            ss_pad_push (SS_PAD (this->src), buf_in);
        }
    }
}

static GstElementStateReturn
ss_sequencer_change_state (GstElement *element)
{
    SSSequencer *this = (SSSequencer*)element;

    switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_READY_TO_PAUSED:
        this->need_call = TRUE;
        break;
    case GST_STATE_PAUSED_TO_READY:
        break;
    default:
        break;
    }

    if (GST_ELEMENT_CLASS (parent_class)->change_state)
        return GST_ELEMENT_CLASS (parent_class)->change_state (element);

    return GST_STATE_SUCCESS;
}

gboolean 
init_ss_sequencer (GstPlugin *plugin) 
{
    gst_element_register (plugin, "sequencer", GST_RANK_NONE, SS_TYPE_SEQUENCER);
    return TRUE;
}
