/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-slider-control.c: Control a soundscrape parameter with a slider bar
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 */


#include <math.h>

#include "ss-slider-control.h"


enum {
    PROP_0,
    PROP_NAME,
    PROP_VALUE,
    PROP_LOWER,
    PROP_UPPER,
    PROP_WARP,
    PROP_MODEL
};

static void		ss_slider_control_class_init	(SSSliderControlClass *klass);
static void		ss_slider_control_init		(SSSliderControl *control);

static void		ss_slider_control_set_property	(GObject *object, guint prop_id, 
                                                         const GValue *value, GParamSpec *pspec);
static void		ss_slider_control_get_property	(GObject *object, guint prop_id, 
                                                         GValue *value, GParamSpec *pspec);

static void		on_adjustment_value_changed	(GtkAdjustment *adjustment,
                                                         SSSliderControl *control);
static void		model_state_change 		(GstElement *model,
                                                         GstElementState old,
                                                         GstElementState new,
                                                         SSSliderControl *control);

static GtkVBoxClass *parent_class = NULL;

SS_TYPE_MAKE_BOILERPLATE (ss_slider_control_get_type, GTK_TYPE_VBOX,
                          "SSSliderControl", SSSliderControlClass, SSSliderControl,
                          ss_slider_control_class_init, ss_slider_control_init);

static void
ss_slider_control_class_init (SSSliderControlClass *klass) 
{
    GObjectClass *gobject_class;
    GParamSpec *pspec;

    gobject_class = (GObjectClass*)klass;

    parent_class = g_type_class_ref (GTK_TYPE_VBOX);

    gobject_class->set_property = ss_slider_control_set_property;
    gobject_class->get_property = ss_slider_control_get_property;

    pspec= g_param_spec_string ("name", "Name", "The name of the control",
                                "(unnamed)", G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_NAME, pspec);
    pspec = g_param_spec_float ("lower", "Lower", "Lower bound of control",
                                -G_MAXFLOAT, G_MAXFLOAT, -1, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_LOWER, pspec);
    pspec = g_param_spec_float ("upper", "upper", "Upper bound of control",
                                -G_MAXFLOAT, G_MAXFLOAT, 1, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_UPPER, pspec);
    pspec = g_param_spec_float ("value", "Value", "Value of control",
                                -G_MAXFLOAT, G_MAXFLOAT, 0, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_VALUE, pspec);
    pspec = g_param_spec_enum ("warp", "Warp", "Warp type (linear or exponential)",
                               SS_TYPE_WARP, SS_WARP_LINEAR,
                               G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_WARP, pspec);
    pspec = g_param_spec_object ("model", "Model", "Model to control",
                                 SS_TYPE_ELEMENT, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_MODEL, pspec);
}

static void 
ss_slider_control_init (SSSliderControl *control) 
{
    control->name_label = g_object_new (GTK_TYPE_LABEL,
                                        "label", "(unattached)",
                                        "justify", GTK_JUSTIFY_CENTER,
                                        NULL);
    gtk_box_pack_start (GTK_BOX (control), control->name_label, FALSE, FALSE, 0);
    gtk_widget_show (control->name_label);

    control->value = 0.f;
    control->lower = -1.f;
    control->upper = 1.f;
    control->warp = SS_WARP_LINEAR;

    /* adjustment just goes from 0 to 1 */
    control->adjustment = gtk_adjustment_new ((control->value - control->lower) /
                                              (control->upper - control->lower),
                                              0, 1, 0.05, 0.25, 0);
    control->scale = gtk_vscale_new (GTK_ADJUSTMENT (control->adjustment));
    gtk_scale_set_draw_value (GTK_SCALE (control->scale), FALSE);
    gtk_scale_set_digits (GTK_SCALE (control->scale), -1);
    gtk_range_set_inverted (GTK_RANGE (control->scale), TRUE);
    g_signal_connect (control->adjustment, "value_changed",
                      G_CALLBACK (on_adjustment_value_changed), control);
    gtk_box_pack_start (GTK_BOX (control), control->scale, TRUE, TRUE, 0);
    gtk_widget_show (control->scale);

    control->value_label = gtk_label_new ("");
    gtk_box_pack_start (GTK_BOX (control), control->value_label, FALSE, FALSE, 0);
    gtk_widget_show (control->value_label);
    
    on_adjustment_value_changed (GTK_ADJUSTMENT (control->adjustment), control);
}

static void
ss_slider_control_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
    SSSliderControl *control;

    control = SS_SLIDER_CONTROL (object);

    if (prop_id == PROP_MODEL) {
        if (control->model) {
            gst_element_disable_threadsafe_properties ((GstElement*)control->model);
            g_signal_handlers_disconnect_by_func (control->model,
                                                  model_state_change, control);
        }
        control->model = g_value_get_object (value);
        if (control->model) {
            gst_element_enable_threadsafe_properties ((GstElement*)control->model);
            g_signal_connect (control->model, "state-change",
                              G_CALLBACK (model_state_change), control);
        }
        return;
    } else if (prop_id == PROP_NAME) {
        control->name = g_strdup (g_value_get_string (value));
        g_object_set (control->name_label, "label", control->name, NULL);
        return;
    }

    switch (prop_id) {
    case PROP_LOWER:
        control->lower = g_value_get_float (value);
        break;
    case PROP_UPPER:
        control->upper = g_value_get_float (value);
        break;
    case PROP_VALUE:
        control->value = CLAMP (g_value_get_float (value), control->lower, control->upper);
        break;
    case PROP_WARP:
        control->warp = g_value_get_enum (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        return;
    }

    /* sane bounds checking */
    if (control->warp == SS_WARP_EXPONENTIAL)
        control->lower = control->lower < 0.001 ? 0.001 : control->lower;
    control->upper = control->upper > control->lower ? control->upper : control->lower + 1;
    control->value = CLAMP (control->value, control->lower, control->upper);

    /* adjust exponential parameters */
    if (control->warp == SS_WARP_EXPONENTIAL) {
        control->a = control->lower;
        control->b = log (control->upper / control->lower);
    }

    /* adjust adjustment */
    if (control->adjustment) {
        if (control->warp == SS_WARP_LINEAR)
            GTK_ADJUSTMENT (control->adjustment)->value =
                (control->value - control->lower) / (control->upper - control->lower);
        else if (control->warp == SS_WARP_EXPONENTIAL)
            GTK_ADJUSTMENT (control->adjustment)->value =
                log (control->value / control->a) / control->b;
        gtk_adjustment_value_changed (GTK_ADJUSTMENT (control->adjustment));
    }
}

static void 
ss_slider_control_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
    SSSliderControl *control;
   
    control = SS_SLIDER_CONTROL (object);
   
    switch (prop_id) {
    case PROP_NAME:
        g_value_set_string (value, g_strdup (control->name));
        break;
    case PROP_LOWER:
        g_value_set_float (value, control->lower);
        break;
    case PROP_UPPER:
        g_value_set_float (value, control->upper);
        break;
    case PROP_VALUE:
        g_value_set_float (value, control->value);
        break;
    case PROP_WARP:
        g_value_set_enum (value, control->warp);
        break;
    case PROP_MODEL:
        g_value_set_object (value, control->model);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
on_adjustment_value_changed (GtkAdjustment *adjustment, SSSliderControl *control)
{
    gchar *new_value;

    if (control->warp == SS_WARP_LINEAR)
        control->value = control->lower + adjustment->value * (control->upper - control->lower);
    else if (control->warp == SS_WARP_EXPONENTIAL)
        control->value = control->a * exp (control->b * adjustment->value);
    else
        g_assert_not_reached ();

    new_value = g_strdup_printf ("%2f", control->value);
    g_object_set (control->value_label, "label", new_value, NULL);
    g_free (new_value);

    if (control->model)
        gst_element_set ((GstElement*)control->model, "value", control->value, NULL);
}

static void
model_state_change (GstElement *model, GstElementState old,
                    GstElementState new, SSSliderControl *control)
{
    if (old == GST_STATE_READY && new == GST_STATE_PAUSED) {
        GstPad *pad;
        gchar *label;
        gst_element_set (model, "value", control->value, NULL);
        pad = gst_element_get_pad (model, "src");
        if (pad && GST_PAD_PEER (pad)) {
            GstPad *peer = GST_PAD_PEER (pad);
            label = g_strdup_printf ("%s\n%s",
                                     GST_OBJECT_NAME (GST_OBJECT_PARENT (peer)),
                                     GST_OBJECT_NAME (peer));
            g_object_set (control->name_label, "label", label, NULL);
            g_free (label);
        }
    }
    
    if (old == GST_STATE_READY && new == GST_STATE_NULL)
        gtk_widget_destroy (GTK_WIDGET (control));
}
