/* Soundscrape
 * Copyright (C) 2003 Andy Wingo <wingo at pobox dot com>
 *
 * ss-unary-op.c: Unary operations
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Some code based on gstadder.c from gst-plugins:
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *               2000 Wim Taymans <wim.taymans@chello.be>
 *               2001 Thomas <thomas@apestaart.org>
 *
 * Some code based on UnaryOpUgens.cpp from SuperCollider:
 * Copyright (C) 2002 James McCartney
 */

#include <config.h>
#include "ss-unary-op.h"
#include <math.h>

/* a lot of this could probably be optimized if I used routines from
 * SC_InlineUnaryOp.h */

/* declare the guts of the operations here in the beginning */
#define SS_MINFLOAT 0.000001
static float rlog2 = 0.69314718056f;
static inline float log2 (float a)
{
    return rlog2 * log (a);
}

static gfloat do_neg_c (gfloat a)
{
    return -a;
}
static void do_neg_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = -a[i];
}
static gfloat do_reciprocal_c (gfloat a)
{
    return a == 0.f ? G_MAXFLOAT : 1.f / a;
}
static void do_reciprocal_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] == 0.f ? G_MAXFLOAT : 1.f / a[i];
}
static gfloat do_abs_c (gfloat a)
{
    return a < 0.f ? -a : a;
}
static void do_abs_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -a[i] : a[i];
}
static gfloat do_floor_c (gfloat a)
{
    return floorf (a);
}
static void do_floor_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = floorf (a[i]);
}
static gfloat do_ceil_c (gfloat a)
{
    return ceilf (a);
}
static void do_ceil_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = ceilf (a[i]);
}
static gfloat do_frac_c (gfloat a)
{
    return a - floorf (a);
}
static void do_frac_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] - floorf (a[i]);
}
static gfloat do_sign_c (gfloat a)
{
    return a < 0.f ? -1.f : a == 0.f ? 0.f : 1.f;
}
static void do_sign_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -1.f : a[i] == 0.f ? 0.f : 1.f ;
}
static gfloat do_squared_c (gfloat a)
{
    return a * a;
}
static void do_squared_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] *= a[i];
}
static gfloat do_cube_c (gfloat a)
{
    return a * a * a;
}
static void do_cube_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] *= a[i] * a[i];
}
static gfloat do_sqrt_c (gfloat a)
{
    return a < 0.f ? -sqrt (-a) : sqrt (a);
}
static void do_sqrt_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -sqrt (-a[i]) : sqrt (a[i]);
}
static gfloat do_exp_c (gfloat a)
{
    return exp (a);
}
static void do_exp_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = exp (a[i]);
}
static gfloat do_midi2cps_c (gfloat a)
{
    /* copied from supercollider */
    return 440.f * pow(2.f, (a - 69.f) * 0.083333333333f);
}
static void do_midi2cps_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = 440.f * pow(2.f, (a[i] - 69.f) * 0.083333333333f);
}
static gfloat do_cps2midi_c (gfloat a)
{
    /* copied from supercollider */
    return log2(a * 0.0022727272727f) * 12.f + 69.f;
}
static void do_cps2midi_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = log2(a[i] * 0.0022727272727f) * 12.f + 69.f;;
}
static gfloat do_midi2ratio_c (gfloat a)
{
    /* copied from supercollider */
    return pow(2.f, a * 0.083333333333f);
}
static void do_midi2ratio_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = pow(2.f, a[i] * 0.083333333333f);
}
static gfloat do_ratio2midi_c (gfloat a)
{
    /* copied from supercollider */
    return 12.f * log2 (a);
}
static void do_ratio2midi_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = 12.f * log2 (a[i]);
}
static gfloat do_oct2cps_c (gfloat a)
{
    /* copied from supercollider */
    return 440.f * pow(2.f, a - 4.75f);
}
static void do_oct2cps_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = 440.f * pow(2.f, a[i] - 4.75f);
}
static gfloat do_cps2oct_c (gfloat a)
{
    /* copied from supercollider */
    return log2 (a * 0.0022727272727f) + 4.75f;
}
static void do_cps2oct_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = log2 (a[i] * 0.0022727272727f) + 4.75f;
}
static gfloat do_amp2db_c (gfloat a)
{
    /* copied from supercollider */
    return log10 (a) * 20.f;
}
static void do_amp2db_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = log10 (a[i]) * 20.f;
}
static gfloat do_db2amp_c (gfloat a)
{
    /* copied from supercollider */
    return pow (10.f, a * .05f);
}
static void do_db2amp_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = pow (10.f, a[i] * .05f);
}
static gfloat do_log_c (gfloat a)
{
    return a < 0.f ? -log (-a) : a == 0.f ? 0.f : log (a);
}
static void do_log_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -log (-a[i]) : a[i] == 0.f ? 0.f : log (a[i]);
}
static gfloat do_log2_c (gfloat a)
{
    return a < 0.f ? -log2 (-a) : a == 0.f ? 0.f : log2 (a);
}
static void do_log2_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -log2 (-a[i]) : a[i] == 0.f ? 0.f : log2 (a[i]);
}
static gfloat do_log10_c (gfloat a)
{
    return a < 0.f ? -log10 (-a) : a == 0.f ? 0.f : log10 (a);
}
static void do_log10_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] < 0.f ? -log10 (-a[i]) : a[i] == 0.f ? 0.f : log10 (a[i]);
}
static gfloat do_sin_c (gfloat a)
{
    return sin (a);
}
static void do_sin_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = sin (a[i]);
}
static gfloat do_cos_c (gfloat a)
{
    return cos (a);
}
static void do_cos_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = cos (a[i]);
}
static gfloat do_tan_c (gfloat a)
{
    return tan (a);
}
static void do_tan_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = tan (a[i]);
}
static gfloat do_asin_c (gfloat a)
{
    return asin (a);
}
static void do_asin_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = asin (a[i]);
}
static gfloat do_acos_c (gfloat a)
{
    return acos (a);
}
static void do_acos_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = acos (a[i]);
}
static gfloat do_atan_c (gfloat a)
{
    return atan (a);
}
static void do_atan_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = atan (a[i]);
}
static gfloat do_sinh_c (gfloat a)
{
    return sinh (a);
}
static void do_sinh_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = sinh (a[i]);
}
static gfloat do_cosh_c (gfloat a)
{
    return cosh (a);
}
static void do_cosh_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = cosh (a[i]);
}
static gfloat do_tanh_c (gfloat a)
{
    return tanh (a);
}
static void do_tanh_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = tanh (a[i]);
}
static gfloat do_distort_c (gfloat a)
{
    /* copied from supercollider */
    return a / (1.f + fabsf (a));
}
static void do_distort_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] /= 1.f + fabsf (a[i]);
}
static gfloat do_softclip_c (gfloat a)
{
    /* copied from supercollider */
    float absa = fabsf (a);
    return absa < 0.5 ? a : (absa - 0.25f) / a;
}
static void do_softclip_a (gfloat *a, gint how_many) 
{
    gint i;
    float absa;
    for (i=0; i<how_many; i++) {
        absa = fabsf (a[i]);
        a[i] = absa < 0.5 ? a[i] : (absa - 0.25f) / a[i];
    }
}
static gfloat do_positivep_c (gfloat a)
{
    return a >= 0.f ? 1.f : 0.f;
}
static void do_positivep_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] >= 0.f ? 1.f : 0.f;
}
static gfloat do_negativep_c (gfloat a)
{
    return a <= 0.f ? 1.f : 0.f;
}
static void do_negativep_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] <= 0.f ? 1.f : 0.f;
}
static gfloat do_strictly_positivep_c (gfloat a)
{
    return a > 0.f ? 1.f : 0.f;
}
static void do_strictly_positivep_a (gfloat *a, gint how_many) 
{
    gint i;
    for (i=0; i<how_many; i++)
        a[i] = a[i] > 0.f ? 1.f : 0.f;
}

typedef struct {
    gchar *factory_name;
    gfloat (*do_op_c) (gfloat a);
    void (*do_op_a) (gfloat *a, gint how_many);
    gchar *description;
    GType type;
} SSUnaryOpDetails;

static SSUnaryOpDetails details[] = {
    {"neg", do_neg_c, do_neg_a, "Negation", 0},
    {"reciprocal", do_reciprocal_c, do_reciprocal_a, "Reciprocal", 0},
    {"abs", do_abs_c, do_abs_a, "Abs", 0},
    {"floor", do_floor_c, do_floor_a, "Floor", 0},
    {"ceil", do_ceil_c, do_ceil_a, "Ceil", 0},
    {"frac", do_frac_c, do_frac_a, "Frac", 0},
    {"sign", do_sign_c, do_sign_a, "Sign", 0},
    {"squared", do_squared_c, do_squared_a, "Squared", 0},
    {"cube", do_cube_c, do_cube_a, "Cube", 0},
    {"sqrt", do_sqrt_c, do_sqrt_a, "Sqrt", 0},
    {"exp", do_exp_c, do_exp_a, "Exp", 0},
    {"midi->cps", do_midi2cps_c, do_midi2cps_a, "MidiCps", 0},
    {"cps->midi", do_cps2midi_c, do_cps2midi_a, "CpsMidi", 0},
    {"midi->ratio", do_midi2ratio_c, do_midi2ratio_a, "MidiRatio", 0},
    {"ratio->midi", do_ratio2midi_c, do_ratio2midi_a, "RatioMidi", 0},
    {"oct->cps", do_oct2cps_c, do_oct2cps_a, "OctCps", 0},
    {"cps->oct", do_cps2oct_c, do_cps2oct_a, "CpsOct", 0},
    {"amp->db", do_amp2db_c, do_amp2db_a, "AmpDb", 0},
    {"db->amp", do_db2amp_c, do_db2amp_a, "DbAmp", 0},
    {"log", do_log_c, do_log_a, "Log", 0},
    {"log2", do_log2_c, do_log2_a, "Log2", 0},
    {"log10", do_log10_c, do_log10_a, "Log10", 0},
    {"sin", do_sin_c, do_sin_a, "Sin", 0},
    {"cos", do_cos_c, do_cos_a, "Cos", 0},
    {"tan", do_tan_c, do_tan_a, "Tan", 0},
    {"asin", do_asin_c, do_asin_a, "ASin", 0},
    {"acos", do_acos_c, do_acos_a, "ACos", 0},
    {"atan", do_atan_c, do_atan_a, "ATan", 0},
    {"sinh", do_sinh_c, do_sinh_a, "Sinh", 0},
    {"cosh", do_cosh_c, do_cosh_a, "Cosh", 0},
    {"tanh", do_tanh_c, do_tanh_a, "Tanh", 0},
    {"distort", do_distort_c, do_distort_a, "Distort", 0},
    {"softclip", do_softclip_c, do_softclip_a, "SoftClip", 0},
    {"positive?", do_positivep_c, do_positivep_a, "PositiveP", 0},
    {"negative?", do_negativep_c, do_negativep_a, "NegativeP", 0},
    {"strictly-positive?", do_strictly_positivep_c, do_strictly_positivep_a, "StrictlyPositive", 0},
    {NULL, NULL, NULL, NULL, 0}
};

/* then the generic support */
SS_PAD_TEMPLATE_SRC_FACTORY (ss_unary_op_src_factory, "output");
SS_PAD_TEMPLATE_SINK_FACTORY (ss_unary_op_sink_factory, "input", 0.f);


static void 		ss_unary_op_class_init		(SSUnaryOpClass *klass);
static void 		ss_unary_op_init		(SSUnaryOp *unary_op);

static void 		ss_unary_op_chain  		(GstPad *pad, GstData *data);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_unary_op_get_type, SS_TYPE_ELEMENT,
                          "SSUnaryOp", SSUnaryOpClass, SSUnaryOp,
                          NULL /* abstract type */, NULL /* abstract type */);

static void
ss_unary_op_class_init (SSUnaryOpClass *klass) 
{
    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);
}

static void 
ss_unary_op_init (SSUnaryOp *this) 
{
    SSUnaryOpDetails *d;

    for (d=details; d; d++) {
        if (G_TYPE_FROM_INSTANCE (this) == d->type) {
            this->do_op_c = d->do_op_c;
            this->do_op_a = d->do_op_a;
            break;
        }
    }

    this->srcpad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_unary_op_src_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->srcpad);

    this->sink = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_unary_op_sink_factory ()));
    gst_pad_set_chain_function (this->sink, ss_unary_op_chain);
    gst_element_add_pad (GST_ELEMENT (this), this->sink);
}

static void
ss_unary_op_chain (GstPad *pad, GstData *data)
{
    SSUnaryOp *this;

    this = (SSUnaryOp*)GST_OBJECT_PARENT (pad);

    /* we're not event-aware, so we shouldn't get events */
    if (SS_IS_CONTROL (data)) {
        /* control rate */
        SS_CONTROL_VALUE (data) = this->do_op_c (SS_CONTROL_VALUE (data));
        ss_pad_push ((SSPad*)this->srcpad, data);
    } else {
        /* audio rate */
        this->do_op_a (SS_BUFFER_DATA (data),
                       SS_BUFFER_GET_FRAMES (data));
        ss_pad_push ((SSPad*)this->srcpad, data);
    }
}

gboolean
init_ss_unary_op (GstPlugin *plugin)
{
    SSUnaryOpDetails *d;
    GstElementClass *class;

    for (d=details; d->factory_name!=NULL; d++) {
        GstElementDetails *edetails;
        static const GTypeInfo type_info = {
            sizeof (SSUnaryOpClass), NULL, NULL,
            (GClassInitFunc) ss_unary_op_class_init, NULL, NULL,
            sizeof (SSUnaryOp), 0,
            (GInstanceInitFunc) ss_unary_op_init,
        };

        edetails = g_new0 (GstElementDetails, 1);
        edetails->longname = d->factory_name;
        edetails->klass = "Soundscrape/Operators/Unary";
        edetails->description = d->description;
        edetails->author = "Andy Wingo <wingo at pobox dot com>";

        d->type = g_type_register_static (SS_TYPE_UNARY_OP,
                                          g_strdup_printf ("SSUnaryOp%s", d->description),
                                          &type_info, 0);
        class = g_type_class_ref (d->type);
        gst_element_class_add_pad_template (class, ss_unary_op_sink_factory ());
        gst_element_class_add_pad_template (class, ss_unary_op_src_factory ());
        gst_element_class_set_details (class, edetails);

        gst_element_register (plugin, d->factory_name, GST_RANK_NONE, d->type);
    }
    return TRUE;
}

