;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;A @dfn{stream} is a lazy sequence of values. Streams are implemented
;;in soundscrape with functions. These functions may be called in two
;;ways. If no arguments are passed, the caller is requesting the next
;;value in the sequence. If any arguments are passed, the caller is
;;requesting the stream to reset itself to the beginning. A false return
;;value is interpreted to be the end of the stream.
;;
;;By this definition, a stream can be as simple as @code{(lambda args
;;3)} -- an infinite sequence of @code{3}s. A @code{yielding-thunk}
;;(@pxref{(soundscrape misc)}) is also a stream.
;;
;;This module defines a number of convenience functions to create
;;streams. Of course you can still create them on your own with
;;@code{yielding-thunk} or with closures, but these might be easier to
;;use.
;;
;;List streams all have the prototype @code{(stream:@var{type}
;;repeat-count . args)}. @var{repeat-count} is interpreted differently
;;by different kinds of streams. In all cases, the elements of
;;@var{args} may be either scalar values or other streams. In the case
;;that the element is a stream, the stream will be allowed to run until
;;it ends by returning @code{#f}.
;;
;;; Code:

(define-module (soundscrape streams)
  :use-module (soundscrape misc))

(define-public (stream:apply func . args)
  "Apply a function to a stream."
  (define (get-value)
    (let ((arg-values (map (lambda (x) (if (procedure? x) (x) x)) args)))
      (if (and-map identity arg-values)
          (apply func arg-values)
          #f)))
  (yielding-thunk
   (for-each (lambda (x) (if (procedure? x) (x 'reset))) args)
   (let loop ((val (get-value)))
     (if val (begin (yield val) (loop (get-value)))))
   (while #t (yield #f))))

(define-public (stream:selector pass? stream)
  "Only returns a value x for which @code{(@var{pass?} x)}."
  (if (not (procedure? pass?)) (error "Invalid selector"))
  (if (not (procedure? stream)) (error "Invalid stream"))
  (yielding-thunk
   (stream 'reset)
   (let loop ((val (stream)))
     (if val
         (begin
           (if (pass? val) (yield val))
           (loop (stream)))))
   (while #t (yield #f))))

(define-public (stream:rejector deny? stream)
  "Only returns a value x for which @code{(not (@var{deny?} x))}."
  (if (not (procedure? deny?)) (error "Invalid rejector"))
  (if (not (procedure? stream)) (error "Invalid stream"))
  (yielding-thunk
   (stream 'reset)
   (let loop ((val (stream)))
     (if val
         (begin
           (if (not (deny? val)) (yield val))
           (loop (stream)))))
   (while #t (yield #f))))

(define-public (stream:series start step length)
  "An arithmetic series starting at @var{start}, incrementing by
@var{step}, and with length @var{length}."
  (if (not (and-map number? (list start step length)))
      (error "START, STEP, and LENGTH must all be numbers."))
  (yielding-thunk
   (let loop ((val start) (remaining length))
     (if (positive? remaining)
         (begin
           (yield val)
           (loop (+ val step) (1- remaining)))))
   (while #t (yield #f))))

(define-public (stream:geom start factor length)
  "A geometric series starting at @var{start}, with a factor of
@var{factor}, and with length @var{length}."
  (if (not (and-map number? (list start factor length)))
      (error "START, FACTOR, and LENGTH must all be numbers."))
  (yielding-thunk
   (let loop ((val start) (remaining length))
     (if (positive? remaining)
         (begin
           (yield val)
           (loop (* val factor) (1- remaining)))))
   (while #t (yield #f))))

(define-public (stream:sequence repeat-count . l)
  "Make a stream that returns @var{args} in order. The sequence will
repeat @var{repeat-count} times."
  (if (zero? (length l))
      (error "Bad sequence."))
  (if (not (or (and (exact? repeat-count) (positive? repeat-count))
               (not repeat-count)))
      (error "Bad number of iterations."))
  (yielding-thunk
   (for-each (lambda (x) (if (procedure? x) (x 'reset))) l)
   (repeat i repeat-count
           (let loop ((l l))
             (if (not (null? l))
                 (if (procedure? (car l))
                     (begin
                       (let proc-loop ((val ((car l))))
                         (if val
                             (begin
                               (yield val)
                               (proc-loop ((car l))))))
                       ((car l) 'reset)
                       (loop (cdr l)))
                     (begin
                       (yield (car l))
                       (loop (cdr l)))))))
   (while #t (yield #f))))

(define-public (stream:chooser repeat-count . l)
  "Make a stream that chooses random elements from @var{args}. The stream
will finish after returning @var{repeat-count} items."
  (if (zero? (length l))
      (error "Bad sequence."))
  (if (not (or (and (exact? repeat-count) (positive? repeat-count))
               (not repeat-count)))
      (error "Bad number of iterations."))
  (yielding-thunk
   (for-each (lambda (x) (if (procedure? x) (x 'reset))) l)
   (repeat i repeat-count
           (let ((val (choose l)))
             (if (procedure? val)
                 (begin
                   (let proc-loop ((res (val)))
                     (if res (begin (yield res) (proc-loop (val)))))
                   (val 'reset))
                 (yield val))))
   (while #t (yield #f))))

(define-public (stream:xchooser repeat-count . l)
  "Same as chooser, but will not return the same element twice in a row."
  (if (zero? (length l))
      (error "Bad sequence."))
  (if (not (or (and (exact? repeat-count) (positive? repeat-count))
               (not repeat-count)))
      (error "Bad number of iterations."))
  (yielding-thunk
   (let ((len (length l))
         (last-offset -1))
     (repeat i repeat-count
             (set! last-offset
                   (let lp ((o (random len)))
                     (if (not (eq? o last-offset))
                         o
                         (lp (random len)))))
             (let ((val (list-ref l last-offset)))
               (if (procedure? val)
                   (begin
                     (let proc-loop ((res (val)))
                       (if res (begin (yield res) (proc-loop (val)))))
                     (val 'reset))
                   (yield val)))))
   (while #t (yield #f))))

(define-public (stream:shuffled-sequence repeat-count . l)
  "Same as sequence, but @var{args} are shuffled beforehand."
  (apply stream:sequence repeat-count (shuffle! l)))
