;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Includes modules in the help window. For internal use only.
;;
;;; Code:

(define-module (soundscrape ui help-modules)
  :use-module (srfi srfi-13)
  :use-module (ice-9 documentation)
  :use-module (soundscrape config)
  :use-module (soundscrape util)
  :use-module (soundscrape ui help-tree)
  :use-module (soundscrape ui stext texinfo)
  :use-module (oop goops)
  :use-module (srfi srfi-1)
  :export (find-modules))

;; List for sorting the definitions in a module
(define defs '((deftp . 1)
               
               (defcv . 2)
               (defivar . 1)
               (deftypeivar . 2)
               (defop . 2)
               (deftypeop . 3)
               (defmethod . 1)
               (deftypemethod . 2)

               (defopt . 0)

               (defvr . 1)
               (defvar . 0)
               (deftypevr . 2)
               (deftypevar . 1)

               (deffn . 1)
               (deftypefn . 2)
               (defmac . 0)
               (defspec . 0)
               (defun . 0)
               (deftypefun . 1)))

;; HACK!
(define <class> (module-ref (resolve-interface '(oop goops)) '<class>))

(define make-anchor-name
  (let ((count -1))
    (lambda ()
      (set! count (1+ count))
      (string-append "indexed-module-anchor-" (number->string count)))))
  
(define (find-modules root no-variables-proc)
  "Find modules in path @var{root}, but not documenting the modules for
which @code{(@var{no-variables-proc} module)} is @code{#t}."
  (define (recursable? path entry)
    (and (file-exists? path)
         (file-is-directory? path)
         (not (equal? entry "..")) (not (equal? entry "."))))
  
  (define (module? path)
    (and (file-exists? path)
         (string-suffix? ".scm" path)))
  
  (define (module-file->name file)
    (let ((name-pieces (string-split
                        (substring file 0 (- (string-length file) 4))
                        #\/)))
      (string-append "(" (string-join name-pieces " ") ")")))

  (define (module-file->list file)
    (let ((name-pieces (string-split
                        (substring file 0 (- (string-length file) 4))
                        #\/)))
      (map string->symbol name-pieces)))
    
  (define (make-stext-func module-file)
    (lambda ()
      (let* ((module-name (module-file->name module-file))
             (module-stext (with-input-from-string (file-commentary module-file)
                             texinfo->stext))
             (module-list (module-file->list module-file))
             (module (if (no-variables-proc module-list)
                         #f
                         (resolve-interface module-list)))
             (sort-proc (lambda args
                          (define (find-index stext)
                            ;;; GRRR! This is srfi-1's list-index.
                            (list-index (lambda (x) (eq? x (caar stext)))
                                        (map car defs)))
                          (let ((indices (map find-index args)))
                            (cond
                             ((not (apply eq? indices))
                              (apply < indices))
                             (else
                              (apply string<=?
                                     (map
                                      (lambda (i)
                                        (list-ref
                                         (string-split (cadar i) #\space)
                                         (cdr (list-ref defs (car indices)))))
                                      args)))))))
             (stext-helper (lambda (deftype args docs)
                             (list
                              (list deftype (string-join args " "))
                              (list 'anchor (make-anchor-name))
                              docs)))
             (variables-stext
              (cons
               'text
               (if (not module)
                   '("The variables in this module are for internal use only. "
                     "This module should not be used by the user.")
                   (sort
                    (module-map
                     (lambda (sym var)
                       (let* ((val (variable-ref var))
                              (doc (object-documentation val))
                              (doc-stext (cond
                                          ((string? doc)
                                           (with-input-from-string doc
                                             texinfo->stext))
                                          ((list? doc)
                                           ;; assume it's stext
                                           doc)
                                          (else
                                           '(text "[undocumented]")))))
                         ;; An interesting dilemma here: How to document
                         ;; generic functions? Really we are just
                         ;; documenting methods on classes, not
                         ;; operations in general. Let's just document
                         ;; methods with classes. (TODO)
                         (cond
                          ((and (list? (car doc-stext))
                                (memq (caar doc-stext) defs))
                           doc-stext)
                          ((is-a? val <class>)
                           (stext-helper
                            'deftp (list "Class" (symbol->string sym))
                            doc-stext))
                          ((is-a? val <macro>)
                           (stext-helper
                            'defspec (list (symbol->string sym))
                            doc-stext))
                          ((is-a? val <procedure>)
                           (stext-helper
                            'defun (map symbol->string
                                        (cons
                                         sym
                                         (let ((source (procedure-source val)))
                                           (if (and source (list? (cadr source)))
                                               (cadr source)
                                               '()))))
                            doc-stext))
                          (else
                           (stext-helper
                            'defvar (list (symbol->string sym))
                            doc-stext)))))
                     module)
                    sort-proc)))))
        `((stext ,module-name)
          (node ,module-name)
          (subsection "Overview")
          ,(with-input-from-string (file-commentary module-file)
             texinfo->stext)
          (subsection "Usage")
          ,variables-stext))))
  
  (append
   ;; At the top level, we want to look also for a file describing the
   ;; whole tree.
   (let ((module-file (string-append root ".scm")))
     (if (and (file-exists? module-file) (module? module-file))
         (list (make <ss-help-node>
                 :name (module-file->name module-file)
                 :make-value (make-stext-func module-file)))
         '()))
   
   ;; For sublevels we just recurse the directories.
   (let scan-path ((path ""))
     (let ((dir (opendir (build-path root path))))
       (let next-entry ((entries (let read-entries ((entries '()))
                                   (let ((entry (readdir dir)))
                                     (if (eof-object? entry)
                                         (begin
                                           (closedir dir)
                                           (sort! entries string-ci>?))
                                         (read-entries (cons entry entries))))))
                        (modules '()))
         (if (null? entries)
             modules ;; entries were already sorted in reverse
             (let ((full-path (build-path root path (car entries))))
               (cond
                ((recursable? full-path (car entries))
                 (next-entry (cdr entries)
                             (append (scan-path (build-path path (car entries)))
                                     modules)))
                ((module? full-path)
                 (next-entry (cdr entries)
                             (cons (make <ss-help-node>
                                     :name (module-file->name full-path)
                                     :make-value (make-stext-func full-path))
                                   modules)))
                (else
                 (next-entry (cdr entries) modules))))))))))
  
