;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;A help browser for Soundscrape.
;;
;;; Code:

(define-module (soundscrape ui help)
  :use-module (ice-9 optargs)
  :use-module (gnome gtk)
  :use-module (srfi srfi-13)
  :use-module (soundscrape util)
  :use-module (soundscrape config)
  :use-module (soundscrape ui help-tree)
  :use-module (soundscrape ui help-index)
  :use-module (soundscrape ui help-modules)
  :use-module (soundscrape ui help-ugens)
  :use-module (soundscrape ui stext help-nodes)
  :use-module (soundscrape ui stext texinfo)
  :use-module (soundscrape ui stext gtk-text-buffer)
  :export (show-help on-help)
  :re-export (the-help-tree))

(define default-buffer
  (stext->gtk-text-buffer
   '((stext "Soundscrape Help")
     "This is the top text.")))

(define (get-help-buffer iter)
  (get-value the-help-tree iter 2))

(define (get-help-mark iter node-name)
  (get-mark (get-help-buffer iter) (string-append "node-" node-name)))

(define the-help-window #f)

(define-class <ss-ui-help-window> (<gtk-window>)
  topics-treeview
  textbuffer
  textview
  current-node)

(define-method (initialize (w <ss-ui-help-window>) initargs)
  (let* ((textbuffer (make <gtk-text-buffer>))
         (textview (make <gtk-text-view>
                     :editable #f :cursor-visible #f :wrap-mode 'word
                     :pixels-above-lines 2 :pixels-below-lines 6
                     :pixels-inside-wrap 1 ;; :justification 'fill <- not supported yet
                     :right-margin 10 :left-margin 10))
         (pane (make <gtk-hpaned> :position 200))
         (notebook (make <gtk-notebook> :show-border #f))
         (text-scroll (make <gtk-scrolled-window>
                        :hscrollbar-policy 'automatic :vscrollbar-policy 'automatic
                        :shadow-type 'in)))

    (add w pane)
    (pack1 pane notebook #f #t)
    (pack2 pane text-scroll #t #t)
    (add text-scroll textview)
    (set-default-size w 700 400)
    (set w 'title "soundscrape: help")
    
    (connect w 'delete-event (lambda (w e) (hide w) #t))

    ;; The "Topics" page
    (let* ((treemodel the-help-tree)
           (treeview (make <gtk-tree-view> :model treemodel :headers-visible #f))
           (scroll (make <gtk-scrolled-window>
                     :hscrollbar-policy 'automatic :vscrollbar-policy 'automatic
                     :shadow-type 'in))
           (cellrenderer (make <gtk-cell-renderer-text>))
           (column (make <gtk-tree-view-column>))
           (selection (get-selection treeview)))
      (slot-set! w 'topics-treeview treeview)
      (append-page notebook scroll (make <gtk-label> :label "Topics"))
      (add scroll treeview)
      (set-mode selection 'single)
      (pack-start column cellrenderer #t)
      (add-attribute column cellrenderer "text" 0)
      (append-column treeview column)
      (connect
       selection 'changed
       (lambda (selection)
         (call-with-values (lambda () (get-selected selection))
           (lambda (model iter)
             (slot-set! w 'current-node iter)
             (set-buffer textview
                         (if iter (get-help-buffer iter) default-buffer))))))
      (emit selection 'changed))
    
    ;; The "Index" page
    (let* ((vbox (make <gtk-vbox>))
           (entry (make <gtk-entry>))
           (list (make <gtk-tree-view>))
           (treemodel (make <ss-help-index>))
           (scroll (make <gtk-scrolled-window>
                     :hscrollbar-policy 'automatic :vscrollbar-policy 'automatic
                     :shadow-type 'in))
           (treeview (make <gtk-tree-view> :model treemodel :headers-visible #f))
           (cellrenderer (make <gtk-cell-renderer-text>))
           (column (make <gtk-tree-view-column>))
           (selection (get-selection treeview)))
      (append-page notebook vbox (make <gtk-label> :label "Index"))
      (pack-start vbox entry #f #f 0)
      (pack-start vbox scroll #t #t 0)
      (add scroll treeview)
      (set-mode selection 'single)
      (pack-start column cellrenderer #t)
      (add-attribute column cellrenderer "text" 0)
      (append-column treeview column)
      (set-text entry "(indexing takes some time)")
      (select-region entry 0 -1)
      (connect
       entry 'changed
       (lambda (entry)
         (let ((text (get-text entry)))
           (set-filter treemodel
                       (if (equal? text "")
                           #f
                           (lambda (row)
                             (string-contains-ci (car row) text)))))))
      (connect
       selection 'changed
       (lambda (selection)
         (call-with-values (lambda () (get-selected selection))
           (lambda (model iter)
             (if iter
                 (let ((manual-name (get-value model iter 1))
                       (node-name (get-value model iter 2)))
                   (select-node w node-name manual-name))))))))

    (slot-set! w 'textbuffer textbuffer)
    (slot-set! w 'textview textview)
    (show-all w)
    (hide w)))

;; If manual-name is #f, look in the currently selected manual.
;; If node-name is #f, select the top of the manual.
(define-method (select-node (window <ss-ui-help-window>) node-name manual-name)
  (define (find-node-named top name)
    (or-map
     (lambda (node)
       (if (get-help-mark node name)
           node
           (find-node-named node name)))
     (iter-children the-help-tree top)))
  (let* ((manual (if manual-name
                     (let loop ((manual (get-iter-first the-help-tree)))
                       (cond
                        ((not manual)
                         (error "Unknown manual:" manual-name))
                        ((string=? (get-value the-help-tree manual 0)
                                   manual-name)
                         manual)
                        (else
                         (loop (iter-next the-help-tree manual)))))
                     (let ((current-node (slot-ref window 'current-node)))
                       (if (not current-node)
                           (error
                            "The TOP node must specify a manual in its cross-references.")
                           (let find-top ((node current-node))
                             (let ((parent (iter-parent the-help-tree node)))
                               (if (not parent)
                                   node
                                   (find-top parent))))))))
         (node (if (and node-name (not (string=? node-name "top")))
                   (or
                    (find-node-named manual node-name)
                    (error "No such node in manual:" manual-name node-name))
                   manual)))
    (let* ((treeview (slot-ref window 'topics-treeview))
           (path (get-path the-help-tree node)))
      (expand-to-path treeview path)
      (select-path (get-selection treeview) path)
      (scroll-to-cell treeview path #f #f 0 0))
    ;; now the buffer should be correct...
    ;; scroll in an idle loop to give the textview time to compute line heights
    (g-idle-add
     (lambda ()
       (scroll-to-mark (slot-ref window 'textview)
                       (if node-name
                           (get-help-mark node node-name)
                           (create-mark (get-help-buffer node)
                                        #f
                                        (get-start-iter (get-help-buffer node))
                                        #t))
                       0.15 #f 0 0)
       #f))))

(define (ensure-help-window)
  (if (not the-help-window)
      (begin
        (populate-help)
        (set! the-help-window (make <ss-ui-help-window>))))
  the-help-window)

(define* (show-help :optional (node-name #f) (manual-name #f))
  "Show the help window. If @var{manual-name} is @code{#f}, look in the
currently selected manual. If @var{node-name} is @code{#f}, select the
top of the manual."
  (let ((window (ensure-help-window)))
    (show window)
    (if (or node-name manual-name)
        (select-node window node-name manual-name))
    (present window)))


(define (on-help manual-symbol)
  "Creates a signal handler that will show the manual named @var{manual-symbol}.
If @var{manual-symbol} is @code{#f}, just show the window."
  ;; Actually makes a signal handler -- neat, eh? wish libglade
  ;; unescaped the strings properly...
  (lambda args
    (show-help #f
               (cond
                ((eq? manual-symbol 'Ugens)
                 "Unit Generators")
                ((not manual-symbol)
                 #f)
                (else
                 (symbol->string manual-symbol))))))

(define (populate-help)
  (append-node the-help-tree
               (let ((nodes (stext->help-nodes
                             (texinfo->stext (find-help-file "tutorial.texi")) 2)))
                 (slot-set! nodes 'name "Tutorial")
                 nodes)
               #f)
  (append-node the-help-tree
               (let ((nodes (stext->help-nodes
                             (texinfo->stext (find-help-file "reference.texi")) 2)))
                 (slot-set! nodes 'name "Reference")
                 nodes)
               #f)

  (let* ((module-path (build-path
                       (if (not (eq? (string-ref (car (command-line)) 0) #\/))
                           ""
                           ss-data-dir)
                       "soundscrape"))
         (modules (find-modules module-path
                                (lambda (mod-name)
                                  ;; Ignore the g-wrap modules
                                  (string=?
                                   (substring (symbol->string (car (last-pair mod-name)))
                                              0 2)
                                   "gw"))))
         (parent (make <ss-help-node>
                   :name "Modules"
                   :value
                   '((stext "Modules")
                     (node "Modules")
                     "Section of modules, bla bla bla..."))))
    (append-node the-help-tree parent #f)
    (for-each
     (lambda (node) (append-node the-help-tree node parent))
     modules))
  
  (append-node the-help-tree (find-ugens) #f))

(add-hook! stext-buffer-xref-activated-hook show-help)
