;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;A lazy tree model. Instantiating this class will occasionally fail.
;;
;;; Code:

(define-module (soundscrape ui lazy-tree)
  :use-module (gnome gtk)
  :use-module (soundscrape util)
  :export (<ss-lazy-node> <ss-lazy-tree> append-node))

(define-class-with-docs <ss-lazy-node> ()
  "A node in a tree that supports lazy calculation of its value and/or
children."
  (name
   :init-keyword :name)
  (value
   :init-keyword :value
   :init-value #f)
  (make-value
   :init-keyword :make-value
   :init-value #f)
  (parent
   :init-keyword :parent
   :init-value #f)
  (children
   :init-keyword :children
   :init-value #f)
  (make-children
   :init-keyword :make-children
   :init-value #f))

(define-class-with-docs <ss-lazy-tree> (<guile-gtk-tree-model>)
  "A tree model composed of <ss-lazy-node> instances. Suitable for use
with @code{<gtk-tree-view>}."
  top-nodes)

(define-method (initialize (obj <ss-lazy-tree>) initargs)
  (next-method)
  (slot-set! obj 'top-nodes '()))

(define-method (on-get-flags (obj <ss-lazy-tree>))
  #f)

(define-method (on-get-n-columns (obj <ss-lazy-tree>))
  2) ;; name and value

(define-method (on-get-column-type (obj <ss-lazy-tree>) index)
  (case index
    ((0) gtype:gchararray)
    ((1) gtype:gboxed-scm)
    (else (error "Invalid index:" index))))

(define-method (on-get-iter (obj <ss-lazy-tree>) path)
  (let loop ((node #f) (path path))
    (if (null? path)
        node
        (let ((children (if node
                            (slot-ref node 'children)
                            (slot-ref obj 'top-nodes))))
          (if (null? children) ;; can be the case for path == (0)
              #f
              (loop (list-ref children (car path)) (cdr path)))))))

(define-method (on-get-path (obj <ss-lazy-tree>) iter)
  (let loop ((node iter) (path '()))
    (let ((parent (slot-ref node 'parent)))
      (if (not parent)
          (cons (list-index (slot-ref obj 'top-nodes) node) path)
          (loop parent
                (cons (list-index (slot-ref parent 'children) node) path))))))

(define-method (on-get-value (obj <ss-lazy-tree>) iter index)
  (case index
    ((0)
     (slot-ref iter 'name))
    ((1)
     (or (slot-ref iter 'value)
         (if (slot-ref iter 'make-value)
             (or
              (begin
                (slot-set! iter 'value ((slot-ref iter 'make-value)))
                (slot-ref iter 'value))
              (begin
                ;; make-value returned #f, meaning there's no value at
                ;; this node. We set make-value to be #f to indicate
                ;; this.
                (slot-set! iter 'make-value #f)
                ;; And return #f.
                #f))
             #f)))
    (else
     (error "Invalid index" index))))

(define-method (on-iter-next (obj <ss-lazy-tree>) iter)
  (let* ((parent (slot-ref iter 'parent))
         (siblings (if parent
                       (slot-ref parent 'children)
                       (slot-ref obj 'top-nodes)))
         (new-position (1+ (list-index siblings iter))))
    (and (< new-position (length siblings))
         (list-ref siblings new-position))))

(define-method (on-iter-children (obj <ss-lazy-tree>) parent)
  (if (and parent
           (not (slot-ref parent 'children))
           (slot-ref parent 'make-children))
      (slot-set! parent 'children ((slot-ref parent 'make-children))))
  (let ((children (if parent
                      (slot-ref parent 'children)
                      (slot-ref obj 'top-nodes))))
    (and children
         (car children))))

(define-method (on-iter-has-child (obj <ss-lazy-tree>) iter)
  (or (slot-ref iter 'children)
      (slot-ref iter 'make-children)))

(define-method (on-iter-n-children (obj <ss-lazy-tree>) iter)
  (or (and (on-iter-children obj iter) (length (slot-ref iter 'children)))
      0))

(define-method (on-iter-nth-child (obj <ss-lazy-tree>) parent n)
  (let* ((children (if parent
                       (and (on-iter-children obj parent)
                            (slot-ref parent 'children))
                       (slot-ref obj 'top-nodes))))
    (and children
         (< n (length children))
         (list-ref children n))))

(define-method (on-iter-parent (obj <ss-lazy-tree>) iter)
  (slot-ref iter 'parent))

;; To the on-* methods, the iter is just a scheme object. But outside
;; those methods it is a boxed type. So we need to call the Gtk+
;; notification functions (row-inserted, row-has-child-toggled) with the
;; boxed types, not with SCM values.
(define-method (append-node (obj <ss-lazy-tree>) node parent)
  (define (emit-signals path)
    (row-inserted obj path (get-iter obj path)))

  (let ((siblings (if parent
                      (slot-ref parent 'children)
                      (slot-ref obj 'top-nodes)))
        (path #f))
    (if parent
        (slot-set! parent 'children
                   (if siblings
                       (append siblings (list node))
                       (list node)))
        (slot-set! obj 'top-nodes (append siblings (list node))))
    (slot-set! node 'parent parent)
    (set! path (get-path obj (get-iter obj (on-get-path obj node))))
    (if (and parent (not siblings))
        (row-has-child-toggled obj
                               (reverse (cdr (reverse path)))
                               (iter-parent obj (get-iter obj path))))
    (emit-signals path)
    (get-iter obj path)))
