;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;The source editor window.
;;
;;; Code:

(define-module (soundscrape ui source-window)
  :use-module (gnome gtk)
  :use-module (gnome gtk gw-pango)
  :use-module (soundscrape util)
  :use-module (soundscrape ui help)
  :use-module (soundscrape ui util)
  :use-module (soundscrape ui about-window)
  :use-module (gnome gtk gw-glade)
  :use-module (gnome gtk gdk-event)
  :use-module (gnome gnome gw-source-view)

  :export (<ss-gui-source-window>))

(define this-module (current-module))

;; Source windows must be added both to this list and to
;; `child-windows'.
(define source-windows '())

(define (find-source-window widget)
  (let ((window (get-widget (glade-get-widget-tree widget) "source")))
    (let loop ((windows source-windows))
      (cond
       ((null? windows)
        #f)
       ((equal? (slot-ref (car windows) 'gtk-window) window)
        (car windows))
       (else
        (loop (cdr windows)))))))

(define (on-cut widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view))
         (clipboard (gtk-clipboard-get (gdk-atom-intern "CLIPBOARD" #f))))
    (cut-clipboard buf clipboard #t)))

(define (on-copy widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view))
         (clipboard (gtk-clipboard-get (gdk-atom-intern "CLIPBOARD" #f))))
    (copy-clipboard buf clipboard)))

(define (on-paste widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view))
         (clipboard (gtk-clipboard-get (gdk-atom-intern "CLIPBOARD" #f))))
    (paste-clipboard buf clipboard #f #t)))

(define (on-delete widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view)))
    (delete-selection buf #t #t)))

(define (on-select-all widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view)))
    (call-with-values (lambda () (get-bounds buf))
      (lambda (start end)
        (move-mark buf (get-selection-bound buf) start)
        (move-mark buf (get-insert buf) end)))))

(define (on-undo widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view)))
    (undo buf)))

(define (on-redo widget . args)
  (let* ((xml (glade-get-widget-tree widget))
         (view (get-widget xml "source-view"))
         (buf (get-buffer view)))
    (redo buf)))

(define (on-execute-buffer widget . args)
  (let* ((window (find-source-window widget))
         (buf (slot-ref window 'source-buffer)))
    (call-with-values (lambda () (get-bounds buf))
      (lambda (start end)
        (let ((port (open-input-string (get-text buf start end #t))))
          (let loop ((val *unspecified*))
            (let ((exp (read port)))
              (if (eof-object? exp)
                  (if (not (unspecified? val))
                      (begin
                        (newline)
                        (display (slot-ref window 'source-file))
                        (display " ==> ")
                        (display val)
                        (newline)))
                  (loop (eval exp (resolve-module '(guile-user))))))))))))

(define (on-open widget . args)
  (catch
   'cancel
   (lambda ()
     (let* ((source-window (find-source-window widget)))
       (open-source
        source-window
        (prompt-for-file-name
         (slot-ref source-window 'gtk-window)
         "Open File..."
         #f))))
   noop))

(define (on-save widget . args)
  (catch
   'cancel
   (lambda ()
     (let* ((source-window (find-source-window widget)))
       (save-source source-window #f)))
   noop))

(define (on-save-as widget . args)
  (catch
   'cancel
   (lambda ()
     (let* ((source-window (find-source-window widget)))
       (save-source source-window #t)))
   noop))

(define (on-close widget . args)
  (catch
   'cancel
   (lambda ()
     (let* ((source-window (find-source-window widget)))
       (close-window source-window)))
   noop))

(define (on-show-help . args)
  (show-help))

(define (on-about . args)
  (show-about-window))

(define (make-gtk-source-view)
  ;; called by glade's custom handler for the source window
  (let* ((manager (make <gtk-source-languages-manager>))
         (language (get-language-from-mime-type manager "text/x-scheme"))
         (buffer (make <gtk-source-buffer>
                   :language language
                   :check-brackets #t
                   :highlight #t
                   :max-undo-levels 100))
         (view (gtk-source-view-new-with-buffer buffer)))

    (modify-font view (pango-font-description-from-string "Monospace"))
    (set view 'auto-indent #t)
    (set view 'insert-spaces-instead-of-tabs #t)
    (set view 'smart-home-end #t)
    (set view 'tabs-width 2)
    view))

(define-class-with-docs <ss-gui-source-window> ()
  "A class for source windows."
  source-view
  source-buffer
  source-file-set?
  source-file
  gtk-window)

(define-method (set-title! (obj <ss-gui-source-window>))
  (set (slot-ref obj 'gtk-window)
       'title
       (apply
        string-append
        "soundscrape: "
        ((lambda (f)
           (if (get-modified (slot-ref obj 'source-buffer))
               (list f " *")
               (list f)))
         (slot-ref obj 'source-file))))
  (run-hook child-windows-changed-hook))

(define-method (get-title (obj <ss-gui-source-window>))
  (get (slot-ref obj 'gtk-window) 'title))

(define-method (present (obj <ss-gui-source-window>))
  (present (slot-ref obj 'gtk-window)))

(define make-untitled-name
  (let ((count 0))
    (lambda ()
      (set! count (1+ count))
      (string-append "untitled-" (number->string count)))))

(define-method (open-source (obj <ss-gui-source-window>) value)
  (catch
   'new-window
   (lambda ()
     (let ((buffer (slot-ref obj 'source-buffer)))
       (if (get-modified buffer)
           (begin
             (open-source (make <ss-gui-source-window>) value)
             (throw 'new-window)))
       (if value
           (begin
             (begin-not-undoable-action buffer)
             ;; FIXME: catch the file-does-not-exist case
             (set-text buffer (read-file value))
             (set-modified buffer #f)
             (end-not-undoable-action buffer))))
     (slot-set! obj 'source-file-set? (not (not value)))
     (slot-set! obj 'source-file (or value (make-untitled-name)))
     (set-title! obj))
   noop))

;; Might throw 'cancel, check yourself before wrecking same
(define-method (save-source (obj <ss-gui-source-window>) new-filename?)
  (let ((buffer (slot-ref obj 'source-buffer))
        (source-file-set? (slot-ref obj 'source-file-set?))
        (source-file (slot-ref obj 'source-file)))
    (if (or new-filename? (not source-file-set?))
        (begin
          (set! source-file (prompt-for-file-name (slot-ref obj 'gtk-window)
                                                  "Save File As..."
                                                  source-file))
          (slot-set! obj 'source-file-set? #t)
          (slot-set! obj 'source-file source-file)))
    (call-with-values (lambda () (get-bounds buffer))
      (lambda (start end)
        (let ((text (get-text buffer start end #t))
              ;; fixme: handle errors from open-file
              (file (open-file source-file "w")))
          (display text file)
          (close-port file))))
    (set-modified buffer #f)
    (set-title! obj)))

(define-method (close-window (obj <ss-gui-source-window>))
  (let ((buffer (slot-ref obj 'source-buffer)))
    (if (get-modified buffer)
        (let ((response (question
                         (format #f "~A \"~A\" ~A\n~A ~A"
                                 "Save changes to document"
                                 (slot-ref obj 'source-file)
                                 "before closing?"
                                 "If you close without saving,"
                                 "your changes will be discarded.")
                         '(no "Discard" "gtk-delete")
                         '(cancel "Don't close" "gtk-cancel")
                         '(ok "_Save" "gtk-save"))))
          (case response
            ((ok)
             (save-source obj #f))
            ((no)
             #f)
            (else
             (throw 'cancel))))))
  (destroy (slot-ref obj 'gtk-window)))

(define-method (initialize (obj <ss-gui-source-window>) initargs)
  (next-method)
  (let* ((window (ss-glade-get-widget this-module "soundscrape.glade" "source"))
         (xml (glade-get-widget-tree window))
         (view (get-widget xml "source-view"))
         (buffer (get-buffer view)))
    
    ;; This doesn't seem to work.
;     (set-events window #x3ffffe)

;     (connect window 'selection-notify-event
;              (lambda (w event)
;                (if (eq? (pk (gdk-event-selection:selection event)) 'CLIPBOARD)
;                    (begin
;                      (set (get-widget "menu-paste") 'sensitive #t)
;                      (set (get-widget "button-paste") 'sensitive #t)))
;                #f))
;     (connect window 'selection-clear-event
;              (lambda (w event)
;                (if (eq? (pk (gdk-event-selection:selection event)) 'CLIPBOARD)
;                    (begin
;                      (set (get-widget "menu-paste") 'sensitive #f)
;                      (set (get-widget "button-paste") 'sensitive #f)))
;                #f))

    (slot-set! obj 'gtk-window window)
    (slot-set! obj 'source-view view)
    (slot-set! obj 'source-buffer buffer)

    (set! source-windows (reverse (cons obj (reverse source-windows))))
    (set! child-windows (reverse (cons obj (reverse child-windows))))
    (run-hook child-windows-changed-hook)

    (connect window 'destroy
             (lambda (w)
               (set! source-windows (delq obj source-windows))
               (set! child-windows (delq obj child-windows))
               (run-hook child-windows-changed-hook)
               (slot-set! obj 'gtk-window #f)))

    (connect window 'delete-event
             (lambda (w e)
               (catch 'cancel
                      (lambda () (close-window obj))
                      noop)
               #t))
               
    (connect buffer 'mark-set
             (let ((old-sensitive #f))
               (lambda (buf . args)
                 (call-with-values (lambda () (get-selection-bounds buf))
                   (lambda (start end)
                     (let ((sensitive (if start #t #f)))
                       (if (not (eq? sensitive old-sensitive))
                           (begin
                             (set (get-widget xml "button-cut") 'sensitive sensitive)
                             (set (get-widget xml "button-copy") 'sensitive sensitive)
                             (set (get-widget xml "menu-cut") 'sensitive sensitive)
                             (set (get-widget xml "menu-copy") 'sensitive sensitive)
                             (set (get-widget xml "menu-delete") 'sensitive sensitive)
                             (set! old-sensitive sensitive)))))))))

    (connect buffer 'can-undo
             (lambda (buf can-undo?)
               (set (get-widget xml "button-undo") 'sensitive can-undo?)
               (set (get-widget xml "menu-undo") 'sensitive can-undo?)))
    (connect buffer 'can-redo
             (lambda (buf can-redo?)
               (set (get-widget xml "button-redo") 'sensitive can-redo?)
               (set (get-widget xml "menu-redo") 'sensitive can-redo?)))

    (connect buffer 'modified-changed
             (lambda (buf)
               (set-title! obj)))

    ;; Glade doesn't want to initialize these properly for some reason.
    (set (get-widget xml "button-cut") 'sensitive #f)
    (set (get-widget xml "button-copy") 'sensitive #f)
    (set (get-widget xml "button-undo") 'sensitive #f)
    (set (get-widget xml "button-redo") 'sensitive #f)

    (open-source obj #f)

    (show window)))
