;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Utility procedures for the soundscrape UI.
;;
;;; Code:

(define-module (soundscrape ui util)
  :use-module (gnome gtk)
  :use-module (gnome gobject primitives)
  :use-module (soundscrape util)
  :use-module (gnome gtk gw-glade)
  
  :export (ss-glade-get-widget
           unimplemented
           question
           prompt-for-file-name
           get-main-window
           the-main-window
           close-window
           child-windows
           child-windows-changed-hook
           populate-window-menu))

(define xml-list '())

;; because glade looks for handlers in the current module, we need to do
;; some hackery so that glade-xml-new is actually evaluated from the
;; proper module. Also, the xml will go away if we don't keep a hold on
;; it, so we store it in a list. (hacky!)
(define (ss-glade-get-widget module filename widget)
  "Load up a widget from a glade file. You only need to give the
filename, not the path. Signal handlers and custom widget handlers are
searched for in @var{module}."
  (let ((xml (eval
              `(,glade-xml-new (,find-data-file ,filename) ,widget #f)
              module)))
    (set! xml-list (cons xml xml-list))
    (signal-autoconnect xml module)
    (get-widget xml widget)))

(define (unimplemented . args)
  "Displays an error window when the user tries to access unimplemented
features."
  (let ((message (make <gtk-message-dialog>
                   #:message-type 'info
                   #:text "This feature is unimplemented.
But you know Scheme now, so why not file a patch?")))
    (run message)
    (destroy message)))

(define (question text . button-specs)
  "Ask a user a question. Each additional argument should have the form
@var{(response-symbol text-string stock-icon)}."
  (let ((message (make <gtk-message-dialog>
                   #:message-type 'question
                   #:buttons 'none
                   #:text text)))
    (for-each
     (lambda (spec)
       (let* ((response (gvalue-primitive-get
                         (make <gtk-response-type> :value (car spec))))
              (text (cadr spec))
              (icon (make <gtk-image> :stock (caddr spec)))
              (align (gtk-alignment-new 0.5 0.5 0 0))
              (hbox (make <gtk-hbox>))
              (button (make <gtk-button>))
              (label (make <gtk-label> :label text :use-underline #t :mnemonic-widget button)))
         (add button align)
         (add align hbox)
         (pack-start hbox icon #f #f 1)
         (pack-start hbox label #f #f 1)
         (show-all button)
         ;; doesn't work now, as there's no set-flags interface
         ;; (grab-default button) ;; default is the last button
         (grab-focus button) ;; Why doesn't this work? Ah well..
         (add-action-widget message button response)))
     button-specs)
    (let ((response (run message)))
      (destroy message)
      (genum->symbol (make <gtk-response-type> :value response)))))

(define-method (get-main-window (obj <gtk-widget>))
  "Returns the GtkWindow non-transient parent of a widget."
  (let loop ((toplevel (get-toplevel obj)))
    (let ((parent (get-parent toplevel)))
      (if (not parent)
          toplevel
          (loop (get-parent parent))))))

;; throws the 'cancel exception if the user deletes the window or
;; presses cancel
(define (prompt-for-file-name parent-window window-title default-file-name)
  "Asks the user for a file name. Might throw @code{cancel}."
  (let ((selection (gtk-file-selection-new window-title)))
    (if parent-window
        (set-transient-for selection parent-window))
    (set selection 'destroy-with-parent #t)
    (if default-file-name
        (set-filename selection default-file-name))
    (show selection)
    (let ((response (genum->symbol (make <gtk-response-type> :value (run selection))))
          (filename (get-filename selection)))
      (destroy selection)
      (case response
        ((ok) filename)
        ((cancel delete-event) (throw 'cancel))
        (else #f)))))

(define-with-docs the-main-window
  "The main soundscrape UI window."
  #f)

(define-generic-with-docs close-window
  "A generic function to close a child window. Takes one argument.
Allowed to throw @code{cancel}.")

(define-with-docs child-windows
  "A list of all child windows (source, control, listener, etc)."
  '())
(define-with-docs child-windows-changed-hook
  "A hook to run when @code{child-windows} changes. No arguments."
  (make-hook 0))

(define (populate-window-menu menu)
  "Append a list of all soundscrape windows to a menu @var{menu}."
  (define (on-choose-window title)
    (present
     (or (or-map
          (lambda (w)
            (if (equal? (get-title w) title)
                w
                #f))
          child-windows)
         (error "Unknown window!" title))))
  (let loop ((windows child-windows))
    (if (not (null? windows))
        (let* ((title (get-title (car windows)))
               (item (gtk-menu-item-new-with-label title)))
          (connect item 'activate
                   (lambda args (on-choose-window title)))
          (show item)
          (gtk-menu-shell-append menu item)
          (loop (cdr windows))))))
