;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;This module implements some miscellaneous useful procedures.
;;
;;; Code:

(define-module (soundscrape util)
  :use-module (oop goops)
  :use-module (soundscrape config)
  :use-module (ice-9 optargs)
  :use-module (ice-9 syncase)
  :use-module (ice-9 rw)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-13)
  :re-export (let-optional let-keywords))

(use-modules ((srfi srfi-1)
              :select (filter filter-map zip delete-duplicates)))

(define-public (class-is-a? class parent)
  "@code{#t} if @var{class} is descended from @var{parent}."
  (memq parent (class-precedence-list class)))

(define-public (until pred? list)
  "Returns the first elements of @var{list} that satisfy @var{pred?}."
  (if (or (eq? list '()) (pred? (car list)))
      '()
      (cons (car list) (until pred? (cdr list)))))

;; like let-keywords, but also allows positional arguments. this took me
;; like three fucking hours to get right -- nothing like macros to sap
;; the enthusiasm out of a body.
(defmacro-public let-positional-keywords (REST-ARG BINDINGS . BODY)
  (let ((positional (gensym))
        (keyword (gensym))
        (new-bindings (map (lambda (x) (if (list? x) x (list x #f))) BINDINGS)))
    `(let* ((,positional (until keyword? ,REST-ARG))
            (,keyword (list-tail ,REST-ARG (length ,positional))))
       (if (> (length ,positional) (length ',BINDINGS))
           (error "Too many positional arguments."))
       (let-optional ,positional ,new-bindings
         (let-keywords ,keyword #f ,(map (lambda (x) (list (car x) (or (car x) (cadr x)))) new-bindings) (begin ,@BODY))))))
(set-object-property! let-positional-keywords 'documentation
  "Like let-keywords, but also allows positional arguments.")

(defmacro define-with-docs args
  (define (syntax)
    (error "bad syntax" (list 'define-public args)))
  (define (get-name n)
    (cond
      ((symbol? n) n)
      ((pair? n) (get-name (car n)))
      (else (syntax))))
  (define (get-documentation n)
    (cond
      ((and (pair? n) (string? (car n))) (car n))
      (else (syntax))))
  (cond
    ((null? args)
     (syntax))
    (#t
     (let ((name (get-name (car args)))
	   (object-documentation (get-documentation (cdr args))))
       `(begin
	  (define ,(car args) ,@(cddr args))
	  (set-object-property! ,name 'documentation ,object-documentation)
          *unspecified*)))))
(set-object-property! define-with-docs 'documentation
  "Define a variable with documentation.")

(define-macro (define-generic-with-docs name documentation)
  `(define-with-docs ,name ,documentation
     (make-generic ',name)))
(set-object-property! define-generic-with-docs 'documentation
  "Define a generic function with documentation.")

(define-syntax define-class-with-docs
  (syntax-rules ()
    ((_ name supers docs . slots)
     (begin
       (define-class name supers . slots)
       (set-object-property! name 'documentation docs)
       (if #f #f)))))
(set-object-property! define-class-with-docs 'documentation
  "Define a class with documentation.")

(export define-with-docs define-generic-with-docs define-class-with-docs)

(define-public (if-empty string default)
  "Return @var{string}, unless it is equal to \"\", in which case return
@var{default}."
  (if (zero? (string-length string))
      default
      string))

(define-public lambda-docs
  (procedure->macro
   (lambda (args env)
     (local-eval `(lambda ,(cadr args)
                    ,(local-eval (caddr args) env)
                    ,@(cdddr args))
                 env))))
(set-object-property! lambda-docs 'documentation
  "Like @code{lambda}, but evaluates the first expression of the
function body. This way you can define the documentation of a procedure
semi-dynamically.")

(define-public (keyword->string keyword)
  "Convert @var{keyword} to a string."
  (symbol->string (keyword->symbol keyword)))
(define-public (string->keyword string)
  "Convert @var{string} to a keyword."
  (symbol->keyword (string->symbol string)))

(define-public (build-path . args)
  "Builds a path from elements in @var{args}, ignoring those equal to
\"\"."
  (string-join (filter (lambda (x) (not (equal? x ""))) args) "/"))

(define-public (append-uniq l1 l2)
  "Appends @var{l1} and @var{l2}, then removes duplicate items."
  (delete-duplicates! (append l1 l2)))

(define-public (wrapping-ref l n)
  "If @var{l} is a scalar, return it directly. Otherwise if it is a
list, reference the @var{n}th element, modulo the length of the list."
  (if (list? l)
      (list-ref l (modulo n (length l)))
      l))

(define-public (any-length l)
  "If @var{l} is a list, return its length. Otherwise return 1."
  (if (list? l)
      (length l)
      1))

(define-public (errorf str . args)
  "Throw an error as a formatted string."
  (error (apply format #f str args)))

(define-public (find-package-file installed-path uninstalled-path filename)
  "Find a soundscrape file @var{filename}. First look in
@var{installed-path}, then @var{uninstalled-path}, otherwise throwing an
error."
  (build-path
   (cond
    ((file-exists? (build-path installed-path filename))
     ss-data-dir)
    ((file-exists? (build-path uninstalled-path filename))
     (build-path (getcwd) uninstalled-path))
    (else
     (errorf "\nCouldn't find file ~A.
If you're running uninstalled, be sure to run from the top of the source directory."
             filename)))
   filename))

(define-public (find-data-file filename)
  "Find a soundscrape data file."
  (find-package-file ss-data-dir "src" filename))
(define-public (find-help-file filename)
  "Find a soundscrape help file."
  (find-package-file ss-data-dir "doc" filename))
 
;; Why aren't these part of the distro?
(define-public (file-size f)
  "Returns the size of a file @var{f}."
  (stat:size (stat f)))
(define-public (read-file f)
  "Returns the contents of a file @var{f} as a string."
  (let* ((size (file-size f))
         (str (make-string size #\space)))
    (with-input-from-file f
      (lambda () (if (eq? size (read-string!/partial str)) str "")))))

(define-public (list-join l infix)
  "Infixes @var{infix} into list @var{l}."
  (let loop ((ret '()) (l l))
    (cond
     ((null? l)
      (reverse ret))
     ((null? (cdr l))
      (loop (cons (car l) ret) (cdr l)))
     (else
      (loop (cons* infix (car l) ret) (cdr l))))))

(define-public (texinfo-escape str)
  "Escapes any illegal texinfo characters (currently @{, @}, and @@)."
  (define escaped-chars '(#\} #\{ #\@))
  (let loop ((in (string->list str)) (out '()))
    (if (null? in)
        (apply string (reverse out))
        (if (memq (car in) escaped-chars)
            (loop (cdr in) (cons* (car in) #\@ out))
            (loop (cdr in) (cons (car in) out))))))

(define-syntax re-export-module
  (syntax-rules ()
    ((_ module)
     (module-use! (module-public-interface (current-module))
                  (resolve-interface 'module)))))
(set-object-property! re-export-module 'documentation
  "Re-export the public interface of a module; used like
@code{use-modules}.")
(export re-export-module)
